/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "rose.h"
#include "Globals.h"
#include "Utils.h"
#include "MethodInfo.h"
#include "BlockClasses.h"
#include "RecursiveCallSet.h"
#include "SOATransform.h"

#include <iostream>
#include <fstream>
using namespace std;

using namespace SageBuilder;
using namespace SageInterface;


void BlockClasses::generate(int numNextBlocks) {

	//SageInterface::insertHeader("block.h", PreprocessingInfo::after, false, g_global);
	if (g_afterClassDecsPragma == NULL) {
		SageInterface::addTextForUnparser(g_global, "\n#include \"block.h\"\n", AstUnparseAttribute::e_before);
	} else {
		SageInterface::addTextForUnparser(g_afterClassDecsPragma, "\n#include \"block.h\"\n", AstUnparseAttribute::e_before);
	}

	ofstream file;
	file.open ("block.h");

	file << "#include <vector>\n"
			<< "#include <cassert>\n"
			<< "using namespace std;\n"
			<< "\n"
			<< getClassForwardDeclareString()
			<< "class _Point {\n"
			<< "public:\n";
	for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
		file << "\t" << getTypeString(param->get_type(), true) << " " << string(param->get_name()) << ";\n";
	}

	/*
	 * _Point constructor should not have additional params!  _SpliceNode.add() does not pass the additional param.
	 */
	vector<string> pointArgs, pointParams;
	for (int i = 0; i < g_methodInfo.recursiveMethodLoopVariantParams.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodLoopVariantParams[i];
		pointParams.push_back(getTypeString(param->get_type(), true) + " " + string(param->get_name()));
		pointArgs.push_back(string(param->get_name()));
	}

	file << "	_Point() {}\n"
			<< "	_Point(" << concatStrs(pointParams) << ");\n"
			<< "};\n"
			<< "\n"
			<< "_Point::_Point(" << concatStrs(pointParams) << ") {\n";
	for (int i = 0; i < pointArgs.size(); i++) {
		file << "	this->" << pointArgs[i] << " = " << pointArgs[i] << ";\n";
	}
	file << "}\n"
			<< "\n"
			<< "class _Block {\n"
			<< "public:\n"
			<< "	_Block();\n"
			<< "	~_Block();\n"
			<< "\n"
			<< "	void add(" << getParamsString(g_methodInfo.recursiveMethodAllLoopVariantParams) << ");\n";
	if (g_useSOA) {
		vector<string> params;
		params.push_back("_Block *block, int bi");
		ConstIteratorFor(i, g_methodInfo.recursiveMethodAdditionalLoopVariantParams, ParamVec) {
			SgInitializedName *param = *i;
			string paramStr = getTypeString(param->get_type()) + " " + param->get_name();
			params.push_back(paramStr);
		}
		file << "	void add(" << concatStrs(params) << ");\n";
		file << "	void copy_up(_Block *upBlock);\n";
		file << "	void copy_back();\n";
	}
	if (g_useSplice) {
		file << "	void add(const _Point &p);\n";
	}
	if (!g_useSOA) {
		file << "	_Point& get(int i) { return points[i]; }\n";
	}
	file << "	void recycle() { size = 0; }\n"
			<< "	bool is_full() { return size == max_block; }\n"
			<< "	bool is_empty() { return size == 0; }\n"
			<< "\n";
	if (g_useSOA) {
		for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
			SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
			string paramName = string(param->get_name());
			SgType *baseType = param->get_type();
			SgPointerType *ptype = isSgPointerType(baseType);
			SgReferenceType *rtype = isSgReferenceType(baseType);
			if (ptype == NULL && rtype == NULL) {
				file << "	" << getTypeString(baseType) << "* " << paramName << ";\n";
			} else {
				SgNamedType *ntype = ptype == NULL ? isSgNamedType(rtype->get_base_type()) : isSgNamedType(ptype->get_base_type());
				ROSE_ASSERT(ntype != NULL);
				PointAccessMap::iterator it = SOATransform::pointAccessMap.find(ntype);
				ROSE_ASSERT(it != SOATransform::pointAccessMap.end());
				PointAccess *pa = it->second;
				//cout << ntype->get_name() << "\t" << pa->allFields.size() << endl;
				for (int j = 0; j < pa->allFields.size(); j++) {
					PointField *pf = pa->allFields[j];
					string varName = paramName + "_" + pf->name;
					if (pf->subName != "") varName += "_" + pf->subName;
					file << "	" << pf->type << "* " << varName;
					if (pf->dim > 0) file << "[" << pf->dim << "]";
					file << ";\n";
				}
				if (pa->isPointerComparison) {
					file << "	" << getTypeString(ntype) << "** " << paramName << "_;\n";
				}
			}
		}
		file << "	int* map;\n";
	}
	file << "	_Point* points;\n"
			<< "	int size;\n"
			<< "\n";

	// 	static TriangleKDTree *root_instance;

	if (g_recursiveMemberFunction != NULL) {
		file << "	static " << string(g_recursiveMemberFunction->get_associatedClassDeclaration()->get_name()) << " *root_instance;\n";
	}

	// static TriangleKDTreeNode* root_node;

	for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
		file << "	static " << getTypeString(param->get_type()) << " root_" << string(param->get_name()) << ";\n";
	}

	file << "	static int max_block;\n"
			<< "};\n"
			<< "\n"
			<< "class _BlockSet\n"
			<< "{\n"
			<< "public:\n"
			<< "	_BlockSet() {}\n"
			<< "	~_BlockSet() {}\n"
			<< "\n"
			<< "	_Block *block;\n";
	for (int i = 0; i < numNextBlocks; i++) {
		file << "	_Block nextBlock" << i << ";\n";
	}
	file << "};\n"
			<< "\n"
			<< "class _BlockStack\n"
			<< "{\n"
			<< "public:\n"
			<< "	_BlockStack() {}\n"
			<< "	~_BlockStack();\n"
			<< "\n"
			<< "	_BlockSet* get(int i);\n"
			<< "\n"
			<< "	vector<_BlockSet *> items;\n"
			<< "};\n"
			<< "\n"
			<< "int _Block::max_block = " << g_blockSize << ";\n";

	// TriangleKDTreeNode* _Block::root_node;

	if (g_recursiveMemberFunction != NULL) {
		file << string(g_recursiveMemberFunction->get_associatedClassDeclaration()->get_name()) << " *_Block::root_instance;\n";
	}

	// TriangleKDTree *_Block::root_instance;

	for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
		file << getTypeString(param->get_type()) << " _Block::root_" << string(param->get_name()) << ";\n";
	}

	file << "\n"
			<< "_Block::_Block() {\n";
	if (g_useSOA) {
		for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
			SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
			string paramName = string(param->get_name());
			SgType *baseType = param->get_type();
			SgPointerType *ptype = isSgPointerType(baseType);
			SgReferenceType *rtype = isSgReferenceType(baseType);
			if (ptype == NULL && rtype == NULL) {
				file << "	" << paramName << " = new " << getTypeString(baseType) << "[max_block];\n";
			} else {
				SgNamedType *ntype = ptype == NULL ? isSgNamedType(rtype->get_base_type()) : isSgNamedType(ptype->get_base_type());
				ROSE_ASSERT(ntype != NULL);
				PointAccessMap::iterator it = SOATransform::pointAccessMap.find(ntype);
				ROSE_ASSERT(it != SOATransform::pointAccessMap.end());
				PointAccess *pa = it->second;

				for (int j = 0; j < pa->allFields.size(); j++) {
					PointField *pf = pa->allFields[j];
					string varName = paramName + "_" + pf->name;
					if (pf->subName != "") varName += "_" + pf->subName;
					if (pf->dim > 0) {
						file << "	for (int i = 0; i < " << pf->dim << "; i++) {\n";
						file << "		" << varName << "[i] = new " << pf->type << "[max_block];\n";
						file << "	}\n";
					} else {
						file << "	" << varName << " = new " << pf->type << "[max_block];\n";
					}
				}
				if (pa->isPointerComparison) {
					file << "	" << paramName << "_ = new " << getTypeString(ntype) << "*[max_block];\n";
				}
			}
		}

		file << "	map = new int[max_block];\n";
		file << "	points = NULL;\n";
	} else {
		file << "	points = new _Point[max_block];\n";
	}
	file << "	size = 0;\n"
			<< "}\n"
			<< "\n"
			<< "_Block::~_Block() {\n";
	if (g_useSOA) {

		for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
			SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
			string paramName = string(param->get_name());
			SgType *baseType = param->get_type();
			SgPointerType *ptype = isSgPointerType(baseType);
			SgReferenceType *rtype = isSgReferenceType(baseType);
			if (ptype == NULL && rtype == NULL) {
				file << "	delete " << paramName << ";\n";
			} else {
				SgNamedType *ntype = ptype == NULL ? isSgNamedType(rtype->get_base_type()) : isSgNamedType(ptype->get_base_type());
				ROSE_ASSERT(ntype != NULL);
				PointAccessMap::iterator it = SOATransform::pointAccessMap.find(ntype);
				ROSE_ASSERT(it != SOATransform::pointAccessMap.end());
				PointAccess *pa = it->second;
				for (int j = 0; j < pa->allFields.size(); j++) {
					PointField *pf = pa->allFields[j];
					string varName = paramName + "_" + pf->name;
					if (pf->subName != "") varName += "_" + pf->subName;
					if (pf->dim > 0) {
						file << "	for (int i = 0; i < " << pf->dim << "; i++) {\n";
						file << "		delete " << varName << "[i];\n";
						file << "	}\n";
					} else {
						file << "	delete " << varName << ";\n";
					}
				}
				if (pa->isPointerComparison) {
					file << "	delete " << paramName << "_;\n";
				}
			}
		}
		file << "	delete map;\n";
		file << "	if (points != NULL) {\n"
				<< "		delete points;\n"
				<< "	}\n";
	} else {
		file << "	delete [] points;\n";
	}
	file << "}\n"
			<< "\n";

	if (g_useSOA) {
		/*
void _Block::add(Point* target, float upperDist) {
	assert(size < max_block);
	target_closest_label[size] = target->closest_label;
	for (int i = 0; i < 7; i++) {
		target_coord[i][size] = target->coord[i];
	}
	target_id[size] = target->id;
	target_num_nodes_traversed[size] = target->num_nodes_traversed;
	target_tau[size] = target->tau;
	this->upperDist[size] = upperDist;
	if (points == NULL) {
		points = new _Point[max_block];
	}
	points[size].target = target;
	points[size].upperDist = upperDist;
	size++;
}
		 */
		file << "void _Block::add(" << getParamsString(g_methodInfo.recursiveMethodAllLoopVariantParams) << ") {\n"
				<< "	assert(size < max_block);\n";
		file << "	if (points == NULL) {\n"
				<< "		points = new _Point[max_block];\n"
				<< "	}\n";

		for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
			SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
			string paramName = string(param->get_name());
			SgType *baseType = param->get_type();
			SgPointerType *ptype = isSgPointerType(baseType);
			SgReferenceType *rtype = isSgReferenceType(baseType);
			if (ptype == NULL && rtype == NULL) {
				file << "	this->" << paramName << "[size] = " << paramName << ";\n";
			} else {
				SgNamedType *ntype = ptype == NULL ? isSgNamedType(rtype->get_base_type()) : isSgNamedType(ptype->get_base_type());
				string arrowOrDot = ptype == NULL ? "." : "->";
				ROSE_ASSERT(ntype != NULL);
				PointAccessMap::iterator it = SOATransform::pointAccessMap.find(ntype);
				ROSE_ASSERT(it != SOATransform::pointAccessMap.end());
				PointAccess *pa = it->second;
				for (int j = 0; j < pa->allFields.size(); j++) {
					PointField *pf = pa->allFields[j];
					string pointName = paramName + arrowOrDot + pf->name;
					if (pf->subName != "") pointName += "." + pf->subName;
					string varName = paramName + "_" + pf->name;
					if (pf->subName != "") varName += "_" + pf->subName;
					if (pf->dim > 0) {
						file << "	for (int i = 0; i < " << pf->dim << "; i++) {\n";
						file << "		" << varName << "[i][size] = " << pointName << "[i];\n";
						file << "	}\n";
					} else {
						file << "	" << varName << "[size] = " << pointName << ";\n";
					}
				}
				if (pa->isPointerComparison) {
					// p_[size] = p;
					file << "	" << paramName << "_[size] = ";
					if (rtype != NULL) file << "&"; // p_[size] = &p; for references
					file << paramName << ";\n";
				}
			}
			// is this necessary?...
			file << "	points[size]." << string(param->get_name()) << " = ";
			if (rtype != NULL) file << "&"; // p_[size] = &p; for references
			file << string(param->get_name()) << ";\n";
		}

		file << "	size++;\n"
				<< "}\n"
				<< "\n";


		/*
void _Block::add(_Block *block, int bi, float upperDist) {
	assert(size < max_block);
	target_closest_label[size] = block->target_closest_label[bi];
	for (int i = 0; i < 7; i++) {
		target_coord[i][size] = block->target_coord[i][bi];
	}
	target_id[size] = block->target_id[bi];
	target_num_nodes_traversed[size] = block->target_num_nodes_traversed[bi];
	target_tau[size] = block->target_tau[bi];
	this->upperDist[size] = upperDist;
	map[size] = bi;
	size++;
}
		 */
		vector<string> params;
		params.push_back("_Block *block, int bi");
		ConstIteratorFor(i, g_methodInfo.recursiveMethodAdditionalLoopVariantParams, ParamVec) {
			SgInitializedName *param = *i;
			string paramStr = getTypeString(param->get_type()) + " " + param->get_name();
			params.push_back(paramStr);
		}
		file << "void _Block::add(" << concatStrs(params) << ") {\n"
				<< "	assert(size < max_block);\n";
		for (int i = 0; i < g_methodInfo.recursiveMethodLoopVariantParams.size(); i++) {
			SgInitializedName *param = g_methodInfo.recursiveMethodLoopVariantParams[i];
			string paramName = string(param->get_name());
			SgType *baseType = param->get_type();
			SgPointerType *ptype = isSgPointerType(baseType);
			SgReferenceType *rtype = isSgReferenceType(baseType);
			if (ptype == NULL && rtype == NULL) {
				file << "	" << paramName << "[size] = block->" << paramName << "[bi];\n";
			} else {
				SgNamedType *ntype = ptype == NULL ? isSgNamedType(rtype->get_base_type()) : isSgNamedType(ptype->get_base_type());
				ROSE_ASSERT(ntype != NULL);
				PointAccessMap::iterator it = SOATransform::pointAccessMap.find(ntype);
				ROSE_ASSERT(it != SOATransform::pointAccessMap.end());
				PointAccess *pa = it->second;
				for (int j = 0; j < pa->allFields.size(); j++) {
					PointField *pf = pa->allFields[j];
					string varName = paramName + "_" + pf->name;
					if (pf->subName != "") varName += "_" + pf->subName;
					if (pf->dim > 0) {
						file << "	for (int i = 0; i < " << pf->dim << "; i++) {\n";
						file << "		" << varName << "[i][size] = block->" << varName << "[i][bi];\n";
						file << "	}\n";
					} else {
						file << "	" << varName << "[size] = block->" << varName << "[bi];\n";
					}
				}
				if (pa->isPointerComparison) {
					// p_[size] = block->p_[bi];
					file << "	" << paramName << "_[size] = block->" << paramName << "_[bi];\n";
				}
			}
		}
		ConstIteratorFor(i, g_methodInfo.recursiveMethodAdditionalLoopVariantParams, ParamVec) {
			string paramName = (*i)->get_name();
			file << "	this->" << paramName << "[size] = " << paramName << ";\n";
		}
		file << "	map[size] = bi;\n";
		file << "	size++;\n"
				<< "}\n"
				<< "\n";

/*
void _Block::copy_up(_Block *upBlock) {
	for (int i = 0; i < size; i++) {
		int to = map[i];
		upBlock->target_closest_label[to] = target_closest_label[i];
		upBlock->target_num_nodes_traversed[to] = target_num_nodes_traversed[i];
		upBlock->target_tau[to] = target_tau[i];
	}
}
 */
				file << "void _Block::copy_up(_Block *upBlock) {\n"

				<< "	for (int i = 0; i < size; i++) {\n"
				<< "		int to = map[i];\n";
		for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
			SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
			string paramName = string(param->get_name());
			SgPointerType *ptype = isSgPointerType(param->get_type());
			SgReferenceType *rtype = isSgReferenceType(param->get_type());
			if (ptype == NULL && rtype == NULL) continue;	// non-pointers don't need to copy_up, pass by value can't propagate back anyway
			SgNamedType *ntype = ptype == NULL ? isSgNamedType(rtype->get_base_type()) : isSgNamedType(ptype->get_base_type());
			PointAccessMap::iterator it = SOATransform::pointAccessMap.find(ntype);
			ROSE_ASSERT(it != SOATransform::pointAccessMap.end());
			PointAccess *pa = it->second;
			for (int j = 0; j < pa->writeFields.size(); j++) {
				PointField *pf = pa->writeFields[j];
				string varName = string(param->get_name()) + "_" + pf->name;
				if (pf->subName != "") varName += "_" + pf->subName;
				if (pf->dim > 0) {
					//TODO:
					//					file << "	for (int i = 0; i < " << pf->dim << "; i++) {\n";
					//					file << "		" << varName << "[i][size] = block->" << varName << "[i][bi];\n";
					//					file << "	}\n";
				} else {
					file << "		upBlock->" << varName << "[to] = " << varName << "[i];\n";
				}
			}
		}
		file << "	}\n"
				<< "}\n";
		file << "\n"


/*
void _Block::copy_back() {
	for (int i = 0; i < size; i++) {
		_Point& point = points[i];
		point.target->closest_label = target_closest_label[i];
		point.target->num_nodes_traversed = target_num_nodes_traversed[i];
		point.target->tau = target_tau[i];
	}
}
 */
				<< "void _Block::copy_back() {\n";
		file <<  "		for (int i = 0; i < size; i++) {\n"
				<< "			_Point& point = points[i];\n";

		for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
			SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
			string paramName = string(param->get_name());
			SgPointerType *ptype = isSgPointerType(param->get_type());
			SgReferenceType *rtype = isSgReferenceType(param->get_type());
			if (ptype == NULL && rtype == NULL) continue;	// non-pointers don't need to copy_up, pass by value can't propagate back anyway
			SgNamedType *ntype = ptype == NULL ? isSgNamedType(rtype->get_base_type()) : isSgNamedType(ptype->get_base_type());
			PointAccessMap::iterator it = SOATransform::pointAccessMap.find(ntype);
			ROSE_ASSERT(it != SOATransform::pointAccessMap.end());
			PointAccess *pa = it->second;
			for (int j = 0; j < pa->writeFields.size(); j++) {
				PointField *pf = pa->writeFields[j];
				string pointName = paramName + "->" + pf->name;
				if (pf->subName != "") pointName += "." + pf->subName;
				string varName = paramName + "_" + pf->name;
				if (pf->subName != "") varName += "_" + pf->subName;
				if (pf->dim > 0) {
					//TODO:
					//					file << "	for (int i = 0; i < " << pf->dim << "; i++) {\n";
					//					file << "		" << varName << "[i][size] = block->" << varName << "[i][bi];\n";
					//					file << "	}\n";
				} else {
					file << "			point." << pointName << " = " << varName << "[i];\n";
				}
			}
		}
		file << "		}\n"
				<< "	}\n";
	} else {
		file << "void _Block::add(" << getParamsString(g_methodInfo.recursiveMethodAllLoopVariantParams) << ") {\n"
				<< "	assert(size < max_block);\n";
		ConstIteratorFor(i, g_methodInfo.recursiveMethodAllLoopVariantParams, ParamVec) {
			SgInitializedName *param = *i;
			string refStr;
			if (isReferenceType(param->get_type())) refStr = "&";
			file << "	points[size]." << string(param->get_name()) << " = " << refStr <<
					string(param->get_name()) << ";\n";
		}

		file << "	size++;\n"
				<< "}\n"
				<< "\n";
	}
	if (g_useSplice) {
		file << "void _Block::add(const _Point &p) {\n"
				<< "	assert(size < max_block);\n";

		if (g_useSOA) {
			for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
				SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
				string paramName = string(param->get_name());
				SgType *baseType = param->get_type();
				SgPointerType *ptype = isSgPointerType(baseType);
				SgReferenceType *rtype = isSgReferenceType(baseType);
				if (ptype == NULL && rtype == NULL) {
					file << "	" << paramName << "[size] = p." << paramName << ";\n";
				} else {
					SgNamedType *ntype = ptype == NULL ? isSgNamedType(rtype->get_base_type()) : isSgNamedType(ptype->get_base_type());
					ROSE_ASSERT(ntype != NULL);
					PointAccessMap::iterator it = SOATransform::pointAccessMap.find(ntype);
					ROSE_ASSERT(it != SOATransform::pointAccessMap.end());
					PointAccess *pa = it->second;
					for (int j = 0; j < pa->allFields.size(); j++) {
						PointField *pf = pa->allFields[j];
						string pointName = paramName + "->" + pf->name;
						if (pf->subName != "") pointName += "." + pf->subName;
						string varName = paramName + "_" + pf->name;
						if (pf->subName != "") varName += "_" + pf->subName;
						if (pf->dim > 0) {
							file << "	for (int i = 0; i < " << pf->dim << "; i++) {\n";
							file << "		" << varName << "[i][size] = p." << pointName << "[i];\n";
							file << "	}\n";
						} else {
							file << "	" << varName << "[size] = p." << pointName << ";\n";
						}
					}
					if (pa->isPointerComparison) {
						// p_[size] = p;
						file << "	" << paramName << "_[size] = p." << paramName << ";\n";
					}
				}
			}
			file << "	if (points == NULL) {\n"
					<< "		points = new _Point[max_block];\n"
					<< "	}\n";
		}
		file << "	points[size] = p;\n";
		file << "	size++;\n"
				<< "}\n";
	}
	file << "\n"
			<< "_BlockStack::~_BlockStack() {\n"
			<< "	for (int i = 0; i < items.size(); i++) {\n"
			<< "		delete items[i];\n"
			<< "	}\n"
			<< "}\n"
			<< "\n"
			<< "_BlockSet* _BlockStack::get(int i) {\n"
			<< "	while (i >= items.size()) {\n"
			<< "		items.push_back(new _BlockSet());\n"
			<< "	}\n"
			<< "	return items[i];\n"
			<< "}\n";

	file.close();

}
