/*
   SIMTree, a framework to automatically vectorize tree traversal algorithms.
   Youngjoon Jo (yjo@purdue.edu)

   Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
   */

#include "rose.h"
#include "BlockTransform.h"
#include "Globals.h"
#include "MethodInfo.h"
#include "Utils.h"
#include "RecursiveCallSet.h"
#include "CommonTransform.h"
#include "SOATransform.h"

using namespace SageBuilder;
using namespace SageInterface;

SgFunctionSymbol *g_recurseBlockSym;
SgFunctionSymbol *g_recurseBlockSym_DF;
SgFunctionSymbol *g_recurseBlockSym_DF_i[2];

SgFunctionSymbol *g_nextMapRootSym;

string getLoopVarName(){
    string varName;
    SgForInitStatement * init_stmt = g_recursiveCallEnclosingForStmt->get_for_init_stmt();
    SgStatementPtrList& init_stmt_list = init_stmt->get_init_stmt();

    SgStatement* the_init = *init_stmt_list.begin();
    //Assume: always simple loop init (int i = 0 or i = 0)
    //To Simplify the following transformation, always assume it is in the form of "int i = 0";
    //i.e., the loop iterator is always defined with in the loop init
    //This can simplify the loop removal in DF execution.
    //cout << "Test for loop init: " << the_init->unparseToString() << endl;
    switch(the_init->variantT()){
        case V_SgVariableDeclaration:
            {
                //cout << "Init is a variable declaration" << endl;
                SgVariableDeclaration* decl = isSgVariableDeclaration(the_init);
                SgInitializedNamePtrList& name_list = decl->get_variables();
                SgInitializedName * name = *name_list.begin();
                ROSE_ASSERT(name);
                varName = name->get_name().getString();
                break;
            }
        case V_SgExprStatement:
            {
                //cout << "Init is an expression statement " << endl;
                SgExpression* the_init_expr = isSgExprStatement(the_init)->get_expression();
                SgAssignOp* ao = isSgAssignOp(the_init_expr);
                ROSE_ASSERT(ao);
                ROSE_ASSERT(isSgVarRefExp(ao->get_lhs_operand_i()));
                varName = isSgVarRefExp(ao->get_lhs_operand_i())->get_symbol()->get_name().getString();
                break;
            }
        default:
            {
                cout << the_init->variantT() << endl;
                cout << the_init->unparseToString() << endl;
                cout << "Reach unhandled case!" << endl;
                assert(false);
            }
    }
    return varName;
}


void BlockTransform::transformMapRootRecursive() {

    // create map root recursive function
    string funcName = g_recursiveFunction->get_name();
    SgFunctionParameterList* parameterList = buildFunctionParameterList();
    SgFunctionDeclaration *func;
    if (g_recursiveMemberFunction != NULL) {
        SgMemberFunctionType *memFuncType = buildMemberFunctionType(buildVoidType(), NULL, NULL, 0);
        func = buildDefiningMemberFunctionDeclaration(
                funcName + g_suffixMapRoot, memFuncType, parameterList, g_recursiveMemberFunction->get_class_scope());
    } else{
        func = buildDefiningFunctionDeclaration(
                funcName + g_suffixMapRoot, buildVoidType(), parameterList, g_global);
    }
    // save symbol to use in intermediary mapRoot functions
    g_nextMapRootSym = isSgFunctionSymbol(func->search_for_symbol_from_symbol_table());
    ROSE_ASSERT(g_nextMapRootSym != NULL);

    SgBasicBlock* funcBody = func->get_definition()->get_body();
    insertStatementBefore(g_recursiveFunction, func);

    // add params
    // _Block::root_node = node;

    for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
        SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
        SgInitializedName *newParam = buildInitializedName(param->get_name(), param->get_type());
        appendArg(parameterList, newParam);

        string varName = "root_" + param->get_name();
        SgExpression *assign = buildAssignOp(buildVarRefExp(varName, g_blockClassDec->get_definition()),
                buildVarRefExp(newParam));
        appendStatement(buildExprStatement(assign), funcBody);
    }

    // also save class instance if member function
    // ex) _Block::root_instance = this;

    if (g_recursiveMemberFunction != NULL) {
        string varName = "root_instance";
        SgClassDeclaration *classDec = g_recursiveMemberFunction->get_associatedClassDeclaration();
        SgClassSymbol* classSym = isSgClassSymbol(classDec->search_for_symbol_from_symbol_table());
        ROSE_ASSERT(classSym != NULL);
        SgExpression *assign = buildAssignOp(buildVarRefExp(varName, g_blockClassDec->get_definition()),
                buildThisExp(classSym));
        appendStatement(buildExprStatement(assign), funcBody);
    }

    // add non-defining forward declaration if necessary
    addForwardFuncDecIfNecessary(g_recursiveFunction, func);
}

void BlockTransform::transformMapRootIntermediary(const MethodInfo::IntermediaryInfo &minfo, bool last) {

    SgMemberFunctionDeclaration *memFunc = isSgMemberFunctionDeclaration(minfo.functionDec);

    // create map root intermediary function
    string funcName = minfo.functionDec->get_name();
    SgFunctionParameterList* parameterList = buildFunctionParameterList();
    SgFunctionDeclaration *func;
    if (memFunc != NULL) {
        SgMemberFunctionType *memFuncType = buildMemberFunctionType(buildVoidType(), NULL, NULL, 0);
        func = buildDefiningMemberFunctionDeclaration(
                funcName + g_suffixMapRoot, memFuncType, parameterList, memFunc->get_class_scope());
    } else{
        func = buildDefiningFunctionDeclaration(
                funcName + g_suffixMapRoot, buildVoidType(), parameterList, g_global);
    }

    SgBasicBlock* funcBody = func->get_definition()->get_body();
    insertStatementBefore(minfo.functionDec, func);

    // add params

    for (int i = 0; i < minfo.loopInvariantParams.size(); i++) {
        SgInitializedName *param = minfo.loopInvariantParams[i];
        SgInitializedName *newParam = buildInitializedName(param->get_name(), param->get_type());
        appendArg(parameterList, newParam);
    }

    // add non-defining forward declaration if necessary
    addForwardFuncDecIfNecessary(minfo.functionDec, func);


    // copy over all statements within original recursive function

    SgBasicBlock *orgFuncBody = minfo.functionDec->get_definition()->get_body();
    const SgStatementPtrList &orgStmts = orgFuncBody->get_statements();
    for (SgStatementPtrList::const_iterator i = orgStmts.begin(); i < orgStmts.end(); i++) {
        appendStatement(copyStatement(*i), funcBody);	// need full copy to avoid changing original function
    }

    // analyze loop variant statements

    set<string> loopVariantArgsPropagation;
    ConstIteratorFor(i, minfo.loopVariantParams, ParamVec) {
        SgInitializedName *param = *i;
        loopVariantArgsPropagation.insert(param->get_name());
    }
    vector<SgStatement *> recursionPointStmts; // empty
    vector<SgStatement *> toRemoveStmts;
    removeLoopVariantStmts(&toRemoveStmts, &loopVariantArgsPropagation, funcBody, recursionPointStmts);

    // remove loop variant statements

    for (vector<SgStatement *>::iterator i = toRemoveStmts.begin(); i != toRemoveStmts.end(); i++) {
        SgStatement *stmt = *i;
        removeStatement(stmt);
    }

    const SgExpressionPtrList &orgArgs = minfo.call->get_args()->get_expressions();
    SgExprListExp *args;
    if (last) {
        args = buildExprListExp();
        for (int i = 0; i < orgArgs.size(); i++) {
            if (vecContains(g_methodInfo.recursiveMethodAllLoopVariantParamsIdxs, i)) continue;
            args->append_expression(copyExpression(orgArgs[i]));
        }
    } else {
        args = getArgsExprList(minfo.loopInvariantArgs);
    }
    SgStatement *callStmt = copyStatement(getEnclosingStatement(minfo.call));
    appendStatement(callStmt, funcBody);
    Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(callStmt, V_SgFunctionCallExp);
    ROSE_ASSERT(calls.size() == 1);
    SgFunctionCallExp *call = isSgFunctionCallExp(calls[0]);
    call->set_args(args);
    SgExpression *callExp = call->get_function();
    SgDotExp *dot = isSgDotExp(callExp);
    if (dot != NULL) callExp = dot->get_rhs_operand();
    SgArrowExp *arrow = isSgArrowExp(callExp);
    if (arrow != NULL) callExp = arrow->get_rhs_operand();
    SgFunctionRefExp *funcRefExp = isSgFunctionRefExp(callExp);
    if (funcRefExp != NULL) funcRefExp->set_symbol(g_nextMapRootSym);
    SgMemberFunctionRefExp *memFuncRefExp = isSgMemberFunctionRefExp(callExp);
    if (memFuncRefExp != NULL) {
        SgMemberFunctionSymbol *memFuncSym = isSgMemberFunctionSymbol(g_nextMapRootSym);
        ROSE_ASSERT(memFuncSym != NULL);
        memFuncRefExp->set_symbol(memFuncSym);
    }
    ROSE_ASSERT(funcRefExp != NULL || memFuncRefExp != NULL);

    // save symbol to use in intermediary mapRoot functions
    g_nextMapRootSym = isSgFunctionSymbol(func->search_for_symbol_from_symbol_table());
    ROSE_ASSERT(g_nextMapRootSym != NULL);
}

void BlockTransform::transformMapRoot() {
    transformMapRootRecursive();
    for (int i = g_methodInfo.intermediaryInfos.size() - 1; i >= 0; i--) {
        transformMapRootIntermediary(*g_methodInfo.intermediaryInfos[i], i == g_methodInfo.intermediaryInfos.size() - 1);
    }
}


//Used only for the mapRootCall
SgExprListExp *getArgsExprListFromIndex() {
    SgExprListExp *args = buildExprListExp();
    assert(g_numCallSites == 1);
    SgFunctionCallExp *entry = g_intermediaryCallSites[g_numCallSites - 1];
    const SgExpressionPtrList &oargs = entry->get_args()->get_expressions();

    for (int i = 0; i < oargs.size(); i++) {
        if (vecContains(g_outerForLoopInvariantArgIdxs, i)) {
            SgExpression *arg = oargs[i];
            args->append_expression(copyExpression(arg));
        }
    }
    return args;
}


SgIfStmt* generateBFDFTogglingCall(bool isEntry, SgBasicBlock* funcBody, SgFunctionCallExp* orgCall,
        SgInitializedName * stackVarDec, SgInitializedName * depthVarDec){
    SgBasicBlock * true_body = buildBasicBlock();
    SgBasicBlock * false_body = buildBasicBlock();

    ROSE_ASSERT(orgCall != NULL);

    //Get args from original Call site
    SgExprListExp * BFargs = buildExprListExp();
    const SgExpressionPtrList &orgArgs = orgCall->get_args()->get_expressions();
    for (int j = 0; j < orgArgs.size(); j++) {
        if (vecContains(g_methodInfo.recursiveMethodAllLoopVariantParamsIdxs, j)) continue;
        BFargs->append_expression(copyExpression(orgArgs[j]));
    }

    if (isEntry){
        BFargs->append_expression(buildAddressOfOp(buildVarRefExp(stackVarDec)));
        BFargs->append_expression(buildVarRefExp(depthVarDec));
    }else{
        BFargs->append_expression(buildVarRefExp(stackVarDec));
        BFargs->append_expression(buildAddOp(buildVarRefExp(depthVarDec), buildIntVal(1)));
    }

    //Build the new Call site in true body
    SgFunctionCallExp *newCall_true = buildFunctionCallExp(g_recurseBlockSym, BFargs);
    appendStatement(buildExprStatement(newCall_true), true_body);

    //Build the new Call site in false body
    if (g_recursiveCallEnclosingForStmt){//Benchmarks like nqueens use "for loop" for each children
        //cout <<"Test for stmt: " << g_recursiveCallEnclosingForStmt->unparseToString() << endl;
        SgForStatement* newCallForStmt = isSgForStatement(copyStatement(g_recursiveCallEnclosingForStmt));
        ROSE_ASSERT(newCallForStmt);
        SgBasicBlock * newCallBody = buildBasicBlock();

        string varName = getLoopVarName();
        assert(!varName.empty());
        SgInitializedName *loopVarDec = buildInitializedName(varName, buildIntType());
        SgExprListExp * DFargs = isSgExprListExp(copyExpression(BFargs));
        DFargs->append_expression(buildVarRefExp(loopVarDec));
        SgFunctionCallExp * newCall_false = buildFunctionCallExp(g_recurseBlockSym_DF, DFargs);
        appendStatement(buildExprStatement(newCall_false), newCallBody);

        replaceStatement(newCallForStmt->get_loop_body(), newCallBody);
        appendStatement(newCallForStmt, false_body);
    }else{//Benchmarks like fib, have only left/right children
        for(int i = 0; i < g_num_recursiveCall; ++i){
            SgExprListExp* DFargs_i = isSgExprListExp(copyExpression(BFargs));
            SgFunctionCallExp* newCall_false_i = buildFunctionCallExp(g_recurseBlockSym_DF_i[i], DFargs_i);
            appendStatement(buildExprStatement(newCall_false_i), false_body);
        }
    }
    //Build the if condition
    SgExpression* lhs;
    if (isEntry){
        lhs = buildDotExp(buildVarRefExp("_block", g_global), buildVarRefExp("size",
                    g_blockClassDec->get_definition()));
    }else{
        lhs = buildArrowExp(buildVarRefExp("_nextBlock0", funcBody), buildVarRefExp("size",
                    g_blockClassDec->get_definition()));
    }
    SgExpression * condExpr = buildLessOrEqualOp(lhs, buildDivideOp(
                buildVarRefExp("_expandSize", g_global),
                buildVarRefExp("_minBlockSize", g_global)));
    SgExprStatement * condStmt = buildExprStatement(condExpr);

    //Build the final if statement for toggling code
    SgIfStmt * ifStmt = buildIfStmt(condStmt, true_body, false_body);

    return ifStmt;
}

void BlockTransform::transformEntry(){
    //Assume there is only one outside call-site
    cout << "Start to transform Entry" << endl;
    assert(g_numCallSites == 1);
    SgFunctionCallExp *call = g_intermediaryCallSites[g_numCallSites - 1];
    SgStatement *callStmt = getEnclosingStatement(call);

    //in code, predefine _expandSize, and _minBlockSize
    //_Block::max_block = _expandSize;
    //_BlockStack stack;
    //_Block _block;

    SgExpression * maxBlockAssignExpr = buildAssignOp(buildVarRefExp("max_block", g_blockClassDec->get_definition()),
            buildVarRefExp("_expandSize", g_global));
    insertStatementBefore(callStmt, buildExprStatement(maxBlockAssignExpr));

    SgVariableDeclaration *stackDec = buildVariableDeclaration("_stack", g_blockStackType);
    insertStatementBefore(callStmt, stackDec);

    SgVariableDeclaration *blockDec = buildVariableDeclaration("_block", g_blockType);
    insertStatementBefore(callStmt, blockDec);

    //_block.add(a,j);//Add point
    //int _depth = 0;
    //_stack.get(_depth)->block = &_block;

    //Debug:
    //cout << "Test: g_methodInfo outerForLoopVariantArgs size: " << g_methodInfo.outerForLoopVariantArgs.size() << endl;
    //cout << "Test: g_methodInfo outerForLoopVariantArgIdxs size: " << g_methodInfo.outerForLoopVariantArgIdxs.size() << endl;
    //for(int i = 0; i < g_methodInfo.outerForLoopVariantArgs.size(); ++i){
    //    cout <<"Test: for _block add: " << g_methodInfo.outerForLoopVariantArgs[i]->unparseToString() << endl;
    //}

    SgFunctionCallExp *blockAddCall = buildFunctionCallExp(buildDotExp(buildVarRefExp(blockDec),
                buildMemberFunctionRefExp(g_blockAddSym, false, false)),
            getArgsExprList(g_methodInfo.outerForLoopVariantArgs));
    insertStatementBefore(callStmt, buildExprStatement(blockAddCall));

    SgVariableDeclaration *depthDec = buildVariableDeclaration("_depth", buildIntType(),
            buildAssignInitializer(buildIntVal(0)));
    insertStatementBefore(callStmt, depthDec);

    SgExprListExp *stackGetArgs = buildExprListExp();
    stackGetArgs->append_expression(buildVarRefExp(depthDec));
    SgExpression *stackGet = buildDotExp(buildVarRefExp(stackDec),
            buildMemberFunctionRefExp(g_blockStackGetSym, false, false));
    SgFunctionCallExp *getCall = buildFunctionCallExp(stackGet, stackGetArgs);
    SgExpression *stackAssignBlock = buildAssignOp(
            buildArrowExp(getCall, buildVarRefExp("block", g_blockStackClassDec->get_definition())),
            buildAddressOfOp(buildVarRefExp(blockDec)));
    insertStatementBefore(callStmt, buildExprStatement(stackAssignBlock));

    SgInitializedName *depthVarDec = buildInitializedName("_depth", buildIntType ());
    SgInitializedName *stackVarDec = buildInitializedName("_stack", buildPointerType(g_blockStackType));
    SgIfStmt* ifStmt = generateBFDFTogglingCall(true, NULL, call, stackVarDec, depthVarDec);
    //insertStatementBefore(callStmt, ifStmt);
    replaceStatement(callStmt, ifStmt);
}

void addBlockCall(SgStatement *stmt, SgInitializedName *stackVarDec, SgInitializedName *depthVarDec,
        SgBasicBlock* funcBody, SgBasicBlock *ifBlock) {
    Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(stmt, V_SgFunctionCallExp);
    for (Rose_STL_Container<SgNode *>::iterator it = calls.begin(); it != calls.end(); it++) {
        SgFunctionCallExp *orgCall = isSgFunctionCallExp(*it);
        if (orgCall == NULL) continue;

        SgExprListExp *args = buildExprListExp();
        const SgExpressionPtrList &orgArgs = orgCall->get_args()->get_expressions();
        for (int j = 0; j < orgArgs.size(); j++) {
            if (vecContains(g_methodInfo.recursiveMethodAllLoopVariantParamsIdxs, j)) continue;
            args->append_expression(copyExpression(orgArgs[j]));
        }
        args->append_expression(buildVarRefExp(stackVarDec));
        args->append_expression(buildAddOp(buildVarRefExp(depthVarDec), buildIntVal(1)));
        SgFunctionCallExp *newCall = buildFunctionCallExp(g_recurseBlockSym, args);
        replaceExpression(orgCall, newCall);
    }

    appendStatement(stmt, ifBlock);
}

void addBlockCall_new(SgStatement *stmt, SgInitializedName *stackVarDec, SgInitializedName *depthVarDec,
        SgBasicBlock* funcBody, SgBasicBlock *ifBlock) {
    int count = 0;
    Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(stmt, V_SgFunctionCallExp);
    for (Rose_STL_Container<SgNode *>::iterator it = calls.begin(); it != calls.end(); it++) {
        SgFunctionCallExp *orgCall = isSgFunctionCallExp(*it);
        if (orgCall == NULL) continue;
        SgIfStmt* ifStmt = generateBFDFTogglingCall(false, funcBody, orgCall, stackVarDec, depthVarDec);
        appendStatement(ifStmt, ifBlock);
        count++;
    }

    if (count > 1) cout << "Warning: Generate wrong BFDFToggling Call!" << endl;
}

void BlockTransform::addRecursivePrototype() {
    // create block recursive function
    //Build for BF execution
    string funcName = g_recursiveFunction->get_name();
    SgFunctionParameterList* parameterList = buildFunctionParameterList();

    SgFunctionDeclaration *func;
    if (g_recursiveMemberFunction != NULL) {
        assert(false);//Should never reach here
        SgMemberFunctionType *memFuncType = buildMemberFunctionType(buildVoidType(), NULL, NULL, 0);
        func = buildDefiningMemberFunctionDeclaration(funcName + g_suffixBlock, memFuncType, parameterList, g_recursiveMemberFunction->get_class_scope());
    } else {
        func = buildDefiningFunctionDeclaration(funcName + g_suffixBlock, buildVoidType(), parameterList, g_global);
    }

    // add params
    for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
        SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
        appendArg(parameterList, buildInitializedName(param->get_name(), param->get_type()));
    }
    SgInitializedName *depthVarDec = buildInitializedName("_depth", buildIntType ());
    // in first phase don't have g_blockStackType, must fall back to using opaque type
    SgInitializedName *stackVarDec = buildInitializedName("_stack", buildPointerType(buildOpaqueType("_BlockStack", g_global)));
    appendArg(parameterList, stackVarDec);
    appendArg(parameterList, depthVarDec);

    addForwardFuncDecIfNecessary(g_recursiveFunction, func);

    prependStatement(buildNondefiningClassDeclaration("_BlockStack", g_global), g_global);


    //Build for DF execution
    if (g_recursiveCallEnclosingForStmt){//Benchmarks like nqueens use "for loop" for each children
        string funcName = g_recursiveFunction->get_name();
        SgFunctionParameterList* parameterList = buildFunctionParameterList();

        SgFunctionDeclaration *func;
        func = buildDefiningFunctionDeclaration(funcName + g_suffixBlock_DF, buildVoidType(), parameterList, g_global);

        // add params
        for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
            SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
            appendArg(parameterList, buildInitializedName(param->get_name(), param->get_type()));
        }
        SgInitializedName *depthVarDec = buildInitializedName("_depth", buildIntType ());
        // in first phase don't have g_blockStackType, must fall back to using opaque type
        SgInitializedName *stackVarDec = buildInitializedName("_stack", buildPointerType(buildOpaqueType("_BlockStack", g_global)));
        appendArg(parameterList, stackVarDec);
        appendArg(parameterList, depthVarDec);

        string varName = getLoopVarName();
        assert(!varName.empty());
        SgInitializedName *loopVarDec = buildInitializedName(varName, buildIntType());
        appendArg(parameterList, loopVarDec);
        addForwardFuncDecIfNecessary(g_recursiveFunction, func);
    }else{//Benchmarks like fib, have only left/right children
        string funcName = g_recursiveFunction->get_name();
        for(int i = 0; i < g_num_recursiveCall; ++i){
            SgFunctionParameterList* parameterList = buildFunctionParameterList();

            SgFunctionDeclaration *func;
            func = buildDefiningFunctionDeclaration(funcName + g_suffixBlock_DF_i[i], buildVoidType(), parameterList, g_global);

            // add params
            for (int ii = 0; ii < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); ii++) {
                SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[ii];
                appendArg(parameterList, buildInitializedName(param->get_name(), param->get_type()));
            }
            SgInitializedName *depthVarDec = buildInitializedName("_depth", buildIntType ());
            // in first phase don't have g_blockStackType, must fall back to using opaque type
            SgInitializedName *stackVarDec = buildInitializedName("_stack", buildPointerType(buildOpaqueType("_BlockStack", g_global)));
            appendArg(parameterList, stackVarDec);
            appendArg(parameterList, depthVarDec);

            addForwardFuncDecIfNecessary(g_recursiveFunction, func);
        }
    }

}

void BlockTransform::transformRecursive() {

    // create block recursive function
    string funcName = g_recursiveFunction->get_name();
    SgFunctionParameterList* parameterList = buildFunctionParameterList();
    SgFunctionDeclaration *func;
    if (g_recursiveMemberFunction != NULL) {
        SgMemberFunctionType *memFuncType = buildMemberFunctionType(buildVoidType(), NULL, NULL, 0);
        func = buildDefiningMemberFunctionDeclaration(funcName + g_suffixBlock, memFuncType, parameterList, g_recursiveMemberFunction->get_class_scope());
    } else {
        func = buildDefiningFunctionDeclaration(funcName + g_suffixBlock, buildVoidType(), parameterList, g_global);
    }

    SgBasicBlock* funcBody = func->get_definition()->get_body();
    insertStatementBefore(g_recursiveFunction, func);

    // add params

    for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
        SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
        appendArg(parameterList, buildInitializedName(param->get_name(), param->get_type()));
    }
    SgInitializedName *depthVarDec = buildInitializedName("_depth", buildIntType ());
    SgInitializedName *stackVarDec = buildInitializedName("_stack", buildPointerType(g_blockStackType));
    appendArg(parameterList, stackVarDec);
    appendArg(parameterList, depthVarDec);

    /*
     * 	_BlockSet *_set = _stack->get(_depth);
     _Block *_block = _set->block;
     */

    SgExprListExp *args = buildExprListExp();
    args->append_expression(buildVarRefExp(depthVarDec));
    SgExpression *stackGet = buildArrowExp(buildVarRefExp(stackVarDec),
            buildMemberFunctionRefExp(g_blockStackGetSym, false, false));
    SgFunctionCallExp *getCall = buildFunctionCallExp(stackGet, args);
    SgAssignInitializer *init = buildAssignInitializer(getCall);
    SgVariableDeclaration *blockSetVarDec = buildVariableDeclaration("_set", buildPointerType(g_blockSetType), init);
    appendStatement(blockSetVarDec, funcBody);

    init = buildAssignInitializer(buildArrowExp(buildVarRefExp(blockSetVarDec),
                buildVarRefExp("block", g_blockStackClassDec->get_definition())));
    SgVariableDeclaration *blockVarDec = buildVariableDeclaration("_block",
            buildPointerType(g_blockType), init);
    appendStatement(blockVarDec, funcBody);

    // add block for loop

    SgVariableDeclaration *biVarDec = buildVariableDeclaration(
            "_bi", buildIntType(), buildAssignInitializer(buildIntVal(0)), funcBody);
    SgExpression *biLtBlockSize = buildLessThanOp(buildVarRefExp(biVarDec),
            buildArrowExp(buildVarRefExp(blockVarDec), buildVarRefExp("size", g_blockClassDec->get_definition())));
    SgBasicBlock *blockBlock = buildBasicBlock();
    SgForStatement *blockFor = buildForStatement(biVarDec, buildExprStatement(biLtBlockSize),
            buildPlusPlusOp(buildVarRefExp(biVarDec)), blockBlock);
    appendStatement(blockFor, funcBody);

    if (!g_useSOA) {
        /* add point decs to start
         * 	_Point& _point = _block->get(_bi);
         Point *point = _point.point;
         */

        args = buildExprListExp();
        args->append_expression(buildVarRefExp(biVarDec));
        SgMemberFunctionRefExp *funcRef = buildMemberFunctionRefExp(g_blockGetSym, false, false);
        getCall = buildFunctionCallExp(buildArrowExp(buildVarRefExp(blockVarDec), funcRef), args);
        init = buildAssignInitializer(getCall);
        SgVariableDeclaration *pointDec = buildVariableDeclaration("_point", buildReferenceType(g_pointType), init);
        appendStatement(pointDec, blockBlock);

        for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
            SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
            SgExpression *initExp = buildDotExp(buildVarRefExp(pointDec), buildVarRefExp(param->get_name(), g_pointClassDec->get_definition()));
            SgType *type = param->get_type();
            // dereference pointer if var is reference type
            // ex) class Ray &ray =  *_point._Point::ray;
            if (isSgReferenceType(type)) {
                initExp = buildPointerDerefExp(initExp);
            }
            SgAssignInitializer *init = buildAssignInitializer(initExp);
            SgVariableDeclaration *subPointDec = buildVariableDeclaration(param->get_name(), param->get_type(), init);
            appendStatement(subPointDec, blockBlock);
        }
    }

    // copy over all statements within original recursive function

    SgBasicBlock *orgFuncBody = g_recursiveFunction->get_definition()->get_body();
    const SgStatementPtrList &orgStmts = orgFuncBody->get_statements();
    for (SgStatementPtrList::const_iterator i = orgStmts.begin(); i < orgStmts.end(); i++) {
        appendStatement(copyStatement(*i), blockBlock);	// need full copy to avoid changing original function
    }

    // change returns to continues

    Rose_STL_Container<SgNode *> returns = NodeQuery::querySubTree(blockBlock, V_SgReturnStmt);
    for (Rose_STL_Container<SgNode *>::iterator i = returns.begin(); i != returns.end(); i++) {
        replaceStatement(isSgReturnStmt(*i), buildContinueStmt());
    }

    // analyze call sets - must be done AFTER copy over all statements within original recursive function

    RecursiveCallSet rcs(blockBlock, g_methodInfo.recursiveMethodLoopVariantParams, g_recursiveFunction->get_name());
    cout << "num call sets: " << rcs.callSets.size() << endl;

    /*
     * 	_Block *_nextBlock0 = &_set->nextBlock0;
     _nextBlock0->recycle();
     */
    SgVariableDeclaration **nextBlockVarDecs = new SgVariableDeclaration*[rcs.callSets.size()];

    //	for (int i = 0; i < rcs.callSets.size(); i++) {
    for (int i = 0; i < 1; i++) { //insert only one "BEFORE block handling site"
        stringstream sstm;
        sstm << "nextBlock" << i;
        string nextBlockStr = sstm.str();
        string nextBlockStr2 = "_" + sstm.str();
        init = buildAssignInitializer(buildAddressOfOp(buildArrowExp(buildVarRefExp(blockSetVarDec),
                        buildVarRefExp(nextBlockStr, g_blockSetClassDec->get_definition()))));
        SgVariableDeclaration *nextBlockVarDec = buildVariableDeclaration(nextBlockStr2,
                buildPointerType(g_blockType), init);
        nextBlockVarDecs[i] = nextBlockVarDec;
        insertStatementAfter(blockVarDec, nextBlockVarDec, funcBody);

        SgExprListExp *args = buildExprListExp();
        SgMemberFunctionRefExp *funcRef = buildMemberFunctionRefExp(g_blockRecycleSym, false, false);
        SgFunctionCallExp *recycleCall = buildFunctionCallExp(buildArrowExp(buildVarRefExp(nextBlockVarDec), funcRef), args);
        insertStatementAfter(nextBlockVarDec, buildExprStatement(recycleCall), funcBody);
    }

    // replace call set stmt with block->add()

    for (int i = 0; i < rcs.callSets.size(); i++) {
        CallSet *cset = rcs.callSets[i];
        // assume that all calls have same loop variant args, and simply use first call
        SgFunctionCallExp *call = cset->calls[0];
        const SgExpressionPtrList &orgArgs = call->get_args()->get_expressions();
        SgExprListExp *args = buildExprListExp();
        ConstIteratorFor(j, g_methodInfo.recursiveMethodAllLoopVariantParamsIdxs, IntVec) {
            int idx = *j;
            args->append_expression(copyExpression(orgArgs[idx]));
        }
        stringstream sstm;
        //sstm << "_nextBlock" << cset->id;
        sstm << "_nextBlock0";
        SgFunctionCallExp *addCall = buildFunctionCallExp(buildArrowExp(
                    buildVarRefExp(sstm.str(), funcBody), buildMemberFunctionRefExp(g_blockAddSym, false, false)), args);
        replaceStatement(cset->stmts[0], buildExprStatement(addCall));
        // also need to remove non-first calls for non basic block call sets
        for (int j = 1; j < cset->stmts.size(); j++) {
            removeStatement(cset->stmts[j]);
        }
    }

    // add recursive calls on next block
    /**
     * 	if (!_next_block0->is_empty()) {
     _stack->get(_depth + 1)->block = _next_block0;
     correlation_search_block(node->left, _stack, _depth + 1);
     correlation_search_block(node->right, _stack, _depth + 1);
     }
     */
    //	for (int i = 0; i < rcs.callSets.size(); i++) {
    for (int i = 0; i < 1; i++) {//insert only one "AFTER block handling site"
        CallSet *cset = rcs.callSets[i];
        stringstream sstm;
        sstm << "_nextBlock" << cset->id;

        SgExpression *sizeGtZero = buildGreaterThanOp(buildArrowExp(
                    buildVarRefExp(sstm.str(), funcBody), buildVarRefExp("size", g_blockClassDec->get_definition())),
                buildIntVal(0));
        SgBasicBlock *ifBlock = buildBasicBlock();
        SgIfStmt *ifstmt = buildIfStmt(buildExprStatement(sizeGtZero), ifBlock, NULL);
        appendStatement(ifstmt, funcBody);

        SgExprListExp *args = buildExprListExp();
        args->append_expression(buildAddOp(buildVarRefExp(depthVarDec), buildIntVal(1)));
        SgFunctionCallExp *getCall = buildFunctionCallExp(stackGet, args);


        SgExpression *stackGetBlock = buildArrowExp(getCall,
                buildVarRefExp("block", g_blockSetClassDec->get_definition()));
        SgExpression *assign = buildAssignOp(stackGetBlock, buildVarRefExp(sstm.str(), funcBody));
        appendStatement(buildExprStatement(assign), ifBlock);

        for (int j = 0; j < cset->stmts.size(); j++) {
            assert(j == 0);//Only one call stmt
            cout << "This is original Call Stmt" << cset->stmts[j]->unparseToString() << endl;
            addBlockCall_new(cset->stmts[j], stackVarDec, depthVarDec, funcBody, ifBlock);
        }

    }


    delete [] nextBlockVarDecs;
}

SgArrowExp* findNextBlock0AddExpr(SgBasicBlock* basicBlock, int i){
    Rose_STL_Container<SgNode *> exprs = NodeQuery::querySubTree(basicBlock, V_SgArrowExp);
    for (Rose_STL_Container<SgNode *>::iterator it = exprs.begin(); it != exprs.end(); it++) {
        SgArrowExp* arrow = isSgArrowExp(*it);
        ROSE_ASSERT(arrow);

        //cout << "This is an arrow expression in DF function: " << arrow->unparseToString() << endl;
        SgVarRefExp* lhs = isSgVarRefExp(arrow->get_lhs_operand());
        SgMemberFunctionRefExp* rhs = isSgMemberFunctionRefExp(arrow->get_rhs_operand());

        if (!lhs || !rhs) continue;
        int left = lhs->get_symbol()->get_name().getString().compare("_nextBlock0");
        int right = rhs->get_symbol()->get_name().getString().compare(g_blockAddSym->get_name().getString());
        if (!left && !right) {
            //cout << "This is the _nextBlock0Add: " << arrow->unparseToString() << endl;

            if (i == 0) return arrow;
            else {
                i--;
                continue;
            }
        }
    }
}

SgIfStmt* findCallIfStmt(SgBasicBlock* funcBody){
    Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(funcBody, V_SgFunctionCallExp);
    for(Rose_STL_Container<SgNode *>::iterator it = calls.begin(); it != calls.end(); it++){
        SgFunctionCallExp* call = isSgFunctionCallExp(*it);
        ROSE_ASSERT(call);

        SgFunctionSymbol* symbol = call->getAssociatedFunctionSymbol();
        if (symbol->get_name().getString().compare(g_recurseBlockSym->get_name().getString())) continue;

        //is BF call
        SgNode* ifNode = getEnclosingStatement(call)->get_parent()->get_parent();
        SgIfStmt* ifStmt = isSgIfStmt(ifNode);

        ROSE_ASSERT(ifStmt);
        return ifStmt;
    }

    cout << "BF function call is not found!" << endl;
    cout << "We should never reach here!" << endl;
    assert(false);
}

//Generate DF execution for nqueens like applications
void BlockTransform::generateRecursive_DF(SgFunctionDeclaration * BFFunc){
    SgFunctionDeclaration * DFFunc = NULL;
    //Create DFFunc from BFFunc
    DFFunc = isSgFunctionDeclaration(copyStatement(BFFunc));
    ROSE_ASSERT(DFFunc);
    //Change name
    DFFunc->set_name(g_recursiveFunction->get_name() + g_suffixBlock_DF);

    //Change Parameter List
    string varName = getLoopVarName();
    assert(!varName.empty());
    SgInitializedName *loopVarDec = buildInitializedName(varName, buildIntType());
    appendArg(DFFunc->get_parameterList(), loopVarDec);

    //Remove for iteration outside add
    SgBasicBlock* funcBody = DFFunc->get_definition()->get_body();
    SgArrowExp* nextBlock0AddExpr = findNextBlock0AddExpr(funcBody, 0);
    SgScopeStatement* scopeStmt = findEnclosingLoop(getEnclosingStatement(nextBlock0AddExpr));
    SgForStatement* forStmt = isSgForStatement(scopeStmt);
    ROSE_ASSERT(forStmt);
    SgStatement* forBody = isSgStatement(forStmt->get_loop_body());
    replaceStatement(forStmt, forBody);

    if(!g_useReexpand){//Remove the toggling code in DF execution
        SgIfStmt * callIfStmt = findCallIfStmt(funcBody);
        replaceStatement(callIfStmt, callIfStmt->get_false_body());
    }

    insertStatementBefore(g_recursiveFunction, DFFunc);
}

//Generate DF execution for fib like applications
void BlockTransform::generateRecursive_DF_i(SgFunctionDeclaration * BFFunc, int i){
    SgFunctionDeclaration * DFFunc = NULL;
    //Create DFFunc from BFFunc
    DFFunc = isSgFunctionDeclaration(copyStatement(BFFunc));
    ROSE_ASSERT(DFFunc);
    //Change name
    DFFunc->set_name(g_recursiveFunction->get_name() + g_suffixBlock_DF_i[i]);

    //No need to Change Parameter List

    //Remove the other child (IfStmt outside add) -- handle nested if/else like binomial
    SgBasicBlock* funcBody = DFFunc->get_definition()->get_body();
    SgArrowExp* nextBlock0AddExpr = findNextBlock0AddExpr(funcBody, g_num_recursiveCall- 1 - i);
    ROSE_ASSERT(nextBlock0AddExpr);

    SgScopeStatement* outForScopeStmt = findEnclosingLoop(getEnclosingStatement(nextBlock0AddExpr));
    ROSE_ASSERT(outForScopeStmt);
    //cout << "This is the block for loop outside add: "<< outForScopeStmt->unparseToString() << endl;

    SgScopeStatement* scopeStmt = getScope(isSgNode(getEnclosingStatement(nextBlock0AddExpr)));
    //cout << "This is scope stmt: " << scopeStmt->unparseToString() << endl;
    SgScopeStatement* testScopeStmt = getScope(scopeStmt->get_parent());
    //cout << "This is the testScopeStmt: " << testScopeStmt->unparseToString() << endl;
    while(testScopeStmt->get_parent() != outForScopeStmt){
        scopeStmt = testScopeStmt;
        //cout << "This is scope stmt: " << scopeStmt->unparseToString() << endl;
        testScopeStmt = getScope(scopeStmt->get_parent());
        //cout << "This is the testScopeStmt: " << testScopeStmt->unparseToString() << endl;
    }
    //cout << "This is the scope type to be removed: " << scopeStmt->variantT() << endl;
    //cout << "This is the scope to be removed: " << scopeStmt->unparseToString() << endl;

    SgIfStmt* ifStmt = isSgIfStmt(scopeStmt);
    ROSE_ASSERT(ifStmt);
    removeStatement(ifStmt);

    if(!g_useReexpand){//Remove the toggling code in DF execution
        SgIfStmt * callIfStmt = findCallIfStmt(funcBody);
        replaceStatement(callIfStmt, callIfStmt->get_false_body());
    }

    insertStatementBefore(g_recursiveFunction, DFFunc);

}

//generate DF execution from BF execution
void BlockTransform::transformRecursive_DF() {
    //cout << "Can arrive here!~" << endl;

    //Get the BF recursion Function
    SgFunctionDeclaration * BFFunc = NULL;
    Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(g_global, V_SgFunctionDefinition);
    for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
        SgFunctionDefinition *funcDef = isSgFunctionDefinition(*i);
        SgFunctionDeclaration *funcDec = funcDef->get_declaration();
        string name = funcDec->get_name().getString();
        if (name.compare(g_recursiveFunction->get_name().getString() + g_suffixBlock) == 0){
            //cout <<  funcDec->unparseToString() << endl;
            BFFunc = funcDec;

        }
    }

    //Build the DF recursive functions
    if (g_recursiveCallEnclosingForStmt){//Benchmarks like nqueens use "for loop" for each children
        generateRecursive_DF(BFFunc);
    }else{//Benchmarks like fib, have only left/right children
        for (int i = 0; i < g_num_recursiveCall; ++i){
            generateRecursive_DF_i(BFFunc, i);
        }
    }
}

void BlockTransform::findClasses() {
    g_pointType = isSgNamedType(lookupNamedTypeInParentScopes("_Point", g_global));
    ROSE_ASSERT(g_pointType != NULL);
    g_blockType = isSgNamedType(lookupNamedTypeInParentScopes("_Block", g_global));
    ROSE_ASSERT(g_blockType != NULL);
    g_blockSetType = isSgNamedType(lookupNamedTypeInParentScopes("_BlockSet", g_global));
    ROSE_ASSERT(g_blockSetType != NULL);
    g_blockStackType = isSgNamedType(lookupNamedTypeInParentScopes("_BlockStack", g_global));
    ROSE_ASSERT(g_blockStackType != NULL);

    Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(g_global, V_SgClassDeclaration);
    for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
        SgClassDeclaration *classDec = isSgClassDeclaration(*i);
        if (classDec->get_definition() == NULL) continue;
        if (classDec->get_name() == "_Block") g_blockClassDec = classDec;
        else if (classDec->get_name() == "_BlockStack") g_blockStackClassDec = classDec;
        else if (classDec->get_name() == "_BlockSet") g_blockSetClassDec = classDec;
        else if (classDec->get_name() == "_Point") g_pointClassDec = classDec;
        else if (classDec->get_name() == g_recursiveTypeName) g_recursiveClassDec = classDec;
    }

    g_blockStackGetSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
                SgName("get"), g_blockStackClassDec->get_definition()));
    ROSE_ASSERT(g_blockStackGetSym != NULL);
    if (!g_useSOA) {
        g_blockGetSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
                    SgName("get"), g_blockClassDec->get_definition()));
        ROSE_ASSERT(g_blockGetSym != NULL);
    } else {
        g_blockCopyUpSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
                    SgName("copy_up"), g_blockClassDec->get_definition()));
        ROSE_ASSERT(g_blockCopyUpSym != NULL);
        g_blockCopyBackSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
                    SgName("copy_back"), g_blockClassDec->get_definition()));
        ROSE_ASSERT(g_blockCopyBackSym != NULL);
    }
    g_blockAddSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
                SgName("add"), g_blockClassDec->get_definition()));
    ROSE_ASSERT(g_blockAddSym != NULL);
    g_blockRecycleSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
                SgName("recycle"), g_blockClassDec->get_definition()));
    ROSE_ASSERT(g_blockRecycleSym != NULL);
}

SgFunctionDeclaration* searchFunctionDeclaration(string funcName){
    SgFunctionDeclaration * func = NULL;
    Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(g_global, V_SgFunctionDeclaration);
    for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
        SgFunctionDeclaration* funcDec = isSgFunctionDeclaration(*i);
        string name = funcDec->get_name().getString();
        if (name.compare(funcName) == 0){
            //cout <<  funcDec->unparseToString() << endl;
            func = funcDec;
        }
    }
    return func;
}

void BlockTransform::setupFuncsyms(){
    SgFunctionDeclaration* BFFunc = searchFunctionDeclaration(g_recursiveFunction->get_name().getString() + g_suffixBlock);
    g_recurseBlockSym = isSgFunctionSymbol(BFFunc->get_symbol_from_symbol_table());
    ROSE_ASSERT(g_recurseBlockSym);

    if (g_recursiveCallEnclosingForStmt){//Benchmarks like nqueens use "for loop" for each children
        SgFunctionDeclaration* DFFunc = searchFunctionDeclaration(g_recursiveFunction->get_name().getString() + g_suffixBlock_DF);
        g_recurseBlockSym_DF = isSgFunctionSymbol(DFFunc->get_symbol_from_symbol_table());
        ROSE_ASSERT(g_recurseBlockSym_DF);
    }else{//Benchmarks like fib, have only left/right children
        for(int i = 0; i < g_num_recursiveCall; ++i){
            SgFunctionDeclaration* DFFunc = searchFunctionDeclaration(g_recursiveFunction->get_name().getString() + g_suffixBlock_DF_i[i]);
            g_recurseBlockSym_DF_i[i] = isSgFunctionSymbol(DFFunc->get_symbol_from_symbol_table());
            ROSE_ASSERT(g_recurseBlockSym_DF_i[i]);
        }
    }
}

void BlockTransform::transform(bool useSplice) {
    //transformMapRoot();
    setupFuncsyms();
    transformRecursive();//Generate the BF execution code
    transformRecursive_DF(); //Generate the DF execution code
    transformEntry();
    cout << "We finished the transformation!" << endl;
}
