/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "rose.h"

#include "CommonTransform.h"
#include "Globals.h"
#include "Utils.h"
#include "Includes.h"

using namespace SageBuilder;
using namespace SageInterface;

SgFunctionSymbol *g_nextPrologueSym;
SgFunctionSymbol *g_nextEpilogueSym;

void CommonTransform::findRecursiveClassDefinition() {
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(g_global, V_SgClassDeclaration);
	for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
		SgClassDeclaration *classDec = isSgClassDeclaration(*i);
		if (classDec->get_definition() == NULL) continue;
		if (classDec->get_name() != g_recursiveTypeName) continue;
		g_recursiveClassDec = classDec;
		return;
	}
	ROSE_ASSERT(false);
}

void CommonTransform::markAfterClassDecPragma() {
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(g_global, V_SgPragmaDeclaration);
	for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
		SgPragmaDeclaration *pragma = isSgPragmaDeclaration(*i);
		if (pragma->get_pragma()->get_pragma() == "afterClassDecs") {
			g_afterClassDecsPragma = pragma;
		}
	}
}

void CommonTransform::proEpilogueAnalysis(SgBasicBlock *block, SgStatement *divideStmt,
		vector<SgStatement *> &prologue, vector<SgStatement *> &epilogue,
		vector<SgVariableDeclaration *>& usedInEpilogueVarDecs) {

	// divide block into prologue and epilogue

	const SgStatementPtrList &stmts = block->get_statements();
	vector<SgVariableDeclaration *> prologueVarDecs;
	bool divideStmtFound = false;
	for (SgStatementPtrList::const_iterator i = stmts.begin(); i != stmts.end(); i++) {
		SgStatement *stmt = *i;
		bool divideStmtFoundNow = false;
		if (divideStmt == stmt) {
			divideStmtFoundNow = true;
			divideStmtFound = true;
		}
		if (!divideStmtFoundNow) {
			if (!divideStmtFound) {
				SgVariableDeclaration *varDec = isSgVariableDeclaration(stmt);
				if (varDec != NULL) {
					//cout << "!!" << varDec->unparseToString() << endl;
					prologueVarDecs.push_back(varDec);
				}
				prologue.push_back(stmt);
			} else {
				epilogue.push_back(stmt);
			}
		}
	}

	// all local variables declared in prologue and used in epilogue need to be saved in IntermediaryState
	for (vector<SgVariableDeclaration *>::const_iterator i = prologueVarDecs.begin(); i != prologueVarDecs.end(); i++) {
		SgVariableDeclaration *varDec = *i;
		SgInitializedName* initName = *(varDec->get_variables().begin());
		string varName = initName->get_name();
		bool found = false;
		for (vector<SgStatement *>::const_iterator j = epilogue.begin(); j != epilogue.end(); j++) {
			if (hasVarUse(*j, varName)) {
				//cout << (*i)->unparseToString() << endl;
				usedInEpilogueVarDecs.push_back(*i);
				found = true;
				break;
			}
		}
		if (!found && hasVarUse(divideStmt, varName)) {	// also need to save if used in divideStmt itself
			usedInEpilogueVarDecs.push_back(*i);
		}
	}
}

void CommonTransform::transformProEpilogue(const MethodInfo::IntermediaryInfo &minfo, bool last) {
	SgFunctionDeclaration *orgFuncDec = minfo.functionDec;
	SgBasicBlock* orgFuncBody = orgFuncDec->get_definition()->get_body();

	// divide block into prologue and epilogue

	SgFunctionCallExp *call = minfo.call;
	SgStatement *callStmt = getEnclosingStatement(call);
	StmtVec prologue, epilogue;
	vector<SgVariableDeclaration *> usedInEpilogueVarDecs;
	CommonTransform::proEpilogueAnalysis(orgFuncBody, callStmt, prologue, epilogue, usedInEpilogueVarDecs);

	SgMemberFunctionDeclaration *memFunc = isSgMemberFunctionDeclaration(orgFuncDec);

	// create prologue, epilogue intermediary function
	// prologues should have void type, epilogues have original type

	SgFunctionParameterList* parameterList = buildFunctionParameterList();
	SgFunctionParameterList* epilogueParameterList = buildFunctionParameterList();
	const SgInitializedNamePtrList& params = minfo.functionDec->get_args();
	for (int i = 0; i < params.size(); i++) {
		SgInitializedName *param = params[i];
		ROSE_ASSERT(param != NULL);
		SgInitializedName *newParam = buildInitializedName(param->get_name(), param->get_type());
		SgInitializedName *epilogueNewParam = buildInitializedName(param->get_name(), param->get_type());
		appendArg(parameterList, newParam);
		appendArg(epilogueParameterList, epilogueNewParam);
	}
	const string funcName = minfo.functionDec->get_name();
	SgInitializedName *interBlockParam = buildInitializedName("_interBlock", buildPointerType(g_intermediaryBlockType));
	SgInitializedName *epilogueInterBlockParam = buildInitializedName("_interBlock", buildPointerType(g_intermediaryBlockType));
	appendArg(parameterList, interBlockParam);
	appendArg(epilogueParameterList, epilogueInterBlockParam);

	SgFunctionDeclaration *prologueFunc, *epilogueFunc;
	SgFunctionType *funcType = orgFuncDec->get_type();
	if (memFunc != NULL) {
		SgMemberFunctionType *memFuncType = buildMemberFunctionType(buildVoidType(), NULL, NULL, 0);
		prologueFunc = buildDefiningMemberFunctionDeclaration(
				funcName + g_suffixPrologue, memFuncType, parameterList, memFunc->get_class_scope());
		epilogueFunc = buildDefiningMemberFunctionDeclaration(
				// for some reason passing SgFunctionType results in empty function name
				//funcName + g_suffixEpilogue, new SgFunctionType(funcType), parameterList, memFunc->get_class_scope());
				funcName + g_suffixEpilogue, buildMemberFunctionType(funcType->get_return_type(), NULL, NULL, 0),
				epilogueParameterList, memFunc->get_class_scope());
	} else{
		prologueFunc = buildDefiningFunctionDeclaration(
				funcName + g_suffixPrologue, buildVoidType(), parameterList, g_global);
		epilogueFunc = buildDefiningFunctionDeclaration(
				//funcName + g_suffixEpilogue, new SgFunctionType(funcType), epilogueParameterList, g_global);
				funcName + g_suffixEpilogue, funcType->get_return_type(), epilogueParameterList, g_global);
	}

	insertStatementBefore(minfo.functionDec, prologueFunc);
	insertStatementBefore(minfo.functionDec, epilogueFunc);

	// add non-defining forward declaration if necessary
	addForwardFuncDecIfNecessary(minfo.functionDec, prologueFunc);
	addForwardFuncDecIfNecessary(minfo.functionDec, epilogueFunc);

	SgBasicBlock* prologueFuncBody = prologueFunc->get_definition()->get_body();
	SgBasicBlock* epilogueFuncBody = epilogueFunc->get_definition()->get_body();

	// construct prologue function

	// copy over prologues

	ConstIteratorFor(i, prologue, StmtVec) {
		appendStatement(copyStatement(*i), prologueFuncBody);	// need full copy to avoid changing original function
	}

	if (last) {
		// _interBlock->block->add(point);

		const SgExpressionPtrList &orgArgs = minfo.call->get_args()->get_expressions();
		SgExprListExp *args = buildExprListExp();
		for (int i = 0; i < orgArgs.size(); i++) {
			if (!vecContains(g_methodInfo.recursiveMethodAllLoopVariantParamsIdxs, i)) continue;
			args->append_expression(copyExpression(orgArgs[i]));
		}
		SgVarRefExp *interBlockVarRef = buildVarRefExp(interBlockParam, prologueFuncBody);
		SgVarRefExp *interBlockBlockVarRef = buildVarRefExp("block", g_intermediaryBlockClassDec->get_definition());
		SgExpression *interBlockAdd = buildArrowExp(buildArrowExp(interBlockVarRef, interBlockBlockVarRef),
				buildMemberFunctionRefExp(g_blockAddSym, false, false));
		SgFunctionCallExp *interBlockAddCall = buildFunctionCallExp(interBlockAdd, args);
		appendStatement(buildExprStatement(interBlockAddCall), prologueFuncBody);
	} else {
		SgExprListExp *args = buildExprListExp();
		const SgExpressionPtrList &argList = minfo.call->get_args()->get_expressions();
		for (int i = 0; i < argList.size(); i++) {
			args->append_expression(copyExpression(argList[i]));
		}
		args->append_expression(buildVarRefExp(interBlockParam, prologueFuncBody));

		SgStatement *callStmt = copyStatement(getEnclosingStatement(minfo.call));
		appendStatement(callStmt, prologueFuncBody);
		Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(callStmt, V_SgFunctionCallExp);
		ROSE_ASSERT(calls.size() == 1);
		SgFunctionCallExp *call = isSgFunctionCallExp(calls[0]);
		call->set_args(args);
		SgExpression *callExp = call->get_function();
		SgDotExp *dot = isSgDotExp(callExp);
		if (dot != NULL) callExp = dot->get_rhs_operand();
		SgArrowExp *arrow = isSgArrowExp(callExp);
		if (arrow != NULL) callExp = arrow->get_rhs_operand();
		SgFunctionRefExp *funcRefExp = isSgFunctionRefExp(callExp);
		if (funcRefExp != NULL) funcRefExp->set_symbol(g_nextPrologueSym);
		SgMemberFunctionRefExp *memFuncRefExp = isSgMemberFunctionRefExp(callExp);
		if (memFuncRefExp != NULL) {
			SgMemberFunctionSymbol *memFuncSym = isSgMemberFunctionSymbol(g_nextPrologueSym);
			ROSE_ASSERT(memFuncSym != NULL);
			memFuncRefExp->set_symbol(memFuncSym);
		}
		ROSE_ASSERT(funcRefExp != NULL || memFuncRefExp != NULL);
	}

	// construct epilogue function

	if (!last) {
		SgExprListExp *args = buildExprListExp();
		const SgExpressionPtrList &argList = minfo.call->get_args()->get_expressions();
		for (int i = 0; i < argList.size(); i++) {
			args->append_expression(copyExpression(argList[i]));
		}
		args->append_expression(buildVarRefExp(interBlockParam, prologueFuncBody));

		SgStatement *callStmt = copyStatement(getEnclosingStatement(minfo.call));
		appendStatement(callStmt, epilogueFuncBody);
		Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(callStmt, V_SgFunctionCallExp);
		ROSE_ASSERT(calls.size() == 1);
		SgFunctionCallExp *call = isSgFunctionCallExp(calls[0]);
		call->set_args(args);
		SgExpression *callExp = call->get_function();
		SgDotExp *dot = isSgDotExp(callExp);
		if (dot != NULL) callExp = dot->get_rhs_operand();
		SgArrowExp *arrow = isSgArrowExp(callExp);
		if (arrow != NULL) callExp = arrow->get_rhs_operand();
		SgFunctionRefExp *funcRefExp = isSgFunctionRefExp(callExp);
		if (funcRefExp != NULL) funcRefExp->set_symbol(g_nextEpilogueSym);
		SgMemberFunctionRefExp *memFuncRefExp = isSgMemberFunctionRefExp(callExp);
		if (memFuncRefExp != NULL) {
			SgMemberFunctionSymbol *memFuncSym = isSgMemberFunctionSymbol(g_nextEpilogueSym);
			ROSE_ASSERT(memFuncSym != NULL);
			memFuncRefExp->set_symbol(memFuncSym);
		}
		ROSE_ASSERT(funcRefExp != NULL || memFuncRefExp != NULL);
	}

	// copy over epilogues

	ConstIteratorFor(i, epilogue, StmtVec) {
		appendStatement(copyStatement(*i), epilogueFuncBody);	// need full copy to avoid changing original function
	}

	// save symbol to use in intermediary functions
	g_nextPrologueSym = isSgFunctionSymbol(prologueFunc->search_for_symbol_from_symbol_table());
	ROSE_ASSERT(g_nextPrologueSym != NULL);
	g_nextEpilogueSym = isSgFunctionSymbol(epilogueFunc->search_for_symbol_from_symbol_table());
	ROSE_ASSERT(g_nextEpilogueSym != NULL);

}

void CommonTransform::transformProEpilogues() {
	for (int i = g_methodInfo.intermediaryInfos.size() - 1; i >= 0; i--) {
		transformProEpilogue(*g_methodInfo.intermediaryInfos[i], i == g_methodInfo.intermediaryInfos.size() - 1);
	}
}

void CommonTransform::transform() {
	transformProEpilogues();
}


