/*
   SIMTree, a framework to automatically vectorize tree traversal algorithms.
   Youngjoon Jo (yjo@purdue.edu)

   Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
   */

#include "MethodInfo.h"
#include "Globals.h"
#include "Utils.h"

using namespace SageInterface;

//Modified one, find recursive variant for the call-site
//outside of the recursive function
void MethodInfo::findLoopVariantArgs(){
    //Assumption: We have only one call-site outside of the recursive function
    //This is true for all our benchmarks
    assert(g_numCallSites == 1);
	SgFunctionCallExp *entry = g_intermediaryCallSites[g_numCallSites - 1];
	const SgExpressionPtrList &args = entry->get_args()->get_expressions();

	for (int i = 0; i < args.size(); ++i) {
        //Assumption:The outside call-site must have all recursive variant in SgVarRefExp
        //or Cast_OP
        SgVarRefExp * arg = isSgVarRefExp(args[i]);
        if (arg == NULL){
            SgCastExp * casts = isSgCastExp(args[i]);
            if (casts == NULL) continue;
            arg = isSgVarRefExp(casts->get_operand());
        }
        if (arg == NULL) continue;

        bool isVariant = vecContains(recursiveMethodAllLoopVariantParamsIdxs, i);
		if (isVariant) {
            cout << "Outside call-site recursive variant: " << arg->unparseToString() << endl;
			outerForLoopVariantArgIdxs.push_back(i);
			outerForLoopVariantArgs.push_back(arg);
		} else {
            cout << "Outside call-site recursive invariant: " << arg->unparseToString() << endl;
			outerForLoopInvariantArgs.push_back(arg);
		}
	}

    //If arg is not in SgVarRefExp, use index for recursive InVariant
	for (int i = 0; i < args.size(); ++i) {
        SgExpression* arg = isSgExpression(args[i]);
        assert(arg);

        bool isVariant = vecContains(recursiveMethodAllLoopVariantParamsIdxs, i);
		if (isVariant) {
            //cout << "arg type: " << arg->variantT() << endl;
            cout << "Not SgVarRefExp, Outside call-site recursive variant: " << arg->unparseToString() << endl;
		} else {
            //cout << "arg type: " << arg->variantT() << endl;
            cout << "Not SgVarRefExp, Outside call-site recursive invariant: " << arg->unparseToString() << endl;
			g_outerForLoopInvariantArgIdxs.push_back(i);
		}
	}

}

void MethodInfo::analyze() {
    g_numCallSites = g_intermediaryCallSites.size();
    //cout << "MethodInfo::analyze() - g_numCallSites: " << g_numCallSites << endl;
    analyzeRecursiveMethod();
    findLoopVariantArgs();
}

//Check Recursive Variant for Benchmarks like nqueens with outerloop
//NOTE: Difference between this and the commented one is to handle something like b[i], or a+b
bool isLoopVariant(const SgVarRefExp *var, const Rose_STL_Container<SgNode *> &loopLocals) {
    for (Rose_STL_Container<SgNode *>::const_iterator l = loopLocals.begin(); l != loopLocals.end(); ++l) {
        SgVariableDeclaration *loopLocal = isSgVariableDeclaration(*l);
        const SgInitializedNamePtrList &vars = loopLocal->get_variables();
        for (SgInitializedNamePtrList::const_iterator j = vars.begin(); j != vars.end(); ++j) {
            SgInitializedName *localName = *j;
            string localStr = localName->get_name();
            string varStr = var->get_symbol()->get_name();

            //cout << "localStr: " << localStr << " varStr: " << varStr << endl;
            if (localStr == varStr) return true;
        }
    }
    return false;
}

//TODO: A more complicated data flow analysis is needed
//Cannot handle alias case
bool areTheSame(const SgExpression* expr, const SgExpression* expr_j){
    if (expr->variantT() != expr_j->variantT()) return false;
    if (expr->unparseToString().compare(expr_j->unparseToString()) != 0) return false;

    return true;
}

void MethodInfo::findRecursiveVariant(vector<SgFunctionCallExp*> & allRecursiveCalls){
    assert(allRecursiveCalls.size());
    //cout << "This is allRecursiveCalls.size()" << allRecursiveCalls.size() << endl;

    if (allRecursiveCalls.size() == 1){//like nqueens
        //Assumption: the recursive call-site must be in a for-loop
        SgFunctionCallExp* call = allRecursiveCalls[0];
        SgScopeStatement* scopeStmt = findEnclosingLoop(getEnclosingStatement(call));
        SgForStatement* forStmt = isSgForStatement(scopeStmt);
        assert(forStmt);
        if(forStmt == NULL) return;
        g_recursiveCallEnclosingForStmt = forStmt;
        //cout << forStmt->unparseToString() << endl;

        const SgExpressionPtrList args = call->get_args()->get_expressions();
        Rose_STL_Container<SgNode *> loopLocals = NodeQuery::querySubTree(forStmt, V_SgVariableDeclaration);

        for(int i = 0; i < args.size(); ++i){
            const SgExpression* arg = args[i];
            bool isVariant = false;

            //cout << arg->unparseToString() << endl;
            //cout << arg->variantT() << endl;

            switch(arg->variantT()){
                case V_SgVarRefExp:
                    //Simple Variable
                    { const SgVarRefExp *arg1 = isSgVarRefExp(arg);
                        isVariant = isVariant || isLoopVariant(arg1, loopLocals);
                        break;
                    }

                case V_SgCastExp:
                    //Expressions like (j + 1)
                    {
                        const SgBinaryOp* tmp_arg = isSgBinaryOp(isSgCastExp(arg)->get_operand());
                        const SgExpression * lhs = tmp_arg->get_lhs_operand_i();
                        const SgExpression * rhs = tmp_arg->get_rhs_operand_i();

                        if (isSgVarRefExp(lhs)){
                            const SgVarRefExp *argl = isSgVarRefExp(lhs);
                            isVariant = isVariant || isLoopVariant(argl, loopLocals);
                        }

                        if (isSgVarRefExp(rhs)){
                            const SgVarRefExp *argr = isSgVarRefExp(rhs);
                            isVariant = isVariant || isLoopVariant(argr, loopLocals);
                        }
                        break;
                    }
                case V_SgAddOp:
                    //Expressions like j + 1
                    {
                        const SgAddOp* tmp_arg = isSgAddOp(arg);
                        const SgExpression * lhs = tmp_arg->get_lhs_operand_i();
                        const SgExpression * rhs = tmp_arg->get_rhs_operand_i();

                        if (isSgVarRefExp(lhs)){
                            const SgVarRefExp *argl = isSgVarRefExp(lhs);
                            isVariant = isVariant || isLoopVariant(argl, loopLocals);
                        }

                        if (isSgVarRefExp(rhs)){
                            const SgVarRefExp *argr = isSgVarRefExp(rhs);
                            isVariant = isVariant || isLoopVariant(argr, loopLocals);
                        }
                        break;
                    }
                case V_SgPntrArrRefExp:
                    //Expressions like b[i]
                    {
                        const SgPntrArrRefExp* tmp_arg = isSgPntrArrRefExp(arg);
                        const SgExpression * lhs = tmp_arg->get_lhs_operand_i();
                        const SgExpression * rhs = tmp_arg->get_rhs_operand_i();

                        if (isSgVarRefExp(lhs)){
                            const SgVarRefExp *argl = isSgVarRefExp(lhs);
                            isVariant = isVariant || isLoopVariant(argl, loopLocals);
                        }

                        if (isSgVarRefExp(rhs)){
                            const SgVarRefExp *argr = isSgVarRefExp(rhs);
                            isVariant = isVariant || isLoopVariant(argr, loopLocals);
                        }
                        break;
                    }

                default:
                    {
                        //cout << arg->variantT() << endl;
                        //cout << arg->unparseToString() << endl;
                        cout << "Reach unhandled case!" << endl;
                        assert(false);
                    }
            }

            const SgInitializedNamePtrList &params = g_recursiveFunction->get_parameterList()->get_args();
            SgInitializedName *param = params[i];
            //TODO: Add into proper global arrays for further analysis and transformation
            if (isVariant){
                cout <<"Recursive Variant Variables: " << arg->unparseToString() << endl;
                recursiveMethodLoopVariantParams.push_back(param);
                recursiveMethodAllLoopVariantParams.push_back(param);
                recursiveMethodLoopVariantParamIdxs.push_back(i);
                recursiveMethodAllLoopVariantParamsIdxs.push_back(i);
            }else{
                recursiveMethodAllLoopInvariantParamsIncludingNode.push_back(param);
                recursiveMethodAllLoopInvariantParams.push_back(param);
            }
        }
        } else{//like fib
            //There are multiple independent recursive call-site

            SgFunctionCallExp* call = allRecursiveCalls[0];
            const SgExpressionPtrList args = call->get_args()->get_expressions();
            vector<bool> isInVariant(args.size(), true);

            for (int j = 1; j < allRecursiveCalls.size(); ++j){
                SgFunctionCallExp* call_j = allRecursiveCalls[j];
                const SgExpressionPtrList args_j = call_j->get_args()->get_expressions();

                assert(args.size() == args_j.size());
                for (int i = 0; i < args.size(); ++i){
                    isInVariant[i] = isInVariant[i] && areTheSame(args[i], args_j[i]);
                }
            }

            const SgInitializedNamePtrList &params = g_recursiveFunction->get_parameterList()->get_args();
            for (int i = 0; i < args.size(); ++i){
                SgInitializedName *param = params[i];
                //TODO: Add into proper global arrays for further analysis and transformation
                if (!isInVariant[i]){
                    cout <<"Recursive Variant Variables: " << args[i]->unparseToString() << endl;
                    recursiveMethodLoopVariantParams.push_back(param);
                    recursiveMethodAllLoopVariantParams.push_back(param);
                    recursiveMethodLoopVariantParamIdxs.push_back(i);
                    recursiveMethodAllLoopVariantParamsIdxs.push_back(i);
                }else{
                    recursiveMethodAllLoopInvariantParamsIncludingNode.push_back(param);
                    recursiveMethodAllLoopInvariantParams.push_back(param);
                }
            }
        }
    }

    void MethodInfo::analyzeRecursiveMethod(){
        vector<SgFunctionCallExp*> allRecursiveCalls;

        Rose_STL_Container<SgNode *> allCalls = NodeQuery::querySubTree(g_recursiveFunction, V_SgFunctionCallExp);
        for (Rose_STL_Container<SgNode *>::iterator l = allCalls.begin(); l!= allCalls.end(); ++l){
            SgFunctionCallExp *call = isSgFunctionCallExp(*l);

            SgFunctionSymbol *sym = call->getAssociatedFunctionSymbol();
            if (sym == NULL) continue;
            string callName = sym->get_name();
            string decName = g_recursiveFunction->get_name();
            if (callName != decName) continue;

            cout << "RecursiveCalls inside recursive function: " << call->unparseToString() << endl;
            allRecursiveCalls.push_back(call);
        }

        g_num_recursiveCall = allRecursiveCalls.size();
        findRecursiveVariant(allRecursiveCalls);
    }


