/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "RecursiveStructureAnalysis.h"
#include "Utils.h"
#include "Globals.h"

using namespace SageInterface;

bool findEnclosingLoop(SgProject *project, SgFunctionDeclaration *functionDec) {
	g_intermediaryCallSites.clear();
	g_intermediaryFunctions.clear();
	Rose_STL_Container<SgNode *> allCalls = NodeQuery::querySubTree(project, V_SgFunctionCallExp);
	for (Rose_STL_Container<SgNode *>::iterator l = allCalls.begin(); l != allCalls.end(); ++l) {
		SgFunctionCallExp *call = isSgFunctionCallExp(*l);

		// this could be a way to skip over builtin functions, but there are far too many included functions anyway
		//		string callName = call->get_function()->unparseToString();
		//		cout << callName << endl;
		//		if (callName.find("__builtin_") != string::npos) continue;

		SgFunctionSymbol *sym = call->getAssociatedFunctionSymbol();
		if (sym == NULL) continue;
		string callName = sym->get_name();
		string decName = functionDec->get_name();
		if (callName != decName) continue;
		if (getEnclosingFunctionDeclaration(call) == functionDec) continue;

        cout << "This is call outside recursive function: "
            << call->unparseToString() << endl;

        //Decide whether the recursive call we are going to transform
        //By pragma "recursionCallEntry"
        SgPragmaDeclaration *pragma = isSgPragmaDeclaration(getPreviousStatement(getEnclosingStatement(call)));
        if (pragma == NULL || pragma->get_pragma()->get_pragma() != "recursionCallEntry") {
            cout << "  but not annotated recursionCallEntry" << endl;
            return false;
        }

		// is recursive call site outside recursive function
		g_intermediaryCallSites.push_back(call);
		g_intermediaryFunctions.push_back(functionDec);

        return true;//Get the g_intermediaryCallSites, Skip the rest

		/*SgNode *node = call;
		while (node != NULL) {
			SgForStatement *forStmt = isSgForStatement(node);
			if (forStmt != NULL) {
				cout << "enclosing for loop found!" << endl;
				SgPragmaDeclaration *pragma = isSgPragmaDeclaration(getPreviousStatement(forStmt));
				if (pragma == NULL || pragma->get_pragma()->get_pragma() != "parallelForOnTree") {
					cout << "  but not annotated parallelForOnTree" << endl;
					return false;
				}
				//cout << forStmt->unparseToString() << endl;
				g_enclosingForStmt = forStmt;
				return true;
			}
			SgFunctionDeclaration *intermediaryFuncDec = isSgFunctionDeclaration(node);
			if (intermediaryFuncDec != NULL) {
				bool found = false;
				for (Rose_STL_Container<SgNode *>::iterator i = allCalls.begin(); i != allCalls.end(); ++i) {
					SgFunctionCallExp *intermediaryCall = isSgFunctionCallExp(*i);
					SgFunctionSymbol *intermediarySym = intermediaryCall->getAssociatedFunctionSymbol();
					if (intermediarySym == NULL) continue;
					string callName = intermediarySym->get_name();
					string decName = intermediaryFuncDec->get_name();
					if (callName != decName) continue;
					// assume intermediary functions have single path to recursive entry, and take first call
					// this is just for simplicity of implementation
					node = intermediaryCall;
					g_intermediaryCallSites.push_back(intermediaryCall);
					g_intermediaryFunctions.push_back(intermediaryFuncDec);
					cout << "    intermediary call: " << intermediaryCall->unparseToString() << endl;
					found = true;
					break;
				}
				if (!found) return false;
			}
			else {
				SgNode *parent = node->get_parent();
				if (node == parent) break;
				node = parent;
			}
		}*/
	}
	return false;
}

bool checkRecursiveFunction(SgProject *project, SgFunctionDeclaration *functionDec) {
	string decName = functionDec->get_name();
	if (decName.find("operator") != string::npos) return false;
	Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(functionDec, V_SgFunctionCallExp);
	for (Rose_STL_Container<SgNode*>::iterator j = calls.begin(); j != calls.end(); ++j) {
		SgFunctionCallExp *call = isSgFunctionCallExp(*j);
		SgFunctionSymbol *sym = call->getAssociatedFunctionSymbol();
		if (sym == NULL) continue;

		// getAssociatedFunctionDeclaration can refer to different instances for function prototypes
		// fall back to comparing function names
		//if (call->getAssociatedFunctionDeclaration() != functionDec) continue;
		string callName = sym->get_name();

		if (callName != decName) continue;

		// is recursive function
		cout << "candidate recursive call: " << decName << "\t" << call->unparseToString() << endl;

	    //Bin: Modified for PLDI work -- Remove the recursive structure finder
		const SgExpressionPtrList &arglist = call->get_args()->get_expressions();

		for (SgExpressionPtrList::const_iterator k = arglist.begin(); k != arglist.end(); ++k) {
			/*SgExpression *arg = *k;
			SgCastExp *cast = isSgCastExp(arg);
			if (cast != NULL) {
				arg = cast->get_operand();
			}
			SgArrowExp *arrow = isSgArrowExp(arg);
			if (arrow == NULL) continue;
			SgExpression *lhs = arrow->get_lhs_operand();
			SgExpression *rhs = arrow->get_rhs_operand();
			SgType *lhsType = getBaseType(lhs->get_type());
			SgType *rhsType = getBaseType(lhs->get_type());

			if (lhsType != rhsType) continue;
			// has recursive class arg

			cout <<  "  candidate recursive class arg: " << lhs->unparseToString() << endl;*/

			//Bin: Modified for PLDI work -- Remove the loop finder, just find the intermediarycall
            if (!findEnclosingLoop(project, functionDec)) continue;

			g_recursiveFunction = functionDec;
			g_recursiveMemberFunction = isSgMemberFunctionDeclaration(g_recursiveFunction);

            /*g_recursiveType = isSgPointerType(lhs->get_type());
			ROSE_ASSERT(g_recursiveType != NULL);
			SgNamedType *nt = isSgNamedType(g_recursiveType->get_base_type());
			ROSE_ASSERT(nt != NULL);
			g_recursiveTypeName = nt->get_name();*/
			return true;
		}
	}
	return false;
}

bool RecursiveStructureAnalysis::findRecursiveTraversals(SgProject *project) {
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(project, V_SgFunctionDefinition);
	for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
		SgFunctionDefinition *funcDef = isSgFunctionDefinition(*i);
		SgFunctionDeclaration *funcDec = funcDef->get_declaration();
		//cout <<  funcDec->unparseToString() << endl;
		if (checkRecursiveFunction(project, funcDec)) return true;
	}
	return false;
}
