/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "SOATransform.h"

#include "Globals.h"
#include "MethodInfo.h"
#include "Utils.h"
#include "Includes.h"

using namespace SageBuilder;
using namespace SageInterface;

PointAccessMap SOATransform::pointAccessMap;
SoAFunctionInfoMap SOATransform::soaFunctionMap;

struct IntBinOpPair {
	int pointVarIndex;
	SgBinaryOp *op;

	IntBinOpPair(int i, SgBinaryOp *a) : pointVarIndex(i), op(a) {}
};

void updateFieldAccess(PointAccess *pa, string field, int flag) {
	map<string,int>::iterator it = pa->fields.find(field);
	if (it == pa->fields.end()) {
		pa->fields[field] = flag;
	} else {
		it->second = it->second | flag;
	}
}

void addPointField(PointAccess *pa, int flag, PointField *pf) {
	if (flag & readViaPoint) {
		pa->readFields.push_back(pf);
	}
	if (flag & writeViaPoint) {
		pa->writeFields.push_back(pf);
	}
	if (flag & writeViaPoint || flag & readViaPoint) {
		pa->allFields.push_back(pf);
	}
}

string getAccessDescription(int flag) {
	string ret;
	if (flag & readViaPoint) ret += "readViaPoint ";
	if (flag & writeViaPoint) ret += "writeViaPoint ";
	if (flag & readViaNode) ret += "readViaNode ";
	if (flag & writeViaNode) ret += "writeViaNode ";
	return ret;
}

bool isSOAType(SgType *type) {
	return isSgTypeInt(type) || isSgTypeFloat(type);
}

void SOATransform::markPointerComparison(SgExpression *exp) {
	SgVarRefExp *expVar = isSgVarRefExp(exp);
	if (expVar == NULL) return;
	SgNamedType *lhsNtype = isSgNamedType(isSgPointerType(expVar->get_type())->get_base_type());
	if (lhsNtype == NULL) return;
	PointAccessMap::iterator it = pointAccessMap.find(lhsNtype);
	if (it == pointAccessMap.end()) return;
	PointAccess *pa = it->second;
	pa->isPointerComparison = true;
}

void SOATransform::fieldAccessAnalysisEntry() {
	vector<string> pointVars, nodeVars;
	for (int i = 0; i < g_methodInfo.recursiveMethodLoopVariantParams.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodLoopVariantParams[i];
		pointVars.push_back(param->get_name());
		SgType *type = param->get_type();
		SgPointerType *ptype = isSgPointerType(type);
		SgReferenceType *rtype = isSgReferenceType(type);
		if (ptype != NULL) {
			type = ptype->get_base_type();
			SgNamedType *ntype = isSgNamedType(type);
			if (ntype != NULL) {
				if (pointAccessMap.count(ntype) == 0) {
					pointAccessMap[ntype] = new PointAccess;
				}
			}
		} else if (rtype != NULL) {
			type = rtype->get_base_type();
			SgNamedType *ntype = isSgNamedType(type);
			if (ntype != NULL) {
				if (pointAccessMap.count(ntype) == 0) {
					pointAccessMap[ntype] = new PointAccess;
				}
			}
		}
	}
	nodeVars.push_back(g_nodeParam->get_name());

	SoAFunctionInfo *soaFunctionInfo = new SoAFunctionInfo;
	soaFunctionMap[g_recursiveFunction] = soaFunctionInfo;
//	soaFunctionInfo->orgPointVars = pointVars;
	soaFunctionInfo->pointVars = pointVars;

	fieldAccessAnalysis(g_recursiveFunction, pointVars, nodeVars, 0);
}

void SOATransform::fieldAccessAnalysis(SgFunctionDeclaration *funcDec, const vector<string>& pointVars, const vector<string>& nodeVars, int depth) {
	for (int i = 0; i < depth; i++) cout << " ";
	cout << "SOATransform::fieldAccessAnalysis: " << funcDec->get_name() << endl;

	// propagate var decs
	vector<string> propagatedPointVars = pointVars;
	vector<string> propagatedNodeVars = nodeVars;
	Rose_STL_Container<SgNode *> varDecs = NodeQuery::querySubTree(funcDec, V_SgVariableDeclaration);
	for (Rose_STL_Container<SgNode *>::iterator i = varDecs.begin(); i != varDecs.end(); i++) {
		SgVariableDeclaration *varDec = isSgVariableDeclaration(*i);
		SgInitializedName *var;
		SgAssignInitializer *init = getVarDecInit(varDec, &var);
		if (init == NULL) continue;
		Rose_STL_Container<SgNode *> arrowExps = NodeQuery::querySubTree(init, V_SgArrowExp);
		for (Rose_STL_Container<SgNode *>::iterator j = arrowExps.begin(); j != arrowExps.end(); j++) {
			SgArrowExp *arrow = isSgArrowExp(*j);
			SgExpression *lhs = arrow->get_lhs_operand();
			if (exprContainsVarName(lhs, propagatedPointVars)) {
				propagatedPointVars.push_back(var->get_name());
			}
			if (exprContainsVarName(lhs, propagatedNodeVars)) {
				propagatedNodeVars.push_back(var->get_name());
			}
		}
	}

	// set accesses from arrows or reference dots

	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(funcDec, V_SgBinaryOp);
	for (Rose_STL_Container<SgNode *>::iterator j = nodes.begin(); j != nodes.end(); j++) {
		SgArrowExp *arrow = isSgArrowExp(*j);
		SgDotExp *dot = isSgDotExp(*j);
		SgBinaryOp *binOp = isSgBinaryOp(*j);
		if (arrow == NULL && dot == NULL) continue;

		SgExpression *lhs = binOp->get_lhs_operand();
		SgExpression *rhs = binOp->get_rhs_operand();
		SgVarRefExp *lhsVar = isSgVarRefExp(lhs);

		if (lhsVar == NULL) continue;
		string lhsName = lhsVar->get_symbol()->get_name();
		//cout << "lhs: " << lhsName << "\t" << lhsVar->get_type()->class_name() << endl;
		SgReferenceType *refType = isSgReferenceType(lhsVar->get_type());
		if (dot != NULL && refType == NULL) continue;


		SgNode *parent = binOp->get_parent();
		SgNode *subParent = binOp;
		if (isSgDotExp(parent)) { // to handle sub fields, class within class
			subParent = parent;
			parent = parent->get_parent();
		}
		//cout << parent->class_name() << " " << parent->unparseToString() << endl;
		bool isWrite = false;
		if (isSgPlusPlusOp(parent) || isSgMinusMinusOp(parent) ) {
			isWrite = true;
		} else if (isSgAssignOp(parent) || isSgCompoundAssignOp(parent)) {
			SgBinaryOp *assign = isSgBinaryOp(parent);
			if (subParent == assign->get_lhs_operand()) {	// subParent is lhs and hence a write
				isWrite = true;
			}
		}

		// if arrow, lhsVar must be pointer since it is lhs of arrow
		SgType *baseType = arrow == NULL ? refType->get_base_type() : isSgPointerType(lhsVar->get_type())->get_base_type();
		SgModifierType *mtype = isSgModifierType(baseType);
		if (mtype != NULL) {
			// handle modifiers, e.g., 'const' in  const Point* a
			baseType = mtype->get_base_type();
		}
		SgNamedType *lhsNtype = isSgNamedType(baseType);
		if (lhsNtype == NULL) continue;
		PointAccessMap::iterator it = pointAccessMap.find(lhsNtype);
		if (it == pointAccessMap.end()) continue;
		PointAccess *pa = it->second;
		SgVarRefExp *rhsVar = isSgVarRefExp(rhs);
		string rhsName = rhsVar->get_symbol()->get_name();
		//		cout << "rhs: " << rhsName << endl;
		//		if (rhsName == "coord") {
		//			ConstIteratorFor(i, *pointVars, set<string>) {
		//				cout << *i << " ";
		//			}
		//			cout << endl;
		//			int bb = 0;
		//		}

		if (exprContainsVarName(lhs, propagatedPointVars)) {
			//cout << "\t\tupdate point access: " << lhsNtype->get_name() << "\t" << rhsName << "\t" << isWrite << endl;
			updateFieldAccess(pa, rhsName, readViaPoint);
			if (isWrite) updateFieldAccess(pa, rhsName, writeViaPoint);
		}
		if (exprContainsVarName(lhs, propagatedNodeVars)) {
			//cout << "\t\tupdate node access: " << lhsNtype->get_name() << "\t" << rhsName << "\t" << isWrite << endl;
			updateFieldAccess(pa, rhsName, readViaNode);
			if (isWrite) updateFieldAccess(pa, rhsName, writeViaNode);
		}
	}

	// check for actual pointer comparisons, e.g., if ((p!=node->point))  in Barnes-Hut, need to save 'p' also
	nodes = NodeQuery::querySubTree(funcDec, V_SgNotEqualOp);
	for (Rose_STL_Container<SgNode *>::iterator i = nodes.begin(); i != nodes.end(); i++) {
		SgNotEqualOp *op = isSgNotEqualOp(*i);
		markPointerComparison(op->get_lhs_operand());
		markPointerComparison(op->get_rhs_operand());
	}
	// TODO: check for other comparisons

	// propagate inter-procedurally, this assumes no cycles

	Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(funcDec, V_SgFunctionCallExp);
	for (Rose_STL_Container<SgNode *>::iterator i = calls.begin(); i != calls.end(); i++) {
		SgFunctionCallExp *call = isSgFunctionCallExp(*i);
		//cout << call->unparseToString() << endl;
		SgFunctionDeclaration *nextFuncDec = call->getAssociatedFunctionDeclaration();
		//SgFunctionDeclaration *nextFuncDec = call->getAssociatedFunctionSymbol()->get_declaration();
		nextFuncDec = isSgFunctionDeclaration(nextFuncDec->get_definingDeclaration());
		if (nextFuncDec == NULL) continue;
		string nextFuncName = nextFuncDec->get_name().getString();
		if (nextFuncName.find("operator") != string::npos) continue;	// skip operators
		if (funcDec->get_name() == nextFuncDec->get_name()) continue; // skip recursive calls
		SgFunctionDefinition *nextFuncDef = nextFuncDec->get_definition();
		//ROSE_ASSERT(nextFuncDef != NULL);
		if (nextFuncDef == NULL) continue;
		//nextFuncDec = nextFuncDef->get_declaration();
		if (soaFunctionMap.count(nextFuncDec) > 0) continue;
		SoAFunctionInfo *soaFunctionInfo = new SoAFunctionInfo;
		soaFunctionMap[nextFuncDec] = soaFunctionInfo;

		const SgInitializedNamePtrList &params = nextFuncDec->get_args();
		const SgExpressionPtrList &args = call->get_args()->get_expressions();
		vector<string> nextPointVars, nextNodeVars, nextOrgPointVars;
		for (int j = 0; j < params.size(); j++) {
			if (exprContainsVarName(args[j], pointVars)) {
				nextPointVars.push_back(params[j]->get_name());
				//nextOrgPointVars.push_back(args[j]);
			}
			if (exprContainsVarName(args[j], nodeVars)) {
				nextNodeVars.push_back(params[j]->get_name());
			}
		}
//		soaFunctionInfo->orgPointVars = nextOrgPointVars;
		soaFunctionInfo->pointVars = nextPointVars;
		fieldAccessAnalysis(nextFuncDec, nextPointVars, nextNodeVars, depth + 1);
	}
}

void SOATransform::summarizeForBlockClass() {
	cout << "\n-----SOATransform::summarizeForBlockClass-----" << endl;
	for (PointAccessMap::iterator i = pointAccessMap.begin(); i != pointAccessMap.end(); i++) {
		SgDeclarationStatement *typeDec = i->first->get_declaration();
		SgClassDeclaration *classDec = isSgClassDeclaration(typeDec);
		classDec = isSgClassDeclaration(classDec->get_definingDeclaration());
		SgClassDefinition *classDef = classDec->get_definition();
		//SgClassDefinition *classDef = getClassDefintion(classDec->get_name());
		ROSE_ASSERT(classDef != NULL);
		PointAccess *pa = i->second;
		cout << "class: " << classDec->get_name() << endl;
		bool fullSoA = true;
		for (map<string,int>::iterator j = pa->fields.begin(); j != pa->fields.end(); j++) {
			bool soa = true;
			string name = j->first;
			int flag = j->second;
			cout << " field: " << name << "\t" << getAccessDescription(flag) << endl;

			if (flag & writeViaPoint && (flag & readViaNode || flag & writeViaNode)) {
				cout << "  read/write conflict on field " << name << endl;
				soa = false;
				continue;
			}
			if (flag & writeViaNode && (flag & readViaPoint || flag & writeViaPoint)) {
				cout << "  read/write conflict on field " << name << endl;
				soa = false;
				continue;
			}

			Rose_STL_Container<SgNode *> fieldDecs = NodeQuery::querySubTree(classDef, V_SgVariableDeclaration);
			for (Rose_STL_Container<SgNode *>::iterator k = fieldDecs.begin(); k != fieldDecs.end(); k++) {
				SgVariableDeclaration *fieldDec = isSgVariableDeclaration(*k);
				SgInitializedName *initName = getVarDecInitName(fieldDec);
				string fieldName = initName->get_name();
				if (fieldName != j->first) continue;
				SgType *type = initName->get_type();
				int dim = 0;
				if (isSgArrayType(type)) {
					SgArrayType *atype = isSgArrayType(type);
					type = atype->get_base_type();
					dim = isSgUnsignedLongVal(atype->get_index())->get_value();
				}

				SgClassType *ctype = isSgClassType(type);
				if (isSgPointerType(type)) {
					cout << "  is pointer, excluding from SoA" << endl;
					soa = false;
				} else if (ctype != NULL) {
					// support one level of structure flattening, TODO: generalize to multiple levels

					SgClassDefinition *subClassDef = getClassDefintion(ctype->get_name());
					Rose_STL_Container<SgNode *> subFieldDecs = NodeQuery::querySubTree(subClassDef, V_SgVariableDeclaration);
					for (Rose_STL_Container<SgNode *>::iterator l = subFieldDecs.begin(); l != subFieldDecs.end(); l++) {
						SgVariableDeclaration *subFieldDec = isSgVariableDeclaration(*l);
						SgInitializedName *subInitName = getVarDecInitName(subFieldDec);
						string subFieldName = subInitName->get_name();
						cout << "  flattening sub field: " << subFieldName << endl;
						SgType *subType = subInitName->get_type();
						if (!isSOAType(subType)) {
							cout << "   not primitive int or float, excluding from SoA" << endl;
							soa = false;
							continue;
						}
						string subTypeName = getTypeString(subType);
						PointField *pf = new PointField(subTypeName, fieldName, subFieldName);
						addPointField(pa, flag, pf);
					}
				} else if (isSOAType(type)) {
					string typeName = getTypeString(type);
					PointField *pf = new PointField(typeName, fieldName, dim);
					addPointField(pa, flag, pf);
				} else {
					cout << "  not primitive int or float, excluding from SoA" << endl;
					soa = false;
				}
				break;
			}
			pa->soa[name] = soa;
			fullSoA = fullSoA && soa;
		}
		// preserve isPointerComparison = true, if set in fieldAccessAnalysis (as in BH)
		pa->isPointerComparison = !fullSoA || pa->isPointerComparison;
	}
	cout << "-----------------------------\n" << endl;
}


void SOATransform::analyzeDec(SgFunctionDeclaration *funcDec, SoAFunctionInfo *soaFunctionInfo) {
	const vector<string> &pointVars = soaFunctionInfo->pointVars;
	const vector<string> &orgPointVars = soaFunctionInfo->pointVars;

	bool needTransform = false;
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(funcDec, V_SgBinaryOp);
	for (Rose_STL_Container<SgNode *>::iterator j = nodes.begin(); j != nodes.end(); j++) {
		SgArrowExp *arrow = isSgArrowExp(*j);
		SgDotExp *dot = isSgDotExp(*j);
		SgBinaryOp *binOp = isSgBinaryOp(*j);
		if (arrow == NULL && dot == NULL) continue;

		SgExpression *lhs = binOp->get_lhs_operand();
		SgExpression *rhs = binOp->get_rhs_operand();
		SgVarRefExp *lhsVar = isSgVarRefExp(lhs);

		if (lhsVar == NULL) continue;
		string lhsName = lhsVar->get_symbol()->get_name();
		//cout << "lhs: " << lhsName << "\t" << lhsVar->get_type()->class_name() << endl;
		SgReferenceType *refType = isSgReferenceType(lhsVar->get_type());
		if (dot != NULL && refType == NULL) continue;

		bool found = false;
		int pointVarIndex;
		for (pointVarIndex = 0; pointVarIndex < pointVars.size(); pointVarIndex++) {
			if (pointVars[pointVarIndex] == lhsName) {
				found = true;
				break;
			}
		}
		if (!found) continue;

		// if arrow, lhsVar must be pointer since it is lhs of arrow
		SgType *baseType = arrow == NULL ? refType->get_base_type() : isSgPointerType(lhsVar->get_type())->get_base_type();
		SgModifierType *mtype = isSgModifierType(baseType);
		if (mtype != NULL) {
			// handle modifiers, e.g., 'const' in  const Point* a
			baseType = mtype->get_base_type();
		}
		SgNamedType *ntype = isSgNamedType(baseType);
		ROSE_ASSERT(ntype != NULL);
		PointAccessMap::iterator it = pointAccessMap.find(ntype);
		if (it == pointAccessMap.end()) continue;
		PointAccess *pa = it->second;

		SgVarRefExp *rhsVarRef = isSgVarRefExp(binOp->get_rhs_operand());
		ROSE_ASSERT(rhsVarRef != NULL);
		string rhsName = rhsVarRef->get_symbol()->get_name();
		map<string, bool>::iterator it2 = pa->soa.find(rhsName);
		if (it2 == pa->soa.end()) continue;
		bool soa = it2->second;
		if (!soa) continue;

		needTransform = true;
		break;
	}
	soaFunctionInfo->needTransform = needTransform;
}

void SOATransform::analyzeDecPaths(SgFunctionDeclaration *funcDec, SoAFunctionInfo *soaFunctionInfo) {
	SgBasicBlock *body = funcDec->get_definition()->get_body();
	bool needTransform = soaFunctionInfo->needTransform;
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(body, V_SgFunctionCallExp);
	for (Rose_STL_Container<SgNode *>::iterator i = nodes.begin(); i != nodes.end(); i++) {
		SgFunctionCallExp *call = isSgFunctionCallExp(*i);
		SgFunctionDeclaration *nextFuncDec = call->getAssociatedFunctionDeclaration();
		nextFuncDec = isSgFunctionDeclaration(nextFuncDec->get_definingDeclaration());
		if (nextFuncDec == NULL) continue;
		if (nextFuncDec == funcDec) continue; // skip recursive calls
		SoAFunctionInfoMap::iterator it = soaFunctionMap.find(nextFuncDec);
		if (it == soaFunctionMap.end()) continue;
		SoAFunctionInfo *nextSoaFunctionInfo = it->second;
		needTransform = needTransform || nextSoaFunctionInfo->needTransform;
		analyzeDecPaths(nextFuncDec, nextSoaFunctionInfo);
	}
	soaFunctionInfo->needTransform = needTransform;
}

void transformCall(SgFunctionCallExp *call, SgInitializedName *blockVar, SgInitializedName *indexVar, const vector<string> &pointVars) {
	//cout << call->unparseToString() << endl;

	SgExprListExp *newArgList = buildExprListExp();
	newArgList->append_expression(buildVarRefExp(blockVar));
	newArgList->append_expression(buildVarRefExp(indexVar));
	const SgExpressionPtrList& args = call->get_args()->get_expressions();
	for (int i = 0; i < args.size(); i++) {
		const SgExpression *arg = args[i];
		const SgCastExp *cexp = isSgCastExp(arg);
		if (cexp != NULL) {
			arg = cexp->get_operand();
		}
		//cout << funcName << " " << arg->unparseToString() << " " << arg->class_name() << endl;
		const SgVarRefExp *varArg = isSgVarRefExp(arg);
		bool found = false;
		if (varArg != NULL) {
			string varName = varArg->get_symbol()->get_name();
			for (int j = 0; j < pointVars.size(); j++) {
				if (pointVars[j] == varName) {
					found = true;
					break;
				}
			}
		}
		if (!found) {
			newArgList->append_expression(copyExpression(args[i]));
		}
	}
	call->set_args(newArgList);
	newArgList->set_parent(call);

	//cout << call->unparseToString() << endl;
}

void SOATransform::transformBody(SgBasicBlock *body, SgInitializedName *blockVar, SgInitializedName *indexVar,
		const vector<string> &pointVars, const vector<string> &orgPointVars, int depth) {

	//	cout << "-----" << endl;
	//	cout << body->unparseToString() << endl << endl;
	//	cout << "pointVars: ";
	//	ConstIteratorFor(i, pointVars, vector<string>) {
	//		cout << *i << " ";
	//	}
	//	cout << endl;
	//	cout << "orgPointVars: ";
	//	ConstIteratorFor(i, orgPointVars, vector<string>) {
	//		cout << *i << " ";
	//	}
	//	cout << endl << endl;

	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(body, V_SgVarRefExp);
	vector<IntBinOpPair> ops;
	set<SgFunctionCallExp *> calls;
	for (Rose_STL_Container<SgNode *>::iterator i = nodes.begin(); i != nodes.end(); i++) {
		SgVarRefExp *varRef = isSgVarRefExp(*i);
		if (varRef == NULL) continue;
		string varName = varRef->get_symbol()->get_name();
		bool found = false;
		int j;
		for (j = 0; j < pointVars.size(); j++) {
			if (pointVars[j] == varName) {
				found = true;
				break;
			}
		}
		if (!found) continue;
		SgArrowExp *arrow = isSgArrowExp(varRef->get_parent());
		SgDotExp *dot = isSgDotExp(varRef->get_parent());
		SgFunctionCallExp *call = getEnclosingFunctionCall(varRef);
		//cout << varRef->unparseToString() << "\t" << (call == NULL ? "" : call->unparseToString()) << endl;
		if (arrow != NULL) {
			ops.push_back(IntBinOpPair(j, arrow));
		} else if (dot != NULL) {
			ops.push_back(IntBinOpPair(j, dot));
		}

		if (call != NULL) {
			//cout << "call: " << call->unparseToString() << endl << endl;
			if (calls.count(call) == 0) {
				calls.insert(call);
			}
		}
	}

	// handle calls first

	for (set<SgFunctionCallExp *>::iterator it = calls.begin(); it != calls.end(); it++) {
		SgFunctionCallExp *call = *it;
		const SgExpressionPtrList& args = call->get_args()->get_expressions();
		//cout << call->unparseToString() << endl;
		SgFunctionSymbol *sym = call->getAssociatedFunctionSymbol();
		if (sym == NULL) continue;
		// skip _SpliceNode::add()
		if (sym == g_spliceNodeAddSym) continue;

		string funcName = sym->get_name();
		//if (funcName == "_checkRayIntersection") continue;

		if (sym == g_blockAddSym) {
			transformCall(call, blockVar, indexVar, pointVars);
			continue;
		}

		// then transform func dec, only if not transformed yet
		SgFunctionDeclaration *funcDec = call->getAssociatedFunctionDeclaration();
		funcDec = isSgFunctionDeclaration(funcDec->get_definingDeclaration());
		if (funcDec == NULL) continue;
		SoAFunctionInfoMap::iterator it = soaFunctionMap.find(funcDec);
		if (it == soaFunctionMap.end()) {
			//cout << "!!" << call->unparseToString() << endl;
			continue;
		}
		SoAFunctionInfo *soaFunctionInfo = it->second;
		//cout << "!!!!" << call->unparseToString() << "\t" << soaFunctionInfo->needTransform << "\t" << soaFunctionInfo->transformed << endl;
		if (!soaFunctionInfo->needTransform) continue;


		transformCall(call, blockVar, indexVar, pointVars);


		if (soaFunctionInfo->transformed) continue;
		soaFunctionInfo->transformed = true;




		//SgFunctionDeclaration *nondefiningFuncDec = isSgFunctionDeclaration(funcDec->get_firstNondefiningDeclaration());
		SgFunctionDefinition *funcDef = funcDec->get_definition();
		//cout << funcDec->unparseToString() << endl << endl;
		//cout << call->unparseToString() << endl << endl;
		const SgInitializedNamePtrList &params = funcDec->get_args();

		SgInitializedName *newBlockDec = buildInitializedName("_block", buildPointerType(g_blockType));
		SgInitializedName *newIndexDec = buildInitializedName("_bi", buildIntType ());

		vector<string> newPointVars, newOrgPointVars;
		SgFunctionParameterList *newParamList = buildFunctionParameterList();
		appendArg(newParamList, newBlockDec);
		appendArg(newParamList, newIndexDec);



		//ROSE_ASSERT(params.size() == args.size());
		for (int i = 0; i < params.size(); i++) {
			const SgInitializedName *param = params[i];
			SgExpression *arg = args[i];
			SgCastExp *cexp = isSgCastExp(arg);
			if (cexp != NULL) {
				arg = cexp->get_operand();
			}
			SgVarRefExp *varArg = isSgVarRefExp(arg);
			bool found = false;
			if (varArg != NULL) {
				string varName = varArg->get_symbol()->get_name();
				//cout << i << " " << varName << " " << param->get_name() << endl;
				for (int j = 0; j < pointVars.size(); j++) {
					if (pointVars[j] == varName) {
						newPointVars.push_back(param->get_name());
						newOrgPointVars.push_back(orgPointVars[j]);
						found = true;
						break;
					}
				}
			}
			if (!found) {
				appendArg(newParamList, new SgInitializedName(*param));
			}
		}

		SgFunctionType *funcType = funcDec->get_type();
		SgMemberFunctionDeclaration *memFunc = isSgMemberFunctionDeclaration(funcDec);
		SgFunctionDeclaration *newFuncDec;
		if (memFunc != NULL) {
			SgMemberFunctionType *memFuncType = buildMemberFunctionType(funcType->get_return_type(), NULL, NULL, 0);
			newFuncDec = buildDefiningMemberFunctionDeclaration(funcName, memFuncType, newParamList, memFunc->get_class_scope());
		} else {
			newFuncDec = buildDefiningFunctionDeclaration(funcName, funcType->get_return_type(), newParamList, g_global);
		}


		insertStatementBefore(funcDec, newFuncDec);

		// add non-defining forward declaration if necessary
		addForwardFuncDecIfNecessary(funcDec, newFuncDec);

		SgBasicBlock* newFuncBody = newFuncDec->get_definition()->get_body();
		const SgStatementPtrList &stmts = funcDec->get_definition()->get_body()->get_statements();
		bool divideStmtFound = false;
		for (SgStatementPtrList::const_iterator i = stmts.begin(); i != stmts.end(); i++) {
			appendStatement(copyStatement(*i), newFuncBody);
		}
		//cout << newFuncDec->unparseToString() << endl;
		for (int i = 0; i < depth; i++) cout << " ";
		cout << "SOATransform::transformBody: " << newFuncDec->get_name() << endl;
		transformBody(newFuncBody, newBlockDec, newIndexDec, newPointVars, newOrgPointVars, depth + 1);
	}

	// then modify arrows and dots

	for (vector<IntBinOpPair>::iterator it = ops.begin(); it != ops.end(); it++) {
		SgBinaryOp *op = it->op;
		int pointVarIndex = it->pointVarIndex;
		/* change
		 * point->num_nodes_traversed++;
		 * to
		 * _block->num_nodes_traversed[_bi]++;
		 */
		SgVarRefExp *lhsVar = isSgVarRefExp(op->get_lhs_operand());
		SgPointerType *ptype = isSgPointerType(lhsVar->get_type());
		SgReferenceType *rtype = isSgReferenceType(lhsVar->get_type());
		SgType *baseType = ptype == NULL ? rtype->get_base_type() : ptype->get_base_type();
		SgModifierType *mtype = isSgModifierType(baseType);
		if (mtype != NULL) {
			// handle modifiers, e.g., 'const' in  const Point* a
			baseType = mtype->get_base_type();
		}
		SgNamedType *ntype = isSgNamedType(baseType);
		ROSE_ASSERT(ntype != NULL);
		PointAccessMap::iterator it = pointAccessMap.find(ntype);
		if (it == pointAccessMap.end()) continue;
		PointAccess *pa = it->second;

		SgVarRefExp *rhsVarRef = isSgVarRefExp(op->get_rhs_operand());
		ROSE_ASSERT(rhsVarRef != NULL);
		string rhsName = rhsVarRef->get_symbol()->get_name();
		map<string, bool>::iterator it2 = pa->soa.find(rhsName);
		if (it2 == pa->soa.end()) continue;
		bool soa = it2->second;
		if (!soa) continue;


		string fieldName = orgPointVars[pointVarIndex] + "_" + rhsName;

		SgNode *parent = op->get_parent();
		SgPntrArrRefExp *outerArray = isSgPntrArrRefExp(parent);
		SgDotExp *dot = isSgDotExp(parent);
		//cout << parent->unparseToString() << "\t" << parent->class_name() << endl;
		if (outerArray != NULL) {
			//			cout << outerArray->unparseToString() << endl;
			//			cout << outerArray->get_rhs_operand()->unparseToString() << "\t" << outerArray->get_rhs_operand()->class_name() << endl;
			SgExpression *blockArrow = buildArrowExp(buildVarRefExp(blockVar),
					buildVarRefExp(fieldName, g_blockClassDec->get_definition()));
			SgExpression *array = buildPntrArrRefExp(blockArrow, copyExpression(outerArray->get_rhs_operand()));
			SgExpression *newOuterArray = buildPntrArrRefExp(array, buildVarRefExp(indexVar));
			replaceExpression(outerArray, newOuterArray);
		} else if (dot != NULL) {
			SgVarRefExp *dotRhsVar = isSgVarRefExp(dot->get_rhs_operand());
			ROSE_ASSERT(dotRhsVar != NULL);
			fieldName += "_" + dotRhsVar->get_symbol()->get_name();
			SgExpression *blockArrow = buildArrowExp(buildVarRefExp(blockVar),
					buildVarRefExp(fieldName, g_blockClassDec->get_definition()));
			SgExpression *array = buildPntrArrRefExp(blockArrow, buildVarRefExp(indexVar));
			replaceExpression(dot, array);
		} else {
			SgExpression *blockArrow = buildArrowExp(buildVarRefExp(blockVar),
					buildVarRefExp(fieldName, g_blockClassDec->get_definition()));
			SgExpression *array = buildPntrArrRefExp(blockArrow, buildVarRefExp(indexVar));
			replaceExpression(op, array);
		}
	}
}

void SOATransform::analyze() {
	fieldAccessAnalysisEntry();
	summarizeForBlockClass();
}

void SOATransform::analyzeForTransform() {
	for (SoAFunctionInfoMap::iterator i = soaFunctionMap.begin(); i != soaFunctionMap.end(); i++) {
		SgFunctionDeclaration *funcDec = i->first;
		SoAFunctionInfo *soaFunctionInfo = i->second;
		analyzeDec(funcDec, soaFunctionInfo);
	}

	SoAFunctionInfo *soaFunctionInfo = soaFunctionMap[g_recursiveFunction];
	analyzeDecPaths(g_recursiveFunction, soaFunctionInfo);
}

void SOATransform::transform(SgBasicBlock *body, SgInitializedName *blockVar, SgInitializedName *indexVar) {
	vector<string> pointVars, orgPointVars;
	ConstIteratorFor(i, g_methodInfo.recursiveMethodLoopVariantParams, ParamVec) {
		SgInitializedName *param = *i;
		pointVars.push_back(param->get_name());
		orgPointVars.push_back(param->get_name());
	}
	transformBody(body, blockVar, indexVar, pointVars, orgPointVars, 0);
}




