/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#ifndef SOATRANSFORM_H_
#define SOATRANSFORM_H_

#include "rose.h"
#include "Globals.h"

#include <map>
#include <set>
#include <vector>
using namespace std;

enum {
	readViaPoint = 0x1,
	readViaNode = 0x2,
	writeViaPoint = 0x4,
	writeViaNode = 0x8,
};

struct PointField {
	string type;
	string name;

	/* currently support only one level of structure flattening
	ex) Vec acc becomes
	float* p_acc_x;
	float* p_acc_y;
	float* p_acc_z;
	 */
	string subName;
	int dim;

	PointField(string t, string n, int d) : type(t), name(n), subName(""), dim(d) {}
	PointField(string t, string n, string s) : type(t), name(n), subName(s), dim(0) {}
};

typedef vector<PointField *> PointFieldVector;

struct PointAccess {
	// fieldAccessAnalysis() populates this map
	// string - field name
	// int - access pattern, ex) readViaPoint | readViaNode
	map<string, int> fields;

	// maps field name to whether it has been transformed
	map<string, bool> soa;

	PointFieldVector readFields, writeFields, allFields;
	bool isPointerComparison; // direct pointer comparison requires saving the point at all levels
	// e.g., if ((p!=node->point))
	// Point **p_;

	// Splice+SOA also requires the pointer to save into _SpliceNode
	PointAccess() : isPointerComparison(g_useSplice) {}
};

typedef map<SgNamedType *, PointAccess *> PointAccessMap;

struct SoAFunctionInfo {
	// name of point vars within this function
	vector<string> pointVars;

//	// name of point vars within recursive function, need to appy this name to SoA transformed vars
//	vector<string> orgPointVars;

	bool needTransform;
	bool transformed;

	SoAFunctionInfo() : needTransform(false), transformed(false) {}
};

typedef map<SgFunctionDeclaration *, SoAFunctionInfo *> SoAFunctionInfoMap;

class SOATransform {
public:
	// called for both phases
	// populate pointAccessMap for use in generating block classes, and use in applying SoA transform in second phase
	static void analyze();

	// called for second phase only
	// populate soaFunctionMap
	static void analyzeForTransform();

	// must be called after analyze() and analyzeForTransform()
	// uses pointAccessMap and soaFunctionMap to actually transform code for SoA,
	// including generating SoA versions of functions called from recursive function
	static void transform(SgBasicBlock *body, SgInitializedName *blockVar, SgInitializedName *indexVar);

	static PointAccessMap pointAccessMap;

private:
	static void fieldAccessAnalysisEntry();
	static void summarizeForBlockClass();

	static void fieldAccessAnalysis(SgFunctionDeclaration *funcDec, const vector<string>& pointVars, const vector<string>& nodeVars, int depth);

	// direct pointer comparison requires saving the point at all levels
	// e.g., if ((p!=node->point))
	// Point **p_;
	static void markPointerComparison(SgExpression *exp);

	static void transformBody(SgBasicBlock *body, SgInitializedName *blockVar, SgInitializedName *indexVar,
			const vector<string> &pointVars, const vector<string> &orgPointVars, int depth);

	// called for each function in soaFunctionMap
	// analyze and set needTransform of SoAFunctionInfo if has SoA access
	static void analyzeDec(SgFunctionDeclaration *funcDec, SoAFunctionInfo *soaFunctionInfo);

	// in call path A->B->C
	// if C is needTransform, then set B and A also
	static void analyzeDecPaths(SgFunctionDeclaration *funcDec, SoAFunctionInfo *soaFunctionInfo);

	static SoAFunctionInfoMap soaFunctionMap;
};

#endif /* SOATRANSFORM_H_ */
