/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "rose.h"
#include "SpliceTransform.h"
#include "Globals.h"
#include "MethodInfo.h"
#include "Utils.h"
#include "RecursiveCallSet.h"
#include "RecursiveField.h"
#include "CommonTransform.h"
#include "SOATransform.h"

using namespace SageBuilder;
using namespace SageInterface;


SgNamedType *g_spliceSetType;
SgNamedType *g_spliceNodeType;

SgClassDeclaration *g_spliceSetClassDec;
SgClassDeclaration *g_spliceNodeClassDec;

SgMemberFunctionSymbol *g_spliceSetInitSym;
SgMemberFunctionSymbol *g_spliceSetDoSpliceSym;
SgMemberFunctionSymbol *g_spliceSetMapTreeSym;
SgMemberFunctionSymbol *g_spliceSetGetNodeSym;

SgFunctionSymbol *g_recurseSpliceSym;
SgFunctionSymbol *g_nextMapTreeSym;

void SpliceTransform::findSpliceClasses() {
	g_spliceSetType = isSgNamedType(lookupNamedTypeInParentScopes("_SpliceSet", g_global));
	ROSE_ASSERT(g_spliceSetType != NULL);
	g_spliceNodeType = isSgNamedType(lookupNamedTypeInParentScopes("_SpliceNode", g_global));
	ROSE_ASSERT(g_spliceNodeType != NULL);

	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(g_global, V_SgClassDeclaration);
	for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
		SgClassDeclaration *classDec = isSgClassDeclaration(*i);
		if (classDec->get_name() == "_SpliceSet") g_spliceSetClassDec = classDec;
		else if (classDec->get_name() == "_SpliceNode") g_spliceNodeClassDec = classDec;
	}

	g_spliceSetInitSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
			SgName("init"), g_spliceSetClassDec->get_definition()));
	ROSE_ASSERT(g_spliceSetInitSym != NULL);
	g_spliceSetDoSpliceSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
			SgName("do_splice"), g_spliceSetClassDec->get_definition()));
	ROSE_ASSERT(g_spliceSetDoSpliceSym != NULL);
	g_spliceSetMapTreeSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
			SgName("map_tree"), g_spliceSetClassDec->get_definition()));
	ROSE_ASSERT(g_spliceSetMapTreeSym != NULL);
	g_spliceSetGetNodeSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
			SgName("get_node"), g_spliceSetClassDec->get_definition()));
	ROSE_ASSERT(g_spliceSetGetNodeSym != NULL);

	g_spliceNodeAddSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
			SgName("add"), g_spliceNodeClassDec->get_definition()));
	ROSE_ASSERT(g_spliceSetGetNodeSym != NULL);
}

void SpliceTransform::transformEntry() {
	SgFunctionDeclaration *funcDec = getEnclosingFunctionDeclaration(g_enclosingForStmt);
	SgBasicBlock* funcBody = funcDec->get_definition()->get_body();

	SgFunctionCallExp *call = g_intermediaryCallSites[g_numCallSites - 1];
	SgStatement *callStmt = getEnclosingStatement(call);
	SgBasicBlock *block = isSgBasicBlock(g_enclosingForStmt->get_loop_body());

	vector<SgStatement *> prologue, epilogue;
	vector<SgVariableDeclaration *> usedInEpilogueVarDecs;
	CommonTransform::proEpilogueAnalysis(block, callStmt, prologue, epilogue, usedInEpilogueVarDecs);

	SgForInitStatement *entryForInit = g_enclosingForStmt->get_for_init_stmt();
	const SgStatementPtrList &entryForInits = entryForInit->get_init_stmt();
	ROSE_ASSERT(entryForInits.size() == 1);
	SgStatement *entryForInitStmt = *entryForInits.begin();
	SgVariableDeclaration *entryForVarDec = isSgVariableDeclaration(entryForInitStmt);
	SgInitializedName *indexVar;
	SgAssignInitializer *initVal = getVarDecInit(entryForVarDec, &indexVar);

	SgStatement *testStmt =	g_enclosingForStmt->get_test();
	SgExprStatement *etestStmt = isSgExprStatement(testStmt);
	ROSE_ASSERT(etestStmt != NULL);
	SgExpression *cond = etestStmt->get_expression();
	SgLessThanOp *ltcond = isSgLessThanOp(cond);
	ROSE_ASSERT(ltcond != NULL);
	SgExpression *forUpper = ltcond->get_rhs_operand();

	// find_correlation_mapRoot(root);

	SgFunctionCallExp *mapRootCall = buildFunctionCallExp(g_intermediaryFunctions[g_numCallSites - 1]->get_name() + g_suffixMapRoot,
			buildVoidType(), getArgsExprList(g_methodInfo.outerForLoopInvariantArgs), g_global);
	insertStatementBefore(g_enclosingForStmt, buildExprStatement(mapRootCall));

	/*
  _SpliceSet::init(3);
  _IntermediaryBlock::max_block = npoints;
  _SpliceSet _spliceSet;
  class _BlockStack *_stack = _spliceSet.get_stack();
  class _Block *_block = _spliceSet.get_block();
  class _IntermediaryBlock _interBlock;
  _interBlock._IntermediaryBlock::block = _block;
	 */

	SgExprListExp *spliceSetInitCallArgs = buildExprListExp();
	spliceSetInitCallArgs->append_expression(buildIntVal(g_spliceDepth));
	SgFunctionCallExp *spliceSetInitCall = buildFunctionCallExp(
			buildMemberFunctionRefExp(g_spliceSetInitSym, false, true), spliceSetInitCallArgs);
	insertStatementBefore(g_enclosingForStmt, buildExprStatement(spliceSetInitCall));

	SgExpression *assign = buildAssignOp(
			buildVarRefExp("max_block", g_intermediaryBlockClassDec->get_definition()), copyExpression(forUpper));
	insertStatementBefore(g_enclosingForStmt, buildExprStatement(assign));

	SgVariableDeclaration *spliceSetDec = buildVariableDeclaration("_spliceSet", g_spliceSetType);
	insertStatementBefore(g_enclosingForStmt, spliceSetDec);

	SgVariableDeclaration *stackDec = buildVariableDeclaration("_stack", buildPointerType(g_blockStackType),
			buildAssignInitializer(buildDotExp(buildVarRefExp(spliceSetDec),
					buildVarRefExp("stack", g_spliceSetClassDec->get_definition()))));
	insertStatementBefore(g_enclosingForStmt, stackDec);

	SgVariableDeclaration *blockDec = buildVariableDeclaration("_block", buildPointerType(g_blockType),
			buildAssignInitializer(buildDotExp(buildVarRefExp(spliceSetDec),
					buildVarRefExp("block", g_spliceSetClassDec->get_definition()))));
	insertStatementBefore(g_enclosingForStmt, blockDec);

	SgVariableDeclaration *interBlockDec = buildVariableDeclaration("_interBlock", g_intermediaryBlockType);
	insertStatementBefore(g_enclosingForStmt, interBlockDec);

	SgExpression *interBlockAssign = buildAssignOp(buildDotExp(buildVarRefExp(interBlockDec),
			buildVarRefExp("block", g_intermediaryBlockClassDec->get_definition())),
			buildVarRefExp(blockDec));
	insertStatementBefore(g_enclosingForStmt, buildExprStatement(interBlockAssign));


	// ::find_correlation_mapTree(root, &_spliceSet);

	SgExprListExp *mapTreeCallArgs = getArgsExprList(g_methodInfo.outerForLoopInvariantArgs);
	if (g_numCallSites == 1) {
		mapTreeCallArgs->append_expression(buildIntVal(0));
		mapTreeCallArgs->append_expression(buildIntVal(1));
	}
	mapTreeCallArgs->append_expression(buildAddressOfOp(buildVarRefExp(spliceSetDec)));
	SgFunctionCallExp *mapTreeCall = buildFunctionCallExp(
			g_intermediaryFunctions[g_numCallSites - 1]->get_name() + g_suffixMapTree,
			buildVoidType(), mapTreeCallArgs, g_global);
	insertStatementBefore(g_enclosingForStmt, buildExprStatement(mapTreeCall));


	// _interBlock.reset();

	SgFunctionCallExp *interBlockResetCall = buildFunctionCallExp(buildDotExp(buildVarRefExp(interBlockDec),
			buildMemberFunctionRefExp(g_interBlockResetSym, false, false)), buildExprListExp());
	insertStatementBefore(g_enclosingForStmt, buildExprStatement(interBlockResetCall));

	// for (int _start = start; _start < npoints; _start += _Block::max_block) {

	SgVariableDeclaration *startVarDec = buildVariableDeclaration(
			"_start", buildIntType(), isSgInitializer(copyExpression(initVal)), funcBody);
	SgExpression *startLtRhs = buildLessThanOp(buildVarRefExp(startVarDec), copyExpression(ltcond->get_rhs_operand()));
	SgBasicBlock *blockBlock = buildBasicBlock();
	SgForStatement *blockFor = buildForStatement(startVarDec, buildExprStatement(startLtRhs),
			buildPlusAssignOp(buildVarRefExp(startVarDec),
					buildVarRefExp("max_block", g_blockClassDec->get_definition())), blockBlock);
	insertStatementBefore(g_enclosingForStmt, blockFor);

	// int _end = min(_start + _Block::max_block, npoints);

	SgExprListExp *args = buildExprListExp();
	args->append_expression(buildAddOp(buildVarRefExp(startVarDec),
			buildVarRefExp("max_block", g_blockClassDec->get_definition())));
	args->append_expression(copyExpression(ltcond->get_rhs_operand()));
	SgFunctionCallExp *minCall = buildFunctionCallExp("min", buildVoidType(), args, g_global);
	SgVariableDeclaration *endVarDec = buildVariableDeclaration(
			"_end", buildIntType(), buildAssignInitializer(minCall), blockBlock);
	appendStatement(endVarDec, blockBlock);

	// prologue
	// for (int i = _start; i < _end; i++) {

	SgVariableDeclaration *innerIndexVarDec = buildVariableDeclaration(
			indexVar->get_name(), buildIntType(), buildAssignInitializer(buildVarRefExp(startVarDec)), blockBlock);
	SgExpression *innerIndexLtEnd = buildLessThanOp(buildVarRefExp(innerIndexVarDec), buildVarRefExp(endVarDec));
	SgBasicBlock *prologueBlock = buildBasicBlock();
	SgForStatement *prologueFor = buildForStatement(innerIndexVarDec, buildExprStatement(innerIndexLtEnd),
			buildPlusPlusOp(buildVarRefExp(innerIndexVarDec)), prologueBlock);
	appendStatement(prologueFor, blockBlock);

	// copy prologue statements

	for (int i = 0; i < prologue.size(); i++) {
		appendStatement(copyStatement(prologue[i]), prologueBlock);
	}

	if (g_numCallSites == 1) {
		// _block->add(p);
		SgFunctionCallExp *blockAddCall = buildFunctionCallExp(buildArrowExp(buildVarRefExp(blockDec),
				buildMemberFunctionRefExp(g_blockAddSym, false, false)),
				getArgsExprList(g_methodInfo.outerForLoopVariantArgs));
		appendStatement(buildExprStatement(blockAddCall), prologueBlock);
	} else {
		// find_correlation_prologue(p,root,&_interBlock);

		const SgFunctionCallExp *orgEntryCall = g_intermediaryCallSites[g_numCallSites - 1];
		SgExprListExp *prologueEntryCallArgs = isSgExprListExp(copyExpression(orgEntryCall->get_args()));
		prologueEntryCallArgs->append_expression(buildAddressOfOp(buildVarRefExp(interBlockDec)));
		SgFunctionCallExp *prologueEntryCall = buildFunctionCallExp(
				g_intermediaryFunctions[g_numCallSites - 1]->get_name() + g_suffixPrologue, orgEntryCall->get_type(),
				prologueEntryCallArgs, prologueBlock);
		appendStatement(buildExprStatement(prologueEntryCall), prologueBlock);
	}

	// _IntermediaryState *_interState = _interBlock. next ();

	SgFunctionCallExp *interBlockNextCall = buildFunctionCallExp(buildDotExp(buildVarRefExp(interBlockDec),
			buildMemberFunctionRefExp(g_interBlockNextSym, false, false)), buildExprListExp());
	SgVariableDeclaration *interStateVarDec = buildVariableDeclaration(
			"_interState", buildPointerType(g_intermediaryStateType),
			buildAssignInitializer(interBlockNextCall), prologueBlock);
	appendStatement(interStateVarDec, prologueBlock);

	/*
	 * 	_interState->a_prev = a_prev;
		_interState->p = p;
	 */

	for (vector<SgVariableDeclaration *>::const_iterator i = usedInEpilogueVarDecs.begin();
			i != usedInEpilogueVarDecs.end(); i++) {
		SgVariableDeclaration *varDec = *i;
		SgInitializedName* initName = *(varDec->get_variables().begin());
		SgExpression *lhs = buildArrowExp(buildVarRefExp(interStateVarDec),
				buildVarRefExp(initName->get_name(), g_intermediaryStateClassDec->get_definition()));

		SgType *type = varDec->get_definition()->get_type();
		SgExpression *rhs = buildVarRefExp(varDec);
		// dereference pointer if var is reference type
		// ex) _interState -> _IntermediaryState::p = &p;
		if (isSgReferenceType(type)) {
			rhs = buildAddressOfOp(rhs);
		}
		SgExpression *assign = buildAssignOp(lhs, rhs);
		appendStatement(buildExprStatement(assign), prologueBlock);
	}


	// _stack->get(0)->block = _block;

	SgExprListExp *stackGetArgs = buildExprListExp();
	stackGetArgs->append_expression(buildIntVal(0));
	SgExpression *stackGet = buildArrowExp(buildVarRefExp(stackDec),
			buildMemberFunctionRefExp(g_blockStackGetSym, false, false));
	SgFunctionCallExp *getCall = buildFunctionCallExp(stackGet, stackGetArgs);
	SgExpression *stackAssignBlock = buildAssignOp(
			buildArrowExp(getCall, buildVarRefExp("block", g_blockStackClassDec->get_definition())),
			buildVarRefExp(blockDec));
	appendStatement(buildExprStatement(stackAssignBlock), blockBlock);

	// correlation_search_block(_Block::root_node, _stack, 0, &_spliceSet);

	SgExprListExp *blockCallArgs = buildExprListExp();
	for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
		blockCallArgs->append_expression(buildVarRefExp(
				"root_" + param->get_name(), g_blockClassDec->get_definition()));
	}
	blockCallArgs->append_expression(buildVarRefExp(stackDec));
	blockCallArgs->append_expression(buildIntVal(0));
	if (g_spliceOptimizationLevel != Splice_NoOrder) blockCallArgs->append_expression(buildIntVal(1));
	blockCallArgs->append_expression(buildAddressOfOp(buildVarRefExp(spliceSetDec)));

	// if recursive function is member function, add root_instance prefix
	// ex) _Block::root_instance->intersectR_block(_Block::root_node,&_stack,0);

	if (g_recursiveMemberFunction != NULL) {
		SgClassDefinition *classScope = g_recursiveMemberFunction->get_class_scope();
		SgMemberFunctionSymbol *memFuncSym = isSgMemberFunctionSymbol(g_recurseSpliceSym);
		SgArrowExp *arrow = buildArrowExp(buildVarRefExp("root_instance", g_blockClassDec->get_definition()),
				buildMemberFunctionRefExp(memFuncSym, false, false));
		SgFunctionCallExp *blockCall = buildFunctionCallExp(arrow, blockCallArgs);
		appendStatement(buildExprStatement(blockCall), blockBlock);

	} else {
		SgFunctionCallExp *blockCall = buildFunctionCallExp(g_recurseSpliceSym, blockCallArgs);
		appendStatement(buildExprStatement(blockCall), blockBlock);
	}

	if (g_useSOA) {
		// _block->copy_back();
		SgFunctionCallExp *blockCopyBackCall = buildFunctionCallExp(buildArrowExp(buildVarRefExp(blockDec),
				buildMemberFunctionRefExp(g_blockCopyBackSym, false, false)), buildExprListExp());
		appendStatement(buildExprStatement(blockCopyBackCall), blockBlock);
	}

	// _block->recycle();

	SgFunctionCallExp *blockRecycleCall = buildFunctionCallExp(buildArrowExp(buildVarRefExp(blockDec),
			buildMemberFunctionRefExp(g_blockRecycleSym, false, false)), buildExprListExp());
	appendStatement(buildExprStatement(blockRecycleCall), blockBlock);

	// _spliceSet.do_splice()

	SgFunctionCallExp *doSpliceCall = buildFunctionCallExp(buildDotExp(buildVarRefExp(spliceSetDec),
			buildMemberFunctionRefExp(g_spliceSetDoSpliceSym, false, false)), buildExprListExp());
	insertStatementBefore(g_enclosingForStmt, buildExprStatement(doSpliceCall));

	// _interBlock.reset();

	insertStatementBefore(g_enclosingForStmt, buildExprStatement(copyExpression(interBlockResetCall)));

	// epilogues
	// for (int i = _start; i < _end; i++) {

	SgVariableDeclaration *epiIndexVarDec = buildVariableDeclaration(
			indexVar->get_name(), buildIntType(), isSgInitializer(copyExpression(initVal)), funcBody);
	SgExpression *epiIndexLtEnd = buildLessThanOp(buildVarRefExp(epiIndexVarDec), copyExpression(forUpper));
	SgBasicBlock *epilogueBlock = buildBasicBlock();
	SgForStatement *epilogueFor = buildForStatement(epiIndexVarDec, buildExprStatement(epiIndexLtEnd),
			buildPlusPlusOp(buildVarRefExp(epiIndexVarDec)), epilogueBlock);
	insertStatementBefore(g_enclosingForStmt, epilogueFor);

	// _IntermediaryState *_interState = _interBlock. next ();

	appendStatement(copyStatement(interStateVarDec), epilogueBlock);

	/*
	 *     	Vec a_prev = _interState->a_prev;
	    	Point *p = _interState->p;
	 */

	for (vector<SgVariableDeclaration *>::const_iterator i = usedInEpilogueVarDecs.begin();
			i != usedInEpilogueVarDecs.end(); i++) {
		SgVariableDeclaration *varDec = *i;
		SgInitializedName* initName = *(varDec->get_variables().begin());
		SgType *type = initName->get_type();
		SgExpression *rhs = buildArrowExp(buildVarRefExp(interStateVarDec),
				buildVarRefExp(initName->get_name(), g_intermediaryStateClassDec->get_definition()));
		// dereference pointer if var is reference type
		// ex) class Ray &ray =  *_point._Point::ray;
		if (isSgReferenceType(type)) {
			rhs = buildPointerDerefExp(rhs);
		}
		varDec = buildVariableDeclaration(
				initName->get_name(), type, buildAssignInitializer(rhs), epilogueBlock);
		appendStatement(varDec, epilogueBlock);
	}

	if (g_numCallSites != 1) {
		// find_correlation_epilogue(p,root,&_interBlock);

		const SgFunctionCallExp *orgEntryCall = g_intermediaryCallSites[g_numCallSites - 1];
		SgExprListExp *epilogueEntryCallArgs = isSgExprListExp(copyExpression(orgEntryCall->get_args()));
		epilogueEntryCallArgs->append_expression(buildAddressOfOp(buildVarRefExp(interBlockDec)));
		SgFunctionCallExp *epilogueEntryCall = buildFunctionCallExp(
				g_intermediaryFunctions[g_numCallSites - 1]->get_name() + g_suffixEpilogue, orgEntryCall->get_type(),
				epilogueEntryCallArgs, epilogueBlock);
		appendStatement(buildExprStatement(epilogueEntryCall), epilogueBlock);
	}

	// copy epilogue statements

	for (int i = 0; i < epilogue.size(); i++) {
		appendStatement(copyStatement(epilogue[i]), epilogueBlock);
	}

	// remove original for loop

	removeStatement(g_enclosingForStmt);
}

void SpliceTransform::transformMapTreeRecursive() {

	// create map root recursive function
	string funcName = g_recursiveFunction->get_name();
	SgFunctionParameterList* parameterList = buildFunctionParameterList();
	SgFunctionDeclaration *func;
	if (g_recursiveMemberFunction != NULL) {
		SgMemberFunctionType *memFuncType = buildMemberFunctionType(buildVoidType(), NULL, NULL, 0);
		func = buildDefiningMemberFunctionDeclaration(
				funcName + g_suffixMapTree, memFuncType, parameterList, g_recursiveMemberFunction->get_class_scope());
	} else{
		func = buildDefiningFunctionDeclaration(
					funcName + g_suffixMapTree, buildVoidType(), parameterList, g_global);
	}
	// save symbol to use in intermediary mapRoot functions
	g_nextMapTreeSym = isSgFunctionSymbol(func->search_for_symbol_from_symbol_table());
	ROSE_ASSERT(g_nextMapTreeSym != NULL);

	SgBasicBlock* funcBody = func->get_definition()->get_body();
	insertStatementBefore(g_recursiveFunction, func);

	// add params
	// _Block::root_node = node;

	for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
		SgInitializedName *newParam = buildInitializedName(param->get_name(), param->get_type());
		appendArg(parameterList, newParam);

		//		string varName = "root_" + param->get_name();
		//		SgExpression *assign = buildAssignOp(buildVarRefExp(varName, g_blockClassDec->get_definition()),
		//				buildVarRefExp(newParam));
		//		appendStatement(buildExprStatement(assign), funcBody);
	}
	SgInitializedName *depthVarDec = buildInitializedName("_depth", buildIntType ());
	SgInitializedName *indexVarDec = buildInitializedName("_index", buildIntType ());
	SgInitializedName *spliceSetVarDec = buildInitializedName("_spliceSet", buildPointerType(g_spliceSetType));
	appendArg(parameterList, depthVarDec);
	appendArg(parameterList, indexVarDec);
	appendArg(parameterList, spliceSetVarDec);

	// if (_depth > _SpliceSet::depth_param + 1) return;

	SgExpression *depthGtDepth = buildGreaterThanOp(
			buildVarRefExp(depthVarDec),
			buildAddOp(buildVarRefExp("depth_param", g_spliceSetClassDec->get_definition()),buildIntVal(1)));
	SgIfStmt *ifstmt = buildIfStmt(buildExprStatement(depthGtDepth), buildReturnStmt(), NULL);
	appendStatement(ifstmt, funcBody);

	// _set->map_tree(_depth, _index, node);

	SgExprListExp *mapTreeCallArgs = getArgsExprList(
			g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode, funcBody);
	mapTreeCallArgs->append_expression(buildVarRefExp(depthVarDec));
	mapTreeCallArgs->append_expression(buildVarRefExp(indexVarDec));
	SgFunctionCallExp *mapTreeCall = buildFunctionCallExp(buildArrowExp(buildVarRefExp(spliceSetVarDec),
			buildMemberFunctionRefExp(g_spliceSetMapTreeSym, false, false)), mapTreeCallArgs);
	appendStatement(buildExprStatement(mapTreeCall), funcBody);

	// copy over all statements within original recursive function

	SgBasicBlock *orgFuncBody = g_recursiveFunction->get_definition()->get_body();
	const SgStatementPtrList &orgStmts = orgFuncBody->get_statements();
	for (SgStatementPtrList::const_iterator i = orgStmts.begin(); i < orgStmts.end(); i++) {
		appendStatement(copyStatement(*i), funcBody);	// need full copy to avoid changing original function
	}

	// analyze loop variant statements

	set<string> loopVariantArgsPropagation;
	for (int i = 0; i < g_methodInfo.recursiveMethodLoopVariantParams.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodLoopVariantParams[i];
		loopVariantArgsPropagation.insert(param->get_name());
	}
	vector<SgFunctionCallExp *> recursionPoints;
	getRecursionPoints(funcBody, &recursionPoints, funcName);
	vector<SgStatement *> recursionPointStmts;
	for (int i = 0; i < recursionPoints.size(); i++) {
		recursionPointStmts.push_back(getEnclosingStatement(recursionPoints[i]));
	}

	vector<SgStatement *> toRemoveStmts;
	removeLoopVariantStmts(&toRemoveStmts, &loopVariantArgsPropagation, funcBody, recursionPointStmts);

	// remove loop variant statements

	for (vector<SgStatement *>::iterator i = toRemoveStmts.begin(); i != toRemoveStmts.end(); i++) {
		SgStatement *stmt = *i;
		removeStatement(stmt);
	}

	/*
	  	correlation_search_mapTree(node->left, _depth + 1, _index * 2, _set);
		correlation_search_mapTree(node->right, _depth + 1, _index * 2 + 1, _set);
	 */
	int numFields = RecursiveField::recursiveFields.size();
	for (vector<SgFunctionCallExp *>::iterator i = recursionPoints.begin(); i != recursionPoints.end(); i++) {
		SgFunctionCallExp *orgCall = isSgFunctionCallExp(*i);

		SgExprListExp *args = buildExprListExp();
		const SgExpressionPtrList &orgArgs = orgCall->get_args()->get_expressions();
		string nodeStr;
		for (int i = 0; i < orgArgs.size(); i++) {
			if (vecContains(g_methodInfo.recursiveMethodLoopVariantParamIdxs, i)) continue;
			if (i == g_nodeParamIdx) {
				SgArrowExp *arrow = isSgArrowExp(orgArgs[i]);
				ROSE_ASSERT(arrow != NULL);
				SgVarRefExp *rhs = isSgVarRefExp(arrow->get_rhs_operand());
				ROSE_ASSERT(rhs != NULL);
				nodeStr = rhs->get_symbol()->get_name();
			}
			args->append_expression(copyExpression(orgArgs[i]));
		}
		args->append_expression(buildAddOp(buildVarRefExp(depthVarDec), buildIntVal(1)));
		args->append_expression(buildAddOp(buildMultiplyOp(buildVarRefExp(indexVarDec), buildIntVal(numFields)),
				buildIntVal(RecursiveField::getFieldId(nodeStr))));
		args->append_expression(buildVarRefExp(spliceSetVarDec));
		SgFunctionCallExp *newCall = buildFunctionCallExp(g_recursiveFunction->get_name() + g_suffixMapTree,
				buildVoidType(), args, funcBody);
		replaceExpression(orgCall, newCall);
	}

	// add non-defining forward declaration if necessary
	addForwardFuncDecIfNecessary(g_recursiveFunction, func);
}

void SpliceTransform::addRecursivePrototype() {
	// create block recursive function
	string funcName = g_recursiveFunction->get_name();
	SgFunctionParameterList* parameterList = buildFunctionParameterList();
	SgFunctionDeclaration *func;
	if (g_recursiveMemberFunction != NULL) {
		SgMemberFunctionType *memFuncType = buildMemberFunctionType(buildVoidType(), NULL, NULL, 0);
		func = buildDefiningMemberFunctionDeclaration(funcName + g_suffixBlockSplice, memFuncType, parameterList, g_recursiveMemberFunction->get_class_scope());
	} else {
		func = buildDefiningFunctionDeclaration(funcName + g_suffixBlockSplice, buildVoidType(), parameterList, g_global);
	}

	// add params
	for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
		appendArg(parameterList, new SgInitializedName(*param));
	}
	SgInitializedName *depthVarDec = buildInitializedName("_depth", buildIntType ());
	// in first phase don't have g_blockStackType, must fall back to using opaque type
	SgInitializedName *stackVarDec = buildInitializedName("_stack", buildPointerType(buildOpaqueType("_BlockStack", g_global)));
	SgInitializedName *indexVarDec = buildInitializedName("_index", buildIntType());
	SgInitializedName *spliceSetVarDec = buildInitializedName("_spliceSet", buildPointerType(buildOpaqueType("_SpliceSet", g_global)));
	appendArg(parameterList, stackVarDec);
	appendArg(parameterList, depthVarDec);
	if (g_spliceOptimizationLevel != Splice_NoOrder) appendArg(parameterList, indexVarDec);
	appendArg(parameterList, spliceSetVarDec);

	addForwardFuncDecIfNecessary(g_recursiveFunction, func);

	prependStatement(buildNondefiningClassDeclaration("_SpliceSet", g_global), g_global);
}

void SpliceTransform::transformRecursive() {

	// create splice recursive function
	string funcName = g_recursiveFunction->get_name();
	SgFunctionParameterList* parameterList = buildFunctionParameterList();
	SgFunctionDeclaration *func;
	if (g_recursiveMemberFunction != NULL) {
		SgMemberFunctionType *memFuncType = buildMemberFunctionType(buildVoidType(), NULL, NULL, 0);
		func = buildDefiningMemberFunctionDeclaration(funcName + g_suffixBlockSplice, memFuncType, parameterList, g_recursiveMemberFunction->get_class_scope());
	} else {
		func = buildDefiningFunctionDeclaration(funcName + g_suffixBlockSplice, buildVoidType(), parameterList, g_global);
	}
	// save symbol to use in transformEntry()
	g_recurseSpliceSym = isSgFunctionSymbol(func->search_for_symbol_from_symbol_table());
	ROSE_ASSERT(g_recurseSpliceSym != NULL);

	SgBasicBlock* funcBody = func->get_definition()->get_body();
	insertStatementBefore(g_recursiveFunction, func);

	// add params

	for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode[i];
		appendArg(parameterList, new SgInitializedName(*param));
	}
	SgInitializedName *depthVarDec = buildInitializedName("_depth", buildIntType());
	SgInitializedName *stackVarDec = buildInitializedName("_stack", buildPointerType(g_blockStackType));
	SgInitializedName *indexVarDec = buildInitializedName("_index", buildIntType());
	SgInitializedName *spliceSetVarDec = buildInitializedName("_spliceSet", buildPointerType(g_spliceSetType));
	appendArg(parameterList, stackVarDec);
	appendArg(parameterList, depthVarDec);
	if (g_spliceOptimizationLevel != Splice_NoOrder) appendArg(parameterList, indexVarDec);
	appendArg(parameterList, spliceSetVarDec);

	// setup addArgs, reused via copyExpression()

	SgExprListExp *addArgs = buildExprListExp();
	for (int j = 0; j < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); j++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[j];
		addArgs->append_expression(buildVarRefExp(param->get_name(), funcBody));
	}

	/*
	 * 	_BlockSet *_set = _stack->get(_depth);
	_Block *_block = _set->block;
	_SpliceNode *_sn = _spliceSet->get_node(_depth);
	 */

	SgExprListExp *args = buildExprListExp();
	args->append_expression(buildVarRefExp(depthVarDec));
	SgExpression *stackGet = buildArrowExp(buildVarRefExp(stackVarDec),
			buildMemberFunctionRefExp(g_blockStackGetSym, false, false));
	SgFunctionCallExp *getCall = buildFunctionCallExp(stackGet, args);
	SgAssignInitializer *init = buildAssignInitializer(getCall);
	SgVariableDeclaration *blockSetVarDec = buildVariableDeclaration("_set", buildPointerType(g_blockSetType), init);
	appendStatement(blockSetVarDec, funcBody);

	init = buildAssignInitializer(buildArrowExp(buildVarRefExp(blockSetVarDec),
			buildVarRefExp("block", g_blockStackClassDec->get_definition())));
	SgVariableDeclaration *blockVarDec = buildVariableDeclaration("_block",
			buildPointerType(g_blockType), init);
	appendStatement(blockVarDec, funcBody);

	SgExprListExp *getNodeArgs = buildExprListExp();
	if (g_spliceOptimizationLevel != Splice_NoOrder) getNodeArgs->append_expression(buildVarRefExp(indexVarDec));
	else getNodeArgs->append_expression(buildVarRefExp(depthVarDec));
	SgExpression *getNode = buildArrowExp(buildVarRefExp(spliceSetVarDec),
			buildMemberFunctionRefExp(g_spliceSetGetNodeSym, false, false));
	SgFunctionCallExp *getNodeCall = buildFunctionCallExp(getNode, getNodeArgs);
	SgAssignInitializer *initGetNode = buildAssignInitializer(getNodeCall);
	SgVariableDeclaration *spliceNodeVarDec = buildVariableDeclaration("_sn", buildPointerType(g_spliceNodeType), initGetNode);
	appendStatement(spliceNodeVarDec, funcBody);


	// add block for loop

	SgVariableDeclaration *biVarDec = buildVariableDeclaration(
			"_bi", buildIntType(), buildAssignInitializer(buildIntVal(0)), funcBody);
	SgExpression *biLtBlockSize = buildLessThanOp(buildVarRefExp(biVarDec),
			buildArrowExp(buildVarRefExp(blockVarDec), buildVarRefExp("size", g_blockClassDec->get_definition())));
	SgBasicBlock *blockBlock = buildBasicBlock();
	SgForStatement *blockFor = buildForStatement(biVarDec, buildExprStatement(biLtBlockSize),
			buildPlusPlusOp(buildVarRefExp(biVarDec)), blockBlock);
	appendStatement(blockFor, funcBody);

	if (!g_useSOA) {
		/* add point decs to start
		 * 	_Point& _point = _block->get(_bi);
		Point *point = _point.point;
		 */

		args = buildExprListExp();
		args->append_expression(buildVarRefExp(biVarDec));
		SgMemberFunctionRefExp *funcRef = buildMemberFunctionRefExp(g_blockGetSym, false, false);
		getCall = buildFunctionCallExp(buildArrowExp(buildVarRefExp(blockVarDec), funcRef), args);
		init = buildAssignInitializer(getCall);
		SgVariableDeclaration *pointDec = buildVariableDeclaration("_point", buildReferenceType(g_pointType), init);
		appendStatement(pointDec, blockBlock);

		for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
			SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
			SgExpression *initExp = buildDotExp(buildVarRefExp(pointDec), buildVarRefExp(param->get_name(), g_pointClassDec->get_definition()));
			SgType *type = param->get_type();
			// dereference pointer if var is reference type
			// ex) class Ray &ray =  *_point._Point::ray;
			if (isSgReferenceType(type)) {
				initExp = buildPointerDerefExp(initExp);
			}
			SgAssignInitializer *init = buildAssignInitializer(initExp);
			SgVariableDeclaration *subPointDec = buildVariableDeclaration(param->get_name(), param->get_type(), init);
			appendStatement(subPointDec, blockBlock);
		}
	} else {
		// Point *p = _block->p_[_bi];

		for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopVariantParams.size(); i++) {
			SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopVariantParams[i];
			SgPointerType *ptype = isSgPointerType(param->get_type());
			SgReferenceType *rtype = isSgReferenceType(param->get_type());
			if (ptype == NULL && rtype == NULL) {

				// float upperDist = _block->upperDist[_bi];

				SgExpression *blockArrow = buildArrowExp(buildVarRefExp(blockVarDec),
						buildVarRefExp(param->get_name(), g_blockClassDec->get_definition()));
				SgExpression *array = buildPntrArrRefExp(blockArrow, buildVarRefExp(biVarDec));
				SgAssignInitializer *init = buildAssignInitializer(array);
				SgVariableDeclaration *subPointDec = buildVariableDeclaration(param->get_name(), param->get_type(), init);
				appendStatement(subPointDec, blockBlock);
			} else {

				// Point *p = _block->p_[_bi];

				SgNamedType *ntype = ptype == NULL ? isSgNamedType(rtype->get_base_type()) : isSgNamedType(ptype->get_base_type());
				if (ntype == NULL) continue;
				PointAccessMap::iterator it = SOATransform::pointAccessMap.find(ntype);
				ROSE_ASSERT(it != SOATransform::pointAccessMap.end());
				PointAccess *pa = it->second;
				if (!pa->isPointerComparison) continue;

				SgExpression *blockArrow = buildArrowExp(buildVarRefExp(blockVarDec),
						buildVarRefExp(param->get_name()  + "_", g_blockClassDec->get_definition()));
				SgExpression *initExp = buildPntrArrRefExp(blockArrow, buildVarRefExp(biVarDec));
				// dereference pointer if var is reference type
				// ex) class Ray &ray =  *_point._Point::ray;
				if (rtype != NULL) initExp = buildPointerDerefExp(initExp);

				SgAssignInitializer *init = buildAssignInitializer(initExp);
				SgVariableDeclaration *subPointDec = buildVariableDeclaration(param->get_name(), param->get_type(), init);
				appendStatement(subPointDec, blockBlock);
			}
		}
	}

	// bool _pointStopped = true;

	SgVariableDeclaration *pointStoppedVarDec = buildVariableDeclaration("_pointStopped", buildBoolType(),
			buildAssignInitializer(buildBoolValExp(true)));
	appendStatement(pointStoppedVarDec, blockBlock);


	// copy over all statements within original recursive function

	SgBasicBlock *orgFuncBody = g_recursiveFunction->get_definition()->get_body();
	const SgStatementPtrList &orgStmts = orgFuncBody->get_statements();
	for (SgStatementPtrList::const_iterator i = orgStmts.begin(); i < orgStmts.end(); i++) {
		appendStatement(copyStatement(*i), blockBlock);	// need full copy to avoid changing original function
	}

	// if (_point_stopped) _sn->add(point); at end of block loop

	SgFunctionCallExp *spliceAddCall = buildFunctionCallExp(buildArrowExp(
			buildVarRefExp(spliceNodeVarDec), buildMemberFunctionRefExp(g_spliceNodeAddSym, false, false)),
			isSgExprListExp(copyExpression(addArgs)));
	SgIfStmt *ifPointStoppedAddStmt = buildIfStmt(buildVarRefExp(pointStoppedVarDec),
			buildExprStatement(spliceAddCall), NULL);
	appendStatement(ifPointStoppedAddStmt, blockBlock);

	// change returns to continues
	// add if (_point_stopped) _sn->add(point); before continues

	Rose_STL_Container<SgNode *> returns = NodeQuery::querySubTree(blockBlock, V_SgReturnStmt);
	for (Rose_STL_Container<SgNode *>::iterator i = returns.begin(); i != returns.end(); i++) {
		SgContinueStmt *contStmt = buildContinueStmt();
		replaceStatement(isSgReturnStmt(*i), contStmt);
		insertStatementBefore(contStmt, copyStatement(ifPointStoppedAddStmt));
	}

	// analyze call sets - must be done AFTER copy over all statements within original recursive function

	RecursiveCallSet rcs(blockBlock, g_methodInfo.recursiveMethodLoopVariantParams, g_recursiveFunction->get_name());
	//cout << "SpliceTransform num call sets: " << rcs.callSets.size() << endl;


	/*
	 * 	_Block *_nextBlock0 = &_set->nextBlock0;
	_nextBlock0->recycle();
	 */
	SgVariableDeclaration **nextBlockVarDecs = new SgVariableDeclaration*[rcs.callSets.size()];
	for (int i = 0; i < rcs.callSets.size(); i++) {
		stringstream sstm;
		sstm << "nextBlock" << i;
		string nextBlockStr = sstm.str();
		string nextBlockStr2 = "_" + sstm.str();
		init = buildAssignInitializer(buildAddressOfOp(buildArrowExp(buildVarRefExp(blockSetVarDec),
				buildVarRefExp(nextBlockStr, g_blockSetClassDec->get_definition()))));
		SgVariableDeclaration *nextBlockVarDec = buildVariableDeclaration(nextBlockStr2,
				buildPointerType(g_blockType), init);
		nextBlockVarDecs[i] = nextBlockVarDec;
		insertStatementAfter(blockVarDec, nextBlockVarDec, funcBody);

		SgExprListExp *args = buildExprListExp();
		SgMemberFunctionSymbol *blockRecycleSym = isSgMemberFunctionSymbol(lookupFunctionSymbolInParentScopes(
				SgName("recycle"), g_blockClassDec->get_definition()));
		ROSE_ASSERT(blockRecycleSym != NULL);
		SgMemberFunctionRefExp *funcRef = buildMemberFunctionRefExp(blockRecycleSym, false, false);
		SgFunctionCallExp *recycleCall = buildFunctionCallExp(buildArrowExp(buildVarRefExp(nextBlockVarDec), funcRef), args);
		insertStatementAfter(nextBlockVarDec, buildExprStatement(recycleCall), funcBody);
	}

	/* replace call set stmt with
	if (_depth < _SpliceSet::depth_param) _nextBlock0->add(point);
	else _spliceSet->get_node(_depth + 1)->add(point);
	_point_stopped = false;
	 */

	int numFields = RecursiveField::recursiveFields.size();
	for (int i = 0; i < rcs.callSets.size(); i++) {
		CallSet *cset = rcs.callSets[i];
		stringstream sstm;
		sstm << "_nextBlock" << cset->id;
		SgFunctionCallExp *blockAddCall = buildFunctionCallExp(buildArrowExp(
				buildVarRefExp(sstm.str(), funcBody), buildMemberFunctionRefExp(g_blockAddSym, false, false)),
				isSgExprListExp(copyExpression(addArgs)));

		SgExprListExp *getNodeArgs = buildExprListExp();
		if (g_spliceOptimizationLevel != Splice_NoOrder) {
			// get field id of first call
			int fieldId = RecursiveField::getFieldId(getNodeVarName(cset->calls[0]));
			getNodeArgs->append_expression(buildAddOp(
					buildMultiplyOp(buildVarRefExp(indexVarDec), buildIntVal(numFields)), buildIntVal(fieldId)));
		} else {
			getNodeArgs->append_expression(buildAddOp(buildVarRefExp(depthVarDec), buildIntVal(1)));
		}
		SgExpression *getNode = buildArrowExp(buildVarRefExp(spliceSetVarDec),
				buildMemberFunctionRefExp(g_spliceSetGetNodeSym, false, false));
		SgFunctionCallExp *getNodeCall = buildFunctionCallExp(getNode, getNodeArgs);
		SgFunctionCallExp *spliceAddCall = buildFunctionCallExp(buildArrowExp(
				getNodeCall, buildMemberFunctionRefExp(g_spliceNodeAddSym, false, false)),
				isSgExprListExp(copyExpression(addArgs)));

		SgIfStmt *ifstmt = buildIfStmt(buildLessThanOp(buildVarRefExp(depthVarDec),
				buildVarRefExp("depth_param", g_spliceSetClassDec->get_definition())),
				buildExprStatement(blockAddCall),
				buildExprStatement(spliceAddCall));

		// TODO: handle multiple for HeatRay
		//ROSE_ASSERT(cset->stmts.size() == 1);
		replaceStatement(cset->stmts[0], ifstmt);

		insertStatementAfter(ifstmt, buildAssignStatement(buildVarRefExp(pointStoppedVarDec),
				buildBoolValExp(false)));
	}

	// add recursive calls on next block
	/**
	 * 	if (!_next_block0->is_empty()) {
		_stack->get(_depth + 1)->block = _nextBlock0;
		correlation_search_blockSplice(node->left, _stack, _depth + 1, _spliceSet);
	}
	 */
	for (int i = 0; i < rcs.callSets.size(); i++) {
		CallSet *cset = rcs.callSets[i];
		stringstream sstm;
		sstm << "_nextBlock" << cset->id;

		SgExpression *sizeGtZero = buildGreaterThanOp(buildArrowExp(
				buildVarRefExp(sstm.str(), funcBody), buildVarRefExp("size", g_blockClassDec->get_definition())),
				buildIntVal(0));
		SgBasicBlock *ifBlock = buildBasicBlock();
		SgIfStmt *ifstmt = buildIfStmt(buildExprStatement(sizeGtZero), ifBlock, NULL);
		appendStatement(ifstmt, funcBody);

		SgExprListExp *args = buildExprListExp();
		args->append_expression(buildAddOp(buildVarRefExp(depthVarDec), buildIntVal(1)));
		SgFunctionCallExp *getCall = buildFunctionCallExp(stackGet, args);


		SgExpression *stackGetBlock = buildArrowExp(getCall,
				buildVarRefExp("block", g_blockSetClassDec->get_definition()));
		SgExpression *assign = buildAssignOp(stackGetBlock, buildVarRefExp(sstm.str(), funcBody));
		appendStatement(buildExprStatement(assign), ifBlock);


		//Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(cset->stmt, V_SgFunctionCallExp);
		for (int j = 0; j < cset->calls.size(); j++) {
			SgFunctionCallExp *orgCall = isSgFunctionCallExp(cset->calls[j]);
			if (j == 0) {
				SgExprListExp *args = buildExprListExp();
				const SgExpressionPtrList &orgArgs = orgCall->get_args()->get_expressions();
				for (int j = 0; j < orgArgs.size(); j++) {
					if (vecContains(g_methodInfo.recursiveMethodLoopVariantParamIdxs, j)) continue;
					args->append_expression(copyExpression(orgArgs[j]));
				}
				args->append_expression(buildVarRefExp(stackVarDec));
				args->append_expression(buildAddOp(buildVarRefExp(depthVarDec), buildIntVal(1)));
				if (g_spliceOptimizationLevel != Splice_NoOrder) {
					// get field id of first call
					int fieldId = RecursiveField::getFieldId(getNodeVarName(orgCall));
					args->append_expression(buildAddOp(
							buildMultiplyOp(buildVarRefExp(indexVarDec), buildIntVal(numFields)), buildIntVal(fieldId)));
				}
				args->append_expression(buildVarRefExp(spliceSetVarDec));
				SgFunctionCallExp *newCall = buildFunctionCallExp(g_recurseSpliceSym, args);
				replaceExpression(orgCall, newCall);
			} else {
				removeStatement(getEnclosingStatement(orgCall));
			}
		}

		appendStatement(cset->stmts[0], ifBlock);

		if (g_useSOA) {
			// _nextBlock0->copy_up(_block);

			SgExprListExp *args = buildExprListExp();
			args->append_expression(buildVarRefExp(blockVarDec));
			SgMemberFunctionRefExp *funcRef = buildMemberFunctionRefExp(g_blockCopyUpSym, false, false);
			SgFunctionCallExp *recycleCall = buildFunctionCallExp(
					buildArrowExp(buildVarRefExp(nextBlockVarDecs[i]), funcRef), args);
			appendStatement(buildExprStatement(recycleCall), ifBlock);
		}
	}

	if (g_useSOA) {
		SOATransform::transform(blockBlock, getVarDecInitName(blockVarDec), getVarDecInitName(biVarDec));
	}

	delete [] nextBlockVarDecs;
}

void SpliceTransform::transformMapTreeIntermediary(const MethodInfo::IntermediaryInfo &minfo, bool last) {

	SgMemberFunctionDeclaration *memFunc = isSgMemberFunctionDeclaration(minfo.functionDec);

	// create map root intermediary function
	string funcName = minfo.functionDec->get_name();
	SgFunctionParameterList* parameterList = buildFunctionParameterList();
	SgFunctionDeclaration *func;
	if (memFunc != NULL) {
		SgMemberFunctionType *memFuncType = buildMemberFunctionType(buildVoidType(), NULL, NULL, 0);
		func = buildDefiningMemberFunctionDeclaration(
				funcName + g_suffixMapTree, memFuncType, parameterList, memFunc->get_class_scope());
	} else{
		func = buildDefiningFunctionDeclaration(
				funcName + g_suffixMapTree, buildVoidType(), parameterList, g_global);
	}

	SgBasicBlock* funcBody = func->get_definition()->get_body();
	insertStatementBefore(minfo.functionDec, func);

	// add params

	for (int i = 0; i < minfo.loopInvariantParams.size(); i++) {
		SgInitializedName *param = minfo.loopInvariantParams[i];
		SgInitializedName *newParam = buildInitializedName(param->get_name(), param->get_type());
		appendArg(parameterList, newParam);
	}
	SgInitializedName *spliceSetVarDec = buildInitializedName("_spliceSet", buildPointerType(g_spliceSetType));
	appendArg(parameterList, spliceSetVarDec);

	// add non-defining forward declaration if necessary
	addForwardFuncDecIfNecessary(minfo.functionDec, func);

	// ::find_correlation_mapTree(root, &_spliceSet);
	// correlation_search_mapTree(node, 0, 1, _set);

	SgExprListExp *args = getArgsExprList(minfo.loopInvariantArgs);
	if (last) {
		args->append_expression(buildIntVal(0));
		args->append_expression(buildIntVal(1));
	}
	args->append_expression(buildVarRefExp(spliceSetVarDec));

	SgStatement *callStmt = copyStatement(getEnclosingStatement(minfo.call));
	appendStatement(callStmt, funcBody);
	Rose_STL_Container<SgNode *> calls = NodeQuery::querySubTree(callStmt, V_SgFunctionCallExp);
	ROSE_ASSERT(calls.size() == 1);
	SgFunctionCallExp *call = isSgFunctionCallExp(calls[0]);
	call->set_args(args);
	SgExpression *callExp = call->get_function();
	SgDotExp *dot = isSgDotExp(callExp);
	if (dot != NULL) callExp = dot->get_rhs_operand();
	SgArrowExp *arrow = isSgArrowExp(callExp);
	if (arrow != NULL) callExp = arrow->get_rhs_operand();
	SgFunctionRefExp *funcRefExp = isSgFunctionRefExp(callExp);
	if (funcRefExp != NULL) funcRefExp->set_symbol(g_nextMapTreeSym);
	SgMemberFunctionRefExp *memFuncRefExp = isSgMemberFunctionRefExp(callExp);
	if (memFuncRefExp != NULL) {
		SgMemberFunctionSymbol *memFuncSym = isSgMemberFunctionSymbol(g_nextMapTreeSym);
		ROSE_ASSERT(memFuncSym != NULL);
		memFuncRefExp->set_symbol(memFuncSym);
	}
	ROSE_ASSERT(funcRefExp != NULL || memFuncRefExp != NULL);

	// save symbol to use in intermediary mapRoot functions
	g_nextMapTreeSym = isSgFunctionSymbol(func->search_for_symbol_from_symbol_table());
	ROSE_ASSERT(g_nextMapTreeSym != NULL);
}

void SpliceTransform::transformMapTree() {
	transformMapTreeRecursive();
	for (int i = g_methodInfo.intermediaryInfos.size() - 1; i >= 0; i--) {
		transformMapTreeIntermediary(*g_methodInfo.intermediaryInfos[i], i == g_methodInfo.intermediaryInfos.size() - 1);
	}
}

void SpliceTransform::optimizationLevelAnalysis(const RecursiveCallSet& rcs) {
	bool nodeBasedOrder = true;
	for (int phase = 0; phase < RecursiveField::recursiveFields.size(); phase++) {
		set<int> counter;
		for (int j = 0; j < rcs.callSets.size(); j++) {
			CallSet *cs = rcs.callSets[j];
			int fieldId = -1;
			if (phase < cs->calls.size()) {
				string fieldName = getNodeVarName(cs->calls[phase]);
				fieldId = RecursiveField::getFieldId(fieldName);
			}
			if (fieldId != -1) {
				pair<set<int>::iterator,bool> r = counter.insert(fieldId);
				nodeBasedOrder = r.second;
			}
			if (!nodeBasedOrder) break;
		}
		if (!nodeBasedOrder) break;
	}

	int optLevel;
	if (rcs.callSets.size() == 1) {
		optLevel = Splice_NoOrder;
	} else if (nodeBasedOrder) {
		optLevel = Splice_NodeBasedOrder;
	} else {
		optLevel = Splice_GeneralOrder;
	}

	cout << "possible splice optimization level: " << optLevel << endl;
	if (optLevel < g_spliceOptimizationLevel) {
		g_spliceOptimizationLevel = optLevel;
	}
	cout << "using splice optimization level: " << g_spliceOptimizationLevel << endl;
	ROSE_ASSERT(g_spliceOptimizationLevel > Splice_GeneralOrder); //TODO: implement general order
}

void SpliceTransform::transform() {
	findSpliceClasses();
	transformRecursive();
	// transformRecursive() must be before transformEntry() to have valid g_recurseSpliceSym
	transformEntry();
	transformMapTree();
}
