/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "Utils.h"
#include "Globals.h"

using namespace SageBuilder;
using namespace SageInterface;

string getIfStr(const vector<IfPart>& ifparts, string prefix) {
	string ret;
	int len = ifparts.size();
	ROSE_ASSERT(len > 0);
	if (ifparts.size() == 1) {
		const IfPart &ifpart = ifparts[0];
		for (int i = 0; i < ifpart.lines.size(); i++) {
			ret += prefix + ifpart.lines[i];
		}
	} else if (ifparts.size() == 2) {
		ret += prefix + "if (" + ifparts[0].cond + ") {\n";
		for (int i = 0; i < ifparts[0].lines.size(); i++) {
			ret += prefix + ifparts[0].lines[i];
		}
		ret += prefix + "} else {\n";
		for (int i = 0; i < ifparts[1].lines.size(); i++) {
			ret += prefix + ifparts[1].lines[i];
		}
		ret += prefix + "}\n";
	} else {
		ret += prefix + "if (" + ifparts[0].cond + ") {\n";
		for (int i = 0; i < ifparts[0].lines.size(); i++) {
			ret += prefix + ifparts[0].lines[i];
		}
		for (int j = 0; j < len - 1; j++) {
			ret += prefix + "} else if (" + ifparts[j].cond + ") {\n";
			for (int i = 0; i < ifparts[j].lines.size(); i++) {
				ret += prefix + ifparts[j].lines[i];
			}
		}
		ret += prefix + "} else {\n";
		for (int i = 0; i < ifparts[len - 1].lines.size(); i++) {
			ret += prefix + ifparts[len - 1].lines[i];
		}
		ret += prefix + "}\n";
	}
	return ret;
}

string getNodeVarName(SgFunctionCallExp *call) {
	const SgExpressionPtrList &args = call->get_args()->get_expressions();
	for (int i = 0; i < args.size(); i++) {
		if (i == g_nodeParamIdx) {
			SgArrowExp *arrow = isSgArrowExp(args[i]);
			ROSE_ASSERT(arrow != NULL);
			SgVarRefExp *rhs = isSgVarRefExp(arrow->get_rhs_operand());
			ROSE_ASSERT(rhs != NULL);
			return rhs->get_symbol()->get_name();
		}
	}
	ROSE_ASSERT(false);
	return string();
}

SgType *getBaseType(SgType *t) {
	while (true) {
		SgModifierType *mt = isSgModifierType(t);
		if (mt == NULL) break;
		t = mt->get_base_type();
	}
	return t;
}

SgFunctionCallExp *getEnclosingFunctionCall(SgNode *node) {
	while (node != NULL) {
		if (isSgFunctionCallExp(node) != NULL || isSgStatement(node) != NULL) break;
		else node = node->get_parent();
	}
	return isSgFunctionCallExp(node);
}

//SgPlusPlusOp *getEnclosingPlusPlusOp(SgNode *node) {
//	while (node != NULL) {
//		if (isSgPlusPlusOp(node) != NULL || isSgStatement(node) != NULL) break;
//		else node = node->get_parent();
//	}
//	return isSgPlusPlusOp(node);
//}

string getArgsString(const vector<SgVarRefExp *> &v) {
	vector<string> args;
	for (int i = 0; i < v.size(); i++) {
		SgVarRefExp *arg = v[i];
		args.push_back(arg->unparseToString());
	}
	return concatStrs(args);
}

SgExprListExp *getArgsExprList(const vector<SgVarRefExp *> &v) {
	SgExprListExp *args = buildExprListExp();
	for (int i = 0; i < v.size(); i++) {
		SgVarRefExp *arg = v[i];
		args->append_expression(copyExpression(arg));
	}
	return args;
}

SgExprListExp *getArgsExprList(const vector<SgInitializedName *> &v, SgScopeStatement *scope) {
	SgExprListExp *args = buildExprListExp();
	for (int i = 0; i < v.size(); i++) {
		SgInitializedName *arg = v[i];
		args->append_expression(buildVarRefExp(arg->get_name(), scope));
	}
	return args;
}

string getParamsString(const vector<SgInitializedName *> &v, bool refToPointer) {
	vector<string> params;
	for (int i = 0; i < v.size(); i++) {
		SgInitializedName *param = v[i];
		params.push_back(getTypeString(param->get_type(), refToPointer) + " " + param->get_name());
	}
	return concatStrs(params);
}

string getArgsString(const vector<SgInitializedName *> &v) {
	vector<string> args;
	for (int i = 0; i < v.size(); i++) {
		SgInitializedName *param = v[i];
		args.push_back(param->get_name());
	}
	return concatStrs(args);
}

string getArgsStringWithPrefix(const vector<SgInitializedName *> &v, string prefix) {
	vector<string> args;
	for (int i = 0; i < v.size(); i++) {
		SgInitializedName *arg = v[i];
		args.push_back(prefix + arg->get_name());
	}
	return concatStrs(args);
}

string concatStrs(const vector<string> &strs) {
	string ret;
	for (int i = 0; i < strs.size(); i++) {
		ret += strs[i];
		if (i != strs.size() - 1) ret += ", ";
	}
	return ret;
}

string getTypeString(const SgType* type, bool refToPointer) {
	// This function is useful for debugging
	// This is the most general case of a function to return a name for an IR node.
	ROSE_ASSERT(type != NULL);

	string returnName;

	const SgNamedType* namedType = isSgNamedType(type);
	if (namedType != NULL)
	{
		returnName = namedType->get_name().getString();
	}
	else
	{
		switch(type->variantT())
		{
		case V_SgPointerType:
		{
			const SgPointerType* pointerType = isSgPointerType(type);
			returnName = getTypeString(pointerType->get_base_type()) + "*";
			break;
		}

		case V_SgReferenceType:
		{
			const SgReferenceType* referenceType = isSgReferenceType(type);
			returnName = getTypeString(referenceType->get_base_type()) + (refToPointer ? "*" : "&");
			break;
		}

//		case V_SgArrayType:
//		{
//			const SgArrayType* arrayType = isSgArrayType(type);
//			returnName = "array_of_";
//			returnName += getTypeString(arrayType->get_base_type());
//			break;
//		}

		case V_SgModifierType:
		{
			const SgModifierType* modifierType = isSgModifierType(type);
			//returnName = getTypeString(&(modifierType->get_typeModifier()));
			returnName += getTypeString(modifierType->get_base_type());
			break;
		}

		default:
		{
			//cout << type->class_name() << endl;
			returnName = type->unparseToString();
		}
		}
	}

	return returnName;
}

bool hasVarUse(SgStatement *stmt, string varName) {
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(stmt, V_SgVarRefExp);
	for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
		SgVarRefExp *var = isSgVarRefExp(*i);
		//cout << var->unparseToString() << "\t" << varName << endl;
		if (var->get_symbol()->get_name() == varName) return true;
	}
	return false;
}

string getClassForwardDeclareString() {
	string code;
	//cout << g_global->get_file_info()->get_filenameString() << endl;
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(g_global, V_SgClassDeclaration);
	for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
		SgClassDeclaration *classDec = isSgClassDeclaration(*i);
		// somewhat a hack to forward declare only classes within the single input file
		// excluding the many classes coming from headers
		// TODO: how to handle this for multiple files?
		if (classDec->get_file_info()->get_filenameString() !=
				g_global->get_file_info()->get_filenameString()) continue;
//		Sg_File_Info *fi = classDec->get_file_info();
//		cout << classDec->get_name() << " " << fi->get_filenameString () <<
//				" " << fi->isCompilerGenerated () << endl;
		//prependStatement(buildNondefiningClassDeclaration(classDec->get_name(), g_global), g_global);
		code += buildNondefiningClassDeclaration(classDec->get_name(), g_global)->unparseToString() + "\n";
	}
	return code;
}

void addForwardFuncDecIfNecessary(SgFunctionDeclaration *orgFunc, SgFunctionDeclaration *newFunc) {
	SgMemberFunctionDeclaration *memOrgFunc = isSgMemberFunctionDeclaration(orgFunc);
	SgScopeStatement *scope;
	if (memOrgFunc == NULL) {
		scope = g_global;
	} else {
		scope = memOrgFunc->get_class_scope();
	}
	string orgName = orgFunc->get_name();
	//SgFunctionDeclaration *orgForwardDec = findFunctionDeclaration(scope, orgName, NULL, false);
	SgFunctionDeclaration *orgForwardDec = NULL;
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(scope, V_SgFunctionDeclaration);
	for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
		SgFunctionDeclaration *funcDec = isSgFunctionDeclaration(*i);
		if (funcDec->get_definition() != NULL) continue;
		if (funcDec->get_name() != orgName) continue;
		orgForwardDec = funcDec;
		break;
	}
	if (orgForwardDec != NULL) {
		SgFunctionParameterList *newList = buildFunctionParameterList();
		const SgInitializedNamePtrList &params = newFunc->get_args();
		for (SgInitializedNamePtrList::const_iterator i = params.begin(); i != params.end(); i++) {
			SgInitializedName *param = *i;
			newList->append_arg(buildInitializedName(param->get_name(), param->get_type()));
		}
		SgFunctionDeclaration *forwardDec;
		if (memOrgFunc == NULL) {
			forwardDec = buildNondefiningFunctionDeclaration(
					newFunc->get_name(), newFunc->get_type()->get_return_type(), newList, scope);
		} else {
			SgMemberFunctionDeclaration *memNewFunc = isSgMemberFunctionDeclaration(newFunc);
			ROSE_ASSERT(memNewFunc != NULL);
			int functionConstVolatileFlags = 0;
			// preserve const modifier
			if (isSgMemberFunctionType(memNewFunc->get_type())->isConstFunc()) functionConstVolatileFlags |= 0x1; //  e_const = 0x00000001
//			forwardDec = buildNondefiningMemberFunctionDeclaration(newFunc->get_name(),
//					buildMemberFunctionType(newFunc->get_type()->get_return_type(), NULL, NULL, 0),
//					newList, scope, NULL, functionConstVolatileFlags);
			forwardDec = buildNondefiningMemberFunctionDeclaration(newFunc->get_name(),
					newFunc->get_type()->get_return_type(), newList, scope, NULL, functionConstVolatileFlags);
			//forwardDec = buildNondefiningMemberFunctionDeclaration(memNewFunc);
		}
		insertStatementBefore(orgForwardDec, forwardDec);
	}
}

bool exprContainsVarName(SgExpression *expr, const set<string> &names) {
	vector<SgVariableSymbol *> vars = getSymbolsUsedInExpression(expr);
	for (int i = 0; i < vars.size(); i++) {
		string name = vars[i]->get_name();
		if (setContains(names, name)) return true;
	}
	return false;
}

bool exprContainsVarName(SgExpression *expr, const vector<string> &names) {
	vector<SgVariableSymbol *> vars = getSymbolsUsedInExpression(expr);
	for (int i = 0; i < vars.size(); i++) {
		string name = vars[i]->get_name();
		if (vecContains(names, name)) return true;
	}
	return false;
}

//bool stmtContainsVarName(SgStatement *root, const set<string> &names) {
//	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(root, V_SgVariableSymbol);
//	for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
//		SgVariableSymbol* sym = isSgVariableSymbol(*i);
//		string name = sym->get_name();
//		if (setContains(names, name)) return true;
//	}
//	return false;
//}

bool stmtContainsStmt(SgStatement *root, const vector<SgStatement *> &stmts) {
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(root, V_SgStatement);
	for (Rose_STL_Container<SgNode*>::iterator i = nodes.begin(); i != nodes.end(); ++i) {
		SgStatement* s = isSgStatement(*i);
		//cout << s->unparseToString() << endl << endl;
		if (vecContains(stmts, s)) return true;
	}
	return false;
}

// for control flow, if either path has loop variant stmts, the parent is also considered loop variant
bool gatherLoopVariantStmts(set<SgStatement *> *loopVariantControlStmts, set<string> *loopVariantArgsPropagation,
		SgStatement *stmt, const vector<SgStatement *> &recursionPointStmts) {
	//cout << loopVariantArgsPropagation->size() << endl;
	//cout << stmt->unparseToString() << endl << endl;
	bool ret = false;
	SgBasicBlock *block = isSgBasicBlock(stmt);
	SgExprStatement *exprStmt = isSgExprStatement(stmt);
	SgIfStmt *ifStmt = isSgIfStmt(stmt);
	SgForStatement *forStmt = isSgForStatement(stmt);
	SgVariableDeclaration *varDec = isSgVariableDeclaration(stmt);
	if (block != NULL) {
		const SgStatementPtrList &stmts = block->get_statements();
		for (int i = 0; i < stmts.size(); i++) {
			bool r = gatherLoopVariantStmts(loopVariantControlStmts, loopVariantArgsPropagation, stmts[i],
					recursionPointStmts);
			ret = r || ret;
		}
	} else if (varDec != NULL) {
		SgInitializedName *var;
		SgAssignInitializer *init = getVarDecInit(varDec, &var);
		if (exprContainsVarName(init, *loopVariantArgsPropagation)) {
			loopVariantArgsPropagation->insert(var->get_name());
		}
		if (exprContainsVarName(init, *loopVariantArgsPropagation) && !vecContains(recursionPointStmts, stmt)) {
			ret = true;
		}
	} else if (exprStmt != NULL) {
		SgExpression *expr = exprStmt->get_expression();
		SgAssignOp *assign = isSgAssignOp(expr);
		if (assign != NULL) {
			if (exprContainsVarName(assign->get_rhs_operand_i(), *loopVariantArgsPropagation)) {
				SgExpression *lhs = assign->get_lhs_operand_i();
				while (true) {
					SgBinaryOp *binary = isSgBinaryOp(lhs);
					if (binary == NULL) break;
					lhs = binary->get_lhs_operand_i();
				}
				SgVarRefExp *varRef = isSgVarRefExp(lhs);
				if (varRef != NULL) {
					loopVariantArgsPropagation->insert(varRef->get_symbol()->get_name());
				}
			}
		}
		if (exprContainsVarName(expr, *loopVariantArgsPropagation) && !vecContains(recursionPointStmts, stmt)) {
			ret = true;
		}
	} else if (ifStmt != NULL) {
		SgExpression *expr = isSgExprStatement(ifStmt->get_conditional())->get_expression();
		if (exprContainsVarName(expr, *loopVariantArgsPropagation)) {
			ret = true;
		}
		bool r = gatherLoopVariantStmts(loopVariantControlStmts, loopVariantArgsPropagation, ifStmt->get_true_body(),
				recursionPointStmts);
		ret = r || ret;
		if (ifStmt->get_false_body() != NULL) {
			bool r = gatherLoopVariantStmts(loopVariantControlStmts, loopVariantArgsPropagation,
					ifStmt->get_false_body(), recursionPointStmts);
			ret = r || ret;
		}
	} else if (forStmt != NULL) {
		bool r = gatherLoopVariantStmts(loopVariantControlStmts, loopVariantArgsPropagation, forStmt->get_loop_body(),
				recursionPointStmts);
		ret = r || ret;
	}
	if (ret) {
		loopVariantControlStmts->insert(stmt);
		//cout << stmt->unparseToString() << endl << endl;
	}
	return ret;
}

void removeLoopVariantStmts(vector<SgStatement *> *toRemoveStmts, set<string> *loopVariantArgsPropagation,
		SgStatement *stmt, const vector<SgStatement *> &recursionPointStmts) {
	SgBasicBlock *block = isSgBasicBlock(stmt);
	SgExprStatement *exprStmt = isSgExprStatement(stmt);
	SgIfStmt *ifStmt = isSgIfStmt(stmt);
	SgForStatement *forStmt = isSgForStatement(stmt);
	SgVariableDeclaration *varDec = isSgVariableDeclaration(stmt);
	if (block != NULL) {
		const SgStatementPtrList &stmts = block->get_statements();
		for (int i = 0; i < stmts.size(); i++) {
			removeLoopVariantStmts(toRemoveStmts, loopVariantArgsPropagation, stmts[i],
					recursionPointStmts);
		}
	} else if (varDec != NULL) {
		const SgInitializedNamePtrList &vars = varDec->get_variables();
		ROSE_ASSERT(vars.size() == 1); //TODO: handle multiple vars in single dec
		SgInitializedName *var = vars[0];
		if (exprContainsVarName(var->get_initializer(), *loopVariantArgsPropagation)) {
			toRemoveStmts->push_back(stmt);
			loopVariantArgsPropagation->insert(var->get_name());
		}
	} else if (exprStmt != NULL) {
		SgExpression *expr = exprStmt->get_expression();
		SgAssignOp *assign = isSgAssignOp(expr);
		if (assign != NULL) {
			if (exprContainsVarName(assign->get_rhs_operand_i(), *loopVariantArgsPropagation)) {
				SgExpression *lhs = assign->get_lhs_operand_i();
				while (true) {
					SgBinaryOp *binary = isSgBinaryOp(lhs);
					if (binary == NULL) break;
					lhs = binary->get_lhs_operand_i();
				}
				SgVarRefExp *varRef = isSgVarRefExp(lhs);
				if (varRef != NULL) {
					loopVariantArgsPropagation->insert(varRef->get_symbol()->get_name());
				}
			}
		}
		if (exprContainsVarName(expr, *loopVariantArgsPropagation) && !vecContains(recursionPointStmts, stmt)) {
			toRemoveStmts->push_back(stmt);
		}
	} else if (ifStmt != NULL) {
		SgExpression *expr = isSgExprStatement(ifStmt->get_conditional())->get_expression();
		SgStatement *thenStmt = ifStmt->get_true_body();
		SgStatement *elseStmt = ifStmt->get_false_body();
		if (exprContainsVarName(expr, *loopVariantArgsPropagation)) {
			if (stmtContainsStmt(thenStmt, recursionPointStmts)) {
				//TODO: this is tricky if then and else statements have different set of recursive calls
				//if (elseStmt != NULL && stmtContainsStmt(elseStmt, recursionPointStmts)) {
					removeLoopVariantStmts(toRemoveStmts, loopVariantArgsPropagation, thenStmt, recursionPointStmts);
					replaceStatement(ifStmt, thenStmt);
				//}
			} else if (elseStmt != NULL && stmtContainsStmt(elseStmt, recursionPointStmts)) {
				removeLoopVariantStmts(toRemoveStmts, loopVariantArgsPropagation, elseStmt, recursionPointStmts);
				replaceStatement(ifStmt, elseStmt);
			} else {
				toRemoveStmts->push_back(stmt);
			}
		} else {
			removeLoopVariantStmts(toRemoveStmts, loopVariantArgsPropagation, thenStmt, recursionPointStmts);
			if (elseStmt != NULL) {
				removeLoopVariantStmts(toRemoveStmts, loopVariantArgsPropagation, elseStmt, recursionPointStmts);
			}
		}
	} else if (forStmt != NULL) {
		removeLoopVariantStmts(toRemoveStmts, loopVariantArgsPropagation, forStmt->get_loop_body(), recursionPointStmts);
	} else {
		//cout << "unhandled removeLoopVariantStmts:" << stmt->class_name() << endl;
	}
}

void getRecursionPoints(SgFunctionDeclaration *funcDec, vector<SgFunctionCallExp *> *recursionPoints) {
	getRecursionPoints(funcDec->get_definition()->get_body(), recursionPoints, funcDec->get_name());
}

void getRecursionPoints(SgBasicBlock *body, vector<SgFunctionCallExp *> *recursionPoints, string orgFuncName) {

	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(body, V_SgFunctionCallExp);
	for (int i = 0; i < nodes.size(); i++) {
		SgFunctionCallExp *call = isSgFunctionCallExp(nodes[i]);
		string callName = call->getAssociatedFunctionSymbol()->get_name();
		//cout << callName << endl;
		if (callName != orgFuncName) continue;
		recursionPoints->push_back(call);
	}
	//cout << "getRecursionPoints\t" << funcDec->get_name() << "\t" << orgFuncName << "\t" << recursionPoints->size() << endl;
}

SgAssignInitializer *getVarDecInit(SgVariableDeclaration *varDec, SgInitializedName **varRet) {
	const SgInitializedNamePtrList &vars = varDec->get_variables();
	ROSE_ASSERT(vars.size() == 1); //TODO: handle multiple vars in single dec
	SgInitializedName *var = vars[0];
	SgAssignInitializer *ret = isSgAssignInitializer(var->get_initializer());
	//ROSE_ASSERT(ret != NULL);
	if (varRet != NULL) {
		*varRet = var;
	}
	return ret;
}

SgInitializedName *getVarDecInitName(SgVariableDeclaration *varDec) {
	const SgInitializedNamePtrList &vars = varDec->get_variables();
	ROSE_ASSERT(vars.size() == 1); //TODO: handle multiple vars in single dec
	return vars[0];
}

SgFunctionDefinition *getFunctionDefintion(string name) {
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(g_global, V_SgFunctionDefinition);
	for (Rose_STL_Container<SgNode*>::iterator j = nodes.begin(); j != nodes.end(); ++j) {
		SgFunctionDefinition *funcDef = isSgFunctionDefinition(*j);
		if (funcDef->get_declaration()->get_name() == name) return funcDef;
	}
	return NULL;
}

SgClassDefinition *getClassDefintion(string name) {
	Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(g_global, V_SgClassDefinition);
	for (Rose_STL_Container<SgNode*>::iterator j = nodes.begin(); j != nodes.end(); ++j) {
		SgClassDefinition *classDef = isSgClassDefinition(*j);
		if (classDef->get_declaration()->get_name() == name) return classDef;
	}
	return NULL;
}

SgNamedType *getNamedType(SgInitializedName *param) {
	SgPointerType *ptype = isSgPointerType(param->get_type());
	ROSE_ASSERT(ptype != NULL);
	SgNamedType *ntype = isSgNamedType(ptype->get_base_type());
	ROSE_ASSERT(ntype != NULL); //TODO: handle non-class points? e.g. int
	return ntype;
}

