/*
   SIMTree-modified, a framework to automatically transform task parallelism code to data parallelism.
   Youngjoon Jo (yjo@purdue.edu)
   Bin Ren (bin.ren@pnnl.gov)

   Copyright 2015, School of Electrical and Computer Engineering, Purdue University.
   */

#include "rose.h"
#include "RecursiveStructureAnalysis.h"
#include "MethodInfo.h"
#include "Globals.h"
#include "BlockTransform.h"
#include "SpliceTransform.h"
#include "BlockClasses.h"
#include "SpliceClasses.h"
#include "CommonTransform.h"
#include "IntermediaryStateClasses.h"
#include "RecursiveCallSet.h"
#include "RecursiveField.h"
#include "SOATransform.h"

struct timeval start_time, end_time;

int get_elapsed_usec() {
    int sec = end_time.tv_sec - start_time.tv_sec;
    int usec = end_time.tv_usec - start_time.tv_usec;
    return (sec * 1000000) + usec;
}

void printUsage() {
    cout << "SIMTree-modified, a framework to automatically transform task parallelism code to data parallelism." << endl;
    cout << "Youngjoon Jo (yjo@purdue.edu)" << endl;
    cout << "Bin Ren (bin.ren@pnnl.gov)" << endl;
    cout << endl;
    cout << "Copyright 2015, School of Electrical and Computer Engineering, Purdue University." << endl;
    cout << endl;
    cout << "./SIMTree [options] <target file>" << endl;
    cout << " --noreexpand            : do not generate reexpansion code in DF execution" << endl;
}

int main(int argc, char* argv[]) {

    int roseargstart;
    for (int i = 1; i < argc; i++) {
        char *arg = argv[i];

        if (strcmp(arg, "--noreexpand") == 0){
            g_useReexpand = false;
        }
        else {
            roseargstart = i;
            break;
        }
    }

    g_useSplice = false;//In case there is any place in the old code using these flags
    g_useSOA = false;

    int roseargc = argc - roseargstart + 1;
    char **roseargv = new char*[roseargc];
    roseargv[0] = argv[0];
    for (int i = roseargstart, j = 1; i < argc; i++, j++) {
        roseargv[j] = argv[i];
    }

    if (argc == 1) {
        printUsage();
        return 0;
    }

    SgProject* project = frontend(roseargc, roseargv);
    ROSE_ASSERT(project != NULL);

    Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(project, V_SgGlobal);
    ROSE_ASSERT(nodes.size() == 1);
    g_global = isSgGlobal(nodes[0]);

    RecursiveStructureAnalysis rsa;
    bool ret = rsa.findRecursiveTraversals(project);
    if (!ret) {
        cout << "no recursive structure, cannot transform" << endl;
        return 0;
    }

    g_methodInfo.analyze();

    for(int i = 0; i < g_num_recursiveCall; ++i){
        ostringstream ss;
        ss << i;
        g_suffixBlock_DF_i.push_back("_block_df" + ss.str());
    }

    gettimeofday(&start_time, NULL);
    string inputFile = roseargv[1];
    bool isSecondPass = inputFile.find("rose_") != string::npos;
    if (!isSecondPass) {
        CommonTransform::markAfterClassDecPragma();
        BlockTransform::addRecursivePrototype();
        BlockClasses::generate(1);
        cout << "generated block classes, ready for second pass" << endl;
    } else {
        BlockTransform::findClasses();
        BlockTransform::transform(false);
        cout << "second pass done" << endl;
    }

    gettimeofday(&end_time, NULL);
    int runtime = get_elapsed_usec() / 1000;
    printf("[runtime:%d]\n", runtime);

    int retval = backend(project);
    return retval;
}






